/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/control/client_pamauth.c,v 1.1.1.1 1999/12/09 01:16:32 koni Exp $ 
  $Author: koni $ $Date: 1999/12/09 01:16:32 $ $Name: slan-release-0-04 $
  $Revision: 1.1.1.1 $

*/
   
#ifndef WIN32
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <syslog.h>
#else
#include "slan_windows.h"
#endif

#include "slan_types.h"
#include "slan_alloc.h"

#include "slan_session.h"
#include "control_tcp.h"
#include "client_pamauth.h"
#include "log_message.h"

#ifdef linux
client_resp_t *linux_terminal_dialog(slan_session_t *, 
				     client_authmsg_t *, 
				     int);
client_resp_t *roboconnect_dialog(slan_session_t *, 
				  client_authmsg_t *, 
				  int);
#ifdef ROBO_CLIENT
client_resp_t *(*client_authenticate_dialog)() = roboconnect_dialog;
#else
client_resp_t *(*client_authenticate_dialog)() = linux_terminal_dialog;
#endif
#else
# ifdef WIN32

#include "slan_config.h"

extern slan_conf_t *slan_global_config;

static client_resp_t *windows_dialog_dialog(slan_session_t *,
                                            client_authmsg_t *,
                                            int);
client_resp_t *(*client_authenticate_dialog)() = windows_dialog_dialog;

# else
# error No client_auth_dialog!
# endif
#endif

#define CLIENTAUTH_TIMEMAX (1200000)

static int client_pamauth(slan_session_t *session_obj) {
  control_msg_t *msg;
  client_resp_t *resp_msg;
  client_authmsg_t *auth_msg;
  int num_msg, i, timeout, rval;
  uint32 length;
  uchar *msg_data;

  for(;;) {
    num_msg = 0;
    auth_msg = NULL;
    
    timeout = CLIENTAUTH_TIMEMAX;
    while(1) {
      rval = control_msg_recv(&msg, session_obj, &timeout);
      if (rval < 0) return CLIENTAUTH_TCP_ERROR;
      if (rval > 0) return CLIENTAUTH_ETIMEOUT;
      if (msg->type == CLIENTAUTH && msg->subtype == PAM_SENDCLIENT) break;
      if (msg->type == CLIENTAUTH && msg->subtype == AUTH_OKAY) {
	control_msg_destruct(msg);
	return 0;
      }
      if (msg->type == CLIENTAUTH && msg->subtype == AUTH_REJECT) {
	control_msg_destruct(msg);
	logmsg(MSG_CRITICAL,"Client authentication failed.\n");
	return 1;
      }
      logmsg(MSG_DEBUG,"Received unexpected control message in client_pamauth() "
	     "protocol = %0lX, type = %0lX\n",msg->type, msg->subtype);
      control_msg_destruct(msg);
    }
    
    /* This basically reconstructs the pam conversation message array
       that was passed to the server, which we will pass to another function 
       that will figure out how to display the prompts/dialogs, etc.*/
    msg_search_tlv(msg, NUM_MSG, &length, &msg_data);
    if (length!=4) goto fail;
    
    memcpy(&num_msg, msg_data, 4); 
    num_msg = ntohl(num_msg);
    auth_msg = (client_authmsg_t *) slan_calloc(num_msg,sizeof(client_authmsg_t));
    if (auth_msg == NULL) {
      logmsg(MSG_ERROR,"Failed allocating memory in client_pamauth() (%d bytes)\n",
	     sizeof(client_authmsg_t *));
      control_msg_destruct(msg);
      return 1;
    }
    
    for(i=0;i<num_msg;i++) {
      uint32 type;

      msg_search_tlv(msg, PAM_MSG + i, &length, &msg_data);
      if (msg_data == NULL || length != 4) goto fail;
      memcpy(&type, msg_data, 4); 
      type = ntohl(type);
      
      msg_search_tlv(msg, MSG_DATA + i, &length, &msg_data);
      if (msg_data == NULL) goto fail;
      auth_msg[i].type = type;
      auth_msg[i].data = msg_data;
    }
    
    /* This function is set (by a pointer) in a header file somewhere,
       it is platform dependant */
    resp_msg = (*client_authenticate_dialog)(session_obj, auth_msg, num_msg);
    if (resp_msg == NULL) goto fail;
    control_msg_destruct(msg);
    msg = new_control_msg(CLIENTAUTH, PAM_CLIENTRESP, session_obj);
    for(i=0;i<num_msg;i++) {
      if (resp_msg[i] != NULL) 
	msg_append_tlv(msg, PAM_RESPONSE + i, strlen(resp_msg[i]),
		       (uchar *) resp_msg[i]);
    }
  
    timeout = CLIENTAUTH_TIMEMAX;
    rval = control_msg_send(msg, session_obj, &timeout);
    control_msg_destruct(msg);
    /* Cleanup mess from doing this stupid dialog shit. */
    if (num_msg > 0) {
      for(i=0;i<num_msg;i++) {
	if (resp_msg[i] != NULL) slan_memfree(resp_msg[i]);
      }
      slan_memfree(auth_msg);
      slan_memfree(resp_msg);
    }
    
    if (rval < 0) return CLIENTAUTH_TCP_ERROR;
    if (rval > 0) return CLIENTAUTH_ETIMEOUT;
  }
  
 fail:
  logmsg(MSG_ERROR,"Client authentication failure, local problem (unknown)\n");
  if (num_msg > 0) {
    for(i=0;i<num_msg;i++) {
      if (auth_msg[i].data != NULL) slan_memfree(auth_msg[i].data);
    }
    slan_memfree(auth_msg);
  }
  return 1;
}

int client_clientauth_wait(slan_session_t *session_obj) {
  control_msg_t *msg;
  int rval, timeout;

  timeout = CLIENTAUTH_TIMEMAX;
  while(1) {
    rval = control_msg_recv(&msg, session_obj, &timeout);
    if (rval < 0) return CLIENTAUTH_TCP_ERROR;
    if (rval > 0) return CLIENTAUTH_ETIMEOUT;
    if (msg->type == CLIENTAUTH && msg->subtype == AUTH_REQUEST) break;
    if (msg->type == CLIENTAUTH && msg->subtype == AUTH_SKIP) {
      control_msg_destruct(msg);
      return 0;
    }
    logmsg(MSG_DEBUG,"Received unexpected control message in client_pamauth() "
	   "protocol = %0lX, type = %0lX\n",msg->type, msg->subtype);
    control_msg_destruct(msg);
  }

  control_msg_destruct(msg);
  msg = new_control_msg(CLIENTAUTH, AUTH_READY, session_obj);
  timeout = CLIENTAUTH_TIMEMAX;
  rval = control_msg_send(msg, session_obj, &timeout);
  control_msg_destruct(msg);
  if (rval < 0) return CLIENTAUTH_TCP_ERROR;
  if (rval > 0) return CLIENTAUTH_ETIMEOUT;
      
  return client_pamauth(session_obj);
}

#ifdef WIN32

client_resp_t *windows_dialog_dialog(slan_session_t *session_obj, 
				     client_authmsg_t *msg, int n_msg) {
  int i;
  client_resp_t *resp;

  if (n_msg <= 0) return NULL;
 
  /* Print out infomational and error/warning messages first. */
  for(i=0;i<n_msg;i++) {
    if (msg[i].type == INFORMATION || msg[i].type == WARNING) {
      fprintf(stdout,"%s\n",msg[i].data);
    }
  }
  fflush(stdout);

  /* Alloocate response array and set all response pointers to NULL */
  resp = slan_calloc(n_msg, sizeof(client_resp_t));

  /* Get the responses */
  for(i=0;i<n_msg;i++) 
  {
    if (msg[i].type == PROMPT_ECHO) 
	{
		char *buf = slan_malloc(128);
		get_data(NULL, msg[i].data, 
			slan_global_config->node_identifier ?
			slan_global_config->node_identifier : "SLAN Server Login", 
			buf, 128, FALSE);
		resp[i] = buf;
    }

    if (msg[i].type == PROMPT_NOECHO) 
	{
		char *buf = slan_malloc(128);
		get_data(NULL, msg[i].data, 
			slan_global_config->node_identifier ?
			slan_global_config->node_identifier : "SLAN Server Login", 
			buf, 128, TRUE);
		resp[i] = buf;
    }
  }

  return resp;
}

#endif

/* Linux specific user-dialog function, for use with terminal I/O to get the
   user's response to PAM's questions. Can be overriden in client_pamauth.h
   with a GUI version or whatever you want. */
#ifdef linux
static unsigned char *safe_fgets(FILE *f) {
  int input_length = 0;
  unsigned char *input = NULL;
  int i;
  
  for(;;) {
    input = slan_realloc(input, input_length + 40);
    for(i=0;i<40;i++) {
      if ((input[input_length++] = getc(f)) == '\n') {
	input[input_length-1] = 0;  
	return input;
      }
    }
  }
}

client_resp_t *linux_terminal_dialog(slan_session_t *session_obj, 
					    client_authmsg_t *msg, int n_msg) {
  int i;
  client_resp_t *resp;

  if (n_msg <= 0) return NULL;
 
  /* Print out infomational and error/warning messages first. */
  for(i=0;i<n_msg;i++) {
    if (msg[i].type == INFORMATION || msg[i].type == WARNING) {
      fprintf(stdout,"%s\n",msg[i].data);
    }
  }
  fflush(stdout);

  /* Alloocate response array and set all response pointers to NULL */
  resp = slan_calloc(n_msg, sizeof(client_resp_t));

  /* Get the responses */
  for(i=0;i<n_msg;i++) {
    if (msg[i].type == PROMPT_ECHO) {
      fprintf(stdout,"%s ",msg[i].data); fflush(stdout);
      resp[i] = safe_fgets(stdin);
    }
    if (msg[i].type == PROMPT_NOECHO) {
      char *p;
      p = getpass(msg[i].data);
      resp[i] = strdup(p);
      /* getpass() returns a pointer to an internal buffer, as annoying as 
	 that is, so we zero it explicitly. */
      memset(p,0x0,strlen(p));
    }
  }

  return resp;
}

client_resp_t *roboconnect_dialog(slan_session_t *session_obj, 
					 client_authmsg_t *msg, int n_msg) {
  int i;
  client_resp_t *resp;

  if (n_msg <= 0) return NULL;
 
  /* Print out infomational and error/warning messages first. */
  for(i=0;i<n_msg;i++) {
    if (msg[i].type == INFORMATION || msg[i].type == WARNING) {
      fprintf(stdout,"%s\n",msg[i].data);
    }
  }
  fflush(stdout);

  /* Alloocate response array and set all response pointers to NULL */
  resp = slan_calloc(n_msg, sizeof(client_resp_t));

  /* Get the responses */
  for(i=0;i<n_msg;i++) {
    if (msg[i].type == PROMPT_ECHO) {
      resp[i] = strdup(session_obj->username);
    }
    if (msg[i].type == PROMPT_NOECHO) {
      resp[i] = strdup(session_obj->password);
    }
  }

  return resp;
}

#endif
