/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/control/control_tcp.c,v 1.2 2002/04/02 02:07:05 koni Exp $ 
  $Author: koni $ $Date: 2002/04/02 02:07:05 $ $Name: slan-release-0-04 $
  $Revision: 1.2 $

*/
   
#ifdef WIN32
#include "slan_windows.h"
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <syslog.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/poll.h>
#include <sys/ioctl.h>
#endif

#include "slan_types.h"
#include "slan_alloc.h"

#include "control_tcp.h"
#include "log_message.h"
#include <assert.h>

/* This function opens, binds, and sets a TCP/IP socket to listen for
   incomming connections. */
SOCKET server_open_socket(int port) {
  struct sockaddr_in address;
  SOCKET tcp_socket;
  int optval;

  tcp_socket = socket(PF_INET, SOCK_STREAM, 0);
  optval = 1;
  setsockopt(tcp_socket, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));
  if (tcp_socket < 0) {
    perror("Can not start SLAN: Can not open TCP socket: ");
#ifndef WIN32
    syslog(LOG_ERR,"Can not start SLAN: Can not open TCP socket (%m)\n");
#endif
    exit(-1);
  }
  
  memset(&address, 0, sizeof(struct sockaddr_in));

  address.sin_family = AF_INET;
  address.sin_addr.s_addr = htonl(INADDR_ANY);
  address.sin_port = htons(port);

  if (bind(tcp_socket, (struct sockaddr *)&address, sizeof(struct sockaddr_in))) {
    perror("Can not start SLAN: Can not bind TCP socket: ");
#ifndef WIN32
    syslog(LOG_CRIT,"Can not start SLAN: Can not bind TCP socket (%m)\n");
#endif

    exit(-1);
  }

  if (listen(tcp_socket, 100)) {
    perror("Can not start SLAN: Cannot make TCP socket listen for "
	   "connections: ");
#ifndef WIN32
    syslog(LOG_CRIT,"Can not start SLAN: Can not make TCP socket listen for "
	   "connections (%m)\n");
#endif
    exit(-1);
  }
  return tcp_socket;
}

/* This function opens a socket for a client, and attempts to connect
   to the server specified by name or ip_address "server_name" */
tcp_socket_t *client_open_socket(unsigned char *server_name, 
			  unsigned int *server_binaddr,
			  int port) {
  tcp_socket_t *tcp_socket;
  SOCKET fd;
  struct sockaddr_in server_addr;
  struct hostent *hostname_hent;

  fd = socket(PF_INET, SOCK_STREAM, 0);
  if (tcp_socket < 0) {
    logmsg(MSG_CRITICAL,"Can not start SLAN: Can not open TCP socket (%s)\n",
	   strerror(errno));
    LEAVE_THREAD(-1);
  }

  memset(&server_addr, 0, sizeof(struct sockaddr_in));

  if((server_addr.sin_addr.s_addr = inet_addr(server_name)) == INADDR_NONE) {

    hostname_hent = gethostbyname(server_name);
    if (hostname_hent == NULL) {
      logmsg(MSG_CRITICAL,"Failed to resolve server hostname "
	     "%s (%s)\n",server_name,strerror(errno));
      LEAVE_THREAD(-1);
    }

    memcpy(&server_addr.sin_addr.s_addr,hostname_hent->h_addr_list[0],
  	   hostname_hent->h_length);
  }

  *server_binaddr = server_addr.sin_addr.s_addr;

  server_addr.sin_port = htons(port);
  server_addr.sin_family = AF_INET;

  if (connect(fd, (struct sockaddr *) &server_addr, sizeof(struct sockaddr_in))) {
    logmsg(MSG_CRITICAL,"Failed to connect to server at %s (%s)\n",
	   server_name,strerror(errno));
    LEAVE_THREAD(-1);
  }

  tcp_socket = malloc(sizeof(tcp_socket_t));
  if (tcp_socket == NULL) {
    logmsg(MSG_CRITICAL,"Failed allocating memory in client_open_socket() "
	   "(%d bytes)\n",sizeof(tcp_socket_t));
    LEAVE_THREAD(-1);
  }

  tcp_socket->fd = fd;
  tcp_socket->start = 0;
  tcp_socket->end = 0;
  tcp_socket->tmp = 0;

  return tcp_socket;
}


int tcp_socket_send(tcp_socket_t *tcp_socket, uchar *message, int length,
		    int *timeout) {
  struct pollfd ufds;
  int rval, bytes_sent, etime;
  struct timeval tv, now;

  ufds.fd = tcp_socket->fd;
  ufds.events = POLLOUT;
  ufds.revents = 0;
  
  while(length > 0) {
    gettimeofday(&tv,NULL);
    rval = poll(&ufds, 1, *timeout);
    if (rval < 0) {
      logmsg(MSG_WARNING,"Error polling tcp socket (%s)\n",strerror(errno));
      return 1;
    }
    
    if (rval == 0) {
      *timeout = 0;
      return 1;
    }

    if (ufds.revents & (POLLERR | POLLHUP | POLLNVAL)) {
      logmsg(MSG_WARNING,"Polling tcp socket returned error\n");
      return 1;
    }

    bytes_sent = send(tcp_socket->fd, message, length, MSG_NOSIGNAL | MSG_DONTWAIT);
    if (bytes_sent < 0) {
      logmsg(MSG_WARNING,"Error sending data on tcp socket (%s)\n",strerror(errno));
      return 1;
    }
    length -= bytes_sent;

    gettimeofday(&now, NULL);
    etime = (now.tv_sec - tv.tv_sec)*1000000 + (now.tv_usec - tv.tv_usec);
    *timeout -= etime/1000;
    if (*timeout <= 0) { 
      *timeout = 0;
      if (length > 0) return 1;
      else return 0;
    }
  }

  return 0;
}

int control_msg_send(control_msg_t *msg, slan_session_t *session_obj, 
		     int *timeout) {
  tlv_t *tlv;
  uchar *p, *q, *buf;
  int message_length;
  uint32 x;
  tcp_socket_t *tcp_socket;

  /* Message layout:

     Control Constant
     Total length (including this word and constant above)
     Message Authentication Code
     Sequence Number
     Protocol Type
     Message Subtype
     TLVs
  */
  
  tcp_socket = session_obj->control_tcp;

  /* Account the space needed at the top for constant and length */
  message_length = 2*sizeof(uint32);

  tlv = msg->tlv_list;
  while(tlv!=NULL) {
    message_length += tlv->length + 2*sizeof(uint32);
    tlv = tlv->next;
  }

  /* Reserve space for authentication code */
  message_length += session_obj->mac->hash_size;

  /* reserve space for sequence number */
  message_length += sizeof(uint32);

  /* reserve space for protocol type and message subtype */
  message_length += sizeof(uint32)*2;

  buf = malloc(message_length + session_obj->mac->header_room + 
	       session_obj->mac->trailer_room);
  if (buf == NULL) {
    logmsg(MSG_ERROR,"Failed allocating memory in control_msg_send() (%d bytes)\n",
	   message_length + session_obj->mac->header_room + 
	   session_obj->mac->trailer_room);
    return 1;
  }

  p = buf + session_obj->mac->header_room + session_obj->mac->hash_size + 
    sizeof(uint32)*2;

  x = htonl(session_obj->control_seq_out);
  memcpy(p, &x, sizeof(uint32));
  p += sizeof(uint32);
  
  x = htonl(msg->type);
  memcpy(p, &x, sizeof(uint32));
  p += sizeof(uint32);

  x = htonl(msg->subtype);
  memcpy(p, &x, sizeof(uint32));
  p += sizeof(uint32);

  tlv = msg->tlv_list;
  while(tlv!=NULL) {
    x = htonl(tlv->type);
    memcpy(p, &x, sizeof(uint32));
    p += sizeof(uint32);
    
    x = htonl(tlv->length);
    memcpy(p, &x, sizeof(uint32));
    p += sizeof(uint32);

    memcpy(p, tlv->value, tlv->length);
    p += tlv->length;

    tlv = tlv->next;
  }

  q = buf + session_obj->mac->header_room + session_obj->mac->hash_size + 
    sizeof(uint32)*2;

  /* Compute and prepend the MAC */
  session_obj->mac->message_nocopy_sign(session_obj->mac, q, p - q,
					q - session_obj->mac->hash_size);
  q -= session_obj->mac->hash_size;
  session_obj->cipher->cbc_encrypt(session_obj->cipher, q, p - q);

  q -= sizeof(uint32);
  x = htonl(message_length);
  memcpy(q, &x, sizeof(uint32));
  
  x = htonl(CONTROL_CONSTANT);
  q -= sizeof(uint32);
  memcpy(q, &x, sizeof(uint32));

  assert((p-q) == message_length);
  
  if (tcp_socket_send(tcp_socket, q, message_length, timeout)) {
    logmsg(MSG_WARNING,"Failed to send control message for session %lu\n",
	   session_obj->id);
    free(buf);
    return 1;
  } 
    
  session_obj->control_seq_out++;
  free(buf);
  return 0;
}

static int tcp_socket_recvdata(tcp_socket_t *tcp_socket, int *timeout, 
			       int expect) {
  int etime;
  struct timeval tv, now;
  int size_ready, recv_bytes, rval;

  if (tcp_socket->end - tcp_socket->start >= expect) return 0;

  /* First, always move data to the beginning of the buffer. This simplifies a
     the complexity greatly, but wastes a little time */
  if (tcp_socket->start != 0) {
    int dist;

    dist = tcp_socket->start;
    memmove(tcp_socket->buf, tcp_socket->buf + tcp_socket->start,
	    tcp_socket->end - tcp_socket->start);
    tcp_socket->start = 0;
    tcp_socket->end -= dist;
    tcp_socket->tmp -= dist;
  }

  if (ioctl(tcp_socket->fd, FIONREAD, &size_ready)) {
    logmsg(MSG_ERROR,"Error on ioctl() in tcp_socket_recvdata() (%s)\n",
	   strerror(errno));
    return -1;
  }

  if (size_ready + tcp_socket->end > CONTROL_MSG_MAX_SIZE) {
    size_ready = CONTROL_MSG_MAX_SIZE - tcp_socket->end;
  }

  /* This is iterated, because of all the flap in the man pages about calls 
     returning less data than expected because a signal was caught, and I
     don't really understand what is happening with that. So, I just keep
     trying to recv() until we get it all. */
  while(size_ready > 0) {
    recv_bytes = recv(tcp_socket->fd, tcp_socket->buf + tcp_socket->end, 
		      size_ready, 0);
    if (recv_bytes < 0) {
      return -1;
    }
    size_ready -= recv_bytes;
    tcp_socket->end += recv_bytes; 
  }

  if (tcp_socket->end - tcp_socket->start >= expect) return 0;

  /* Now we actually do the poll... */
  while(tcp_socket->end - tcp_socket->start < expect) {
#if 0
    struct pollfd ufds;
    ufds.fd = tcp_socket->fd;
    ufds.events = POLLIN | POLLPRI;
    ufds.revents = 0;
    
    gettimeofday(&tv,NULL);
    rval = poll(&ufds, 1, *timeout);
    if (rval < 0) {
      logmsg(MSG_WARNING,"Error polling tcp socket (%s)\n",strerror(errno));
      return -1;
    }

    if (rval == 0) {
      *timeout = 0;
      return 1;
    }

    if (ufds.revents & (POLLERR | POLLHUP | POLLNVAL)) {
      logmsg(MSG_WARNING,"Polling tcp socket returned error\n");
      return -1;
    }
#else
    fd_set readfds;
    fd_set errfds;
    struct timeval stv;
    
    gettimeofday(&tv,NULL);
    FD_ZERO(&readfds);
    FD_ZERO(&errfds);
    FD_SET(tcp_socket->fd, &readfds);
    FD_SET(tcp_socket->fd, &errfds);
    stv.tv_sec = *timeout/1000;
    stv.tv_usec = ((*timeout*1000) % 1000000);
    
    rval = select(tcp_socket->fd+1, &readfds, NULL, &errfds, &stv);
    if (rval < 0) {
      logmsg(MSG_ERROR,"Error on select() in tcp_socket_recvdata() (%s)\n",
	     strerror(errno));
      return -1;
    }
    
    if (rval == 0) {
      *timeout = 0;
      return 1;
    }
    
    if (FD_ISSET(tcp_socket->fd, &errfds)) {
      logmsg(MSG_ERROR,"Error on tcp socket in tcp_socket_recvdata()\n");
      return -1;
    }
#endif

    if (ioctl(tcp_socket->fd, FIONREAD, &size_ready)) {
      logmsg(MSG_ERROR,"Error on ioctl() in tcp_socket_recvdata() (%s)\n",
	     strerror(errno));
      return -1;
    }

    if (size_ready <= 0) {
      /* If select returned with the fds set, the we should have data ready. 
	 But, as I have observered, this doesn't seem to always be true,
	 especially if the other end has "hung up". Really, from what the man
	 pages say, I'd expect that the errfds set should be set for select
	 and that poll should set POLLERR or POLLHUP. It does not. I don't
	 know why! -- so this is fail-safe measure to prevent 100% CPU race
	 on a socket that thinks it can be read, but in fact it can't. We
	 return -1 to indicate TCP error. */
      logmsg(MSG_ERROR,"select()/poll() indicates that data is ready but "
	     "that data can not be read. Assuming TCP error.\n");
      return -1;
    }
    
    if (size_ready + tcp_socket->end > CONTROL_MSG_MAX_SIZE) {
      size_ready = CONTROL_MSG_MAX_SIZE - tcp_socket->end;
    }
    
    /* This is iterated, because of all the flap in the man pages about calls 
       returning less data than expected because a signal was caught, and I
       don't really understand what is happening with that. So, I just keep
       trying to recv() until we get it all. */
    while(size_ready > 0) {
      recv_bytes = recv(tcp_socket->fd, tcp_socket->buf + tcp_socket->end, 
			size_ready, 0);
      if (recv_bytes < 0) {
	return -1;
      }
      size_ready -= recv_bytes;
      tcp_socket->end += recv_bytes; 
    }
    
    gettimeofday(&now, NULL);
    etime = (now.tv_sec - tv.tv_sec)*1000000 + (now.tv_usec - tv.tv_usec);
    *timeout -= etime/1000;
    if (*timeout <= 0) { 
      *timeout = 0;
      if (tcp_socket->end - tcp_socket->start < expect) return 1;
      else return 0;
    }  
  }   
  
  return 0;
}

int control_msg_recv(control_msg_t **r_msg, slan_session_t *session_obj, 
		     int *timeout) {
  int i;
  uint32 x, message_length, hash_length, seq;
  uchar *local_hash, *p;
  tcp_socket_t *tcp_socket;
  control_msg_t *msg;
  tlv_t *tlv;

  tcp_socket = session_obj->control_tcp;
  *r_msg = NULL;
  x = htonl(CONTROL_CONSTANT);

 start_over:
  /* Start by filling our buffer with any waiting data */
  if (tcp_socket_recvdata(tcp_socket, timeout, sizeof(uint32)*2)) {
    if (*timeout != 0) {
      int s_err;
      s_err = errno;
      logmsg(MSG_WARNING,"Error on TCP socket read for session %lu (%s)\n",
	       session_obj->id, strerror(errno));
      //tcp_error(s_err, tcp_socket, session_obj);
      return -1;
    } else {
      return 1;
    }
  }

  /* Start searching for a needle in the haystack */
  while(tcp_socket->tmp < (tcp_socket->end - sizeof(uint32))) {
    if (memcmp(tcp_socket->buf + tcp_socket->tmp, &x, sizeof(uint32))==0) break;
    tcp_socket->tmp++;
  } 

  if (tcp_socket->tmp == tcp_socket->end) {
    tcp_socket->start = tcp_socket->tmp;
    goto start_over;
  }

  /* OK, found the indicator for the control message, see if we can find the
     entire message */
  tcp_socket->start = tcp_socket->tmp;
  if (tcp_socket->end - tcp_socket->tmp < sizeof(uint32)*2) {
    if (tcp_socket_recvdata(tcp_socket, timeout, sizeof(uint32)*2)) {
      if (timeout != 0) {
	int s_err;
	s_err = errno;
	logmsg(MSG_WARNING,"Error on TCP socket read for session %lu (%s)\n",
	       session_obj->id, strerror(errno));
	//tcp_error(s_err, tcp_socket, session_obj);
	return -1;
      } else {
	return 1;
      }
    }
  }

    
  p = tcp_socket->buf + tcp_socket->tmp + sizeof(uint32);
  memcpy(&message_length, p, sizeof(uint32));
  message_length = ntohl(message_length);
    
  if (message_length<CONTROL_MSG_MIN_SIZE || message_length>CONTROL_MSG_MAX_SIZE) {
    logmsg(MSG_WARNING,"Attempt to receive improper size control message for "
	   "session %lu (%d bytes)\n",session_obj->id, message_length);
    tcp_socket->start = tcp_socket->tmp + sizeof(uint32)*2;
    goto start_over;
  }

  /* Try to get the rest of the message */
  if (tcp_socket->end - tcp_socket->tmp < message_length) {
    if (tcp_socket_recvdata(tcp_socket, timeout, sizeof(uint32)*2)) {
      if (timeout != 0) {
	int s_err;
	s_err = errno;
	logmsg(MSG_WARNING,"Error on TCP socket read for session %lu (%s)\n",
	       session_obj->id, strerror(errno));
	//tcp_error(s_err, tcp_socket, session_obj);
	return -1;
      } else {
	return 1;
      }
    }
  }

  /* Verify the message */
  hash_length = session_obj->mac->hash_size;
  local_hash = malloc(hash_length);
  if (local_hash == NULL) {
    /* This had better not happen or we are really in trouble. We'll return as
       if there is a TCP error which should end this session. */
    logmsg(MSG_ERROR,"Failed allocating memory in control_msg_recv() (%d bytes)\n",
	   hash_length);
    return -1;
  }

  p = tcp_socket->buf + tcp_socket->tmp + sizeof(uint32)*2;
  session_obj->cipher->cbc_decrypt(session_obj->cipher, p, 
				   message_length - sizeof(uint32)*2);

  p = tcp_socket->buf + tcp_socket->tmp + sizeof(uint32)*2 + hash_length;
  session_obj->mac->message_sign(session_obj->mac, p, message_length - 
				 (sizeof(uint32)*2 + hash_length), local_hash);
  p = tcp_socket->buf + tcp_socket->tmp + sizeof(uint32)*2;
  for(i=0;i<hash_length;i++) {
    if (p[i] != local_hash[i]) {
      free(local_hash);
      tcp_socket->start = tcp_socket->tmp = tcp_socket->end;
      logmsg(MSG_WARNING,"Message authentication code verificiation failed for "
	     "session id %lu\n",session_obj->id);
      return 1;
    }
  }
  
  p = tcp_socket->buf + tcp_socket->tmp + sizeof(uint32)*2 + hash_length;
  memcpy(&seq, p, sizeof(uint32));
  seq = ntohl(seq);
  
  if (seq != session_obj->control_seq_in) {
    free(local_hash);
    tcp_socket->start = tcp_socket->tmp = tcp_socket->end;
    logmsg(MSG_WARNING,"Dropped out of sequence control message for session id"
	   " %lu (%0lX received, expected %0lX)\n", session_obj->id,
	   seq, session_obj->control_seq_in);
    return 1;
  }
  session_obj->control_seq_in++;

  p = tcp_socket->buf + tcp_socket->tmp + sizeof(uint32)*3 + hash_length;
  free(local_hash);

  /* Parse it up into a control_msg_t struct and return it, clearing the
     buffer. */
  msg = malloc(sizeof(control_msg_t));
  if (msg == NULL) {
    logmsg(MSG_ERROR,"Failed allocating memory in control_msg_recv() (%d bytes)\n",
	   sizeof(control_msg_t));
    return -1;
  }

  memcpy(&x, p, sizeof(uint32));
  p += sizeof(uint32);
  msg->type = ntohl(x);
  
  memcpy(&x, p, sizeof(uint32));
  p += sizeof(uint32);
  msg->subtype = ntohl(x);

  tlv = msg->tlv_list = NULL;
  while(p < (tcp_socket->buf + tcp_socket->end)) {
    uint32 type, length;
    tlv_t *n_tlv;

    memcpy(&x, p, sizeof(uint32));
    p += sizeof(uint32);
    type = ntohl(x);

    memcpy(&x, p, sizeof(uint32));
    p += sizeof(uint32);
    length = ntohl(x);

    if ((tcp_socket->buf + tcp_socket->end) < (p + length)) {
      /* This is not good. Our received/verified message is not long enough
	 to contain this TLV */
      logmsg(MSG_ERROR,"TLV length (%d) exceeds size of received message for "
	     "session id %lu\n",length,session_obj->id);
      tlv = msg->tlv_list;
      while(tlv != NULL) {
	n_tlv = tlv->next;
	if (tlv->value) free(tlv->value);
	free(tlv);
	tlv = n_tlv;
      }
      free(msg);
      return 1;
    }
    
    if (tlv == NULL) {
      tlv = msg->tlv_list = malloc(sizeof(tlv_t));
    } else {
      tlv->next = malloc(sizeof(tlv_t));
      tlv = tlv->next;
    }
    if (tlv == NULL) {
      logmsg(MSG_ERROR,"Failed allocating memory in control_msg_recv() (%d bytes)\n",
	     sizeof(tlv_t));
      tlv = msg->tlv_list;
      while(tlv != NULL) {
	n_tlv = tlv->next;
	if (tlv->value) free(tlv->value);
	free(tlv);
	tlv = n_tlv;
      }
      free(msg);
      return 1;
    }
      
    tlv->type = type;
    tlv->length = length;
    if (length > 0) {
      tlv->value = malloc(length+1);
      if (tlv->value == NULL) {
	logmsg(MSG_ERROR,"Failed allocating memory in control_msg_recv() "
	       "(%d bytes)\n",length+1);
	tlv = msg->tlv_list;
	while(tlv != NULL) {
	  n_tlv = tlv->next;
	  if (tlv->value) free(tlv->value);
	  free(tlv);
	  tlv = n_tlv;
	}
	free(msg);
	return 1;
      }
      tlv->value[length] = 0;
      memcpy(tlv->value, p, length);
      p += length;
    } else {
      tlv->value = NULL;
    }
    tlv->next = NULL;
  }

  tcp_socket->start = tcp_socket->tmp = tcp_socket->end;
  *r_msg = msg;
  return 0;
}

control_msg_t *new_control_msg(uint32 protocol, uint32 type, 
			       slan_session_t *session_obj) {
  control_msg_t *msg;

  msg = slan_calloc(1, sizeof(control_msg_t));
  if (msg == NULL) {
    logmsg(MSG_ERROR,"Failed allocating memory in new_control_msg() (%d bytes)\n",
	   sizeof(control_msg_t));
    return NULL;
  }

  msg->type = protocol;
  msg->subtype = type;
  msg->tlv_list = NULL;
  
  return msg;
}

void control_msg_destruct(control_msg_t *msg) {
  tlv_t *tlv;

  if (msg == NULL) return;
  tlv = msg->tlv_list;
  while(tlv!=NULL) {
    tlv_t *p;

    p = tlv;
    tlv = tlv->next;
    if (p->value) free(p->value);
    free(p);
  }
  free(msg);
}

void msg_append_tlv(control_msg_t *msg, uint32 type, uint32 length, uchar *value) {
  tlv_t *tlv;

  if (msg == NULL) return;
  
  if (msg->tlv_list == NULL) {
    tlv = msg->tlv_list = malloc(sizeof(tlv_t));
  } else {
    tlv = msg->tlv_list;
    while(tlv->next != NULL) tlv = tlv->next;
    tlv->next = malloc(sizeof(tlv_t));
    tlv = tlv->next;
  }
  if (tlv == NULL) {
    logmsg(MSG_ERROR,"Failed to allocate memory in msg_append_tlv() (%d bytes)\n",
	   sizeof(tlv_t));
    return;
  }
  tlv->next = NULL;
  tlv->type = type;
  tlv->length = length;
  tlv->value = malloc(length);
  if (tlv->value == NULL) {
    logmsg(MSG_ERROR,"Failed to allocate memory in msg_append_tlv() (%d bytes)\n",
	   length);
    tlv->length = 0;
    return;
  }
  memcpy(tlv->value, value, length);

}

void msg_search_tlv(control_msg_t *msg, uint32 type, uint32 *length, 
		    uchar **value) {
  tlv_t *tlv;

  if (msg == NULL) return;
  tlv = msg->tlv_list;

  while(tlv!=NULL && tlv->type != type) tlv = tlv->next;

  if (tlv == NULL) {
    *value = NULL;
    *length = 0;
    return;
  }

  *length = tlv->length;
  /* Important! Note that we are not making a copy of the value here --
     the caller must not free() since control_msg_destruct() will do this! */
  *value = tlv->value;
}
