/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/control/mac_conf.c,v 1.2 1999/12/09 02:15:49 koni Exp $ 
  $Author: koni $ $Date: 1999/12/09 02:15:49 $ $Name: slan-release-0-04 $
  $Revision: 1.2 $

*/
   
#ifndef WIN32
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#else
#include "slan_windows.h"
#endif

#include "slan_types.h"
#include "slan_alloc.h"

#include "slan_session.h"

#include "control_tcp.h"
#include "log_message.h"

#define MACCONF_TIMEMAX (10000)

enum { SESSION_ID = 0, MAC_ADDRESS, N_MACCONF_ITEMS };

enum { MACCONF_REQ = 0, MACCONF_ACK, MACCONF_NACK, MACCONF_FAIL, 
       N_IPCONF_MSG };

enum { MACCONF_OK = 0, MACCONF_ETIMEOUT, MACCONF_TCP_ERROR };

static int new_mac_address(unsigned char mac_address[6]) {
  /* We are generating a new mac address by subtracting from the 
     second most significant octet. This should start as 0xFD.

     The MAC address should lead with FE:FD in order to be seperate
     from the space allocated to HW vendors. I am not positive of this yet,
     but seems to be, from what I've read, and the fact that the TUN/TAP
     driver assigns a random MAC address by default using this prefix.

     The least significant bit of the first octet signifies a multicast
     address. The second least significant bit signifies "locally assigned"
     according to various RFCs I've read (RFC 1700?), but some vendors have
     ranges assigned using this bit, so I don't know for sure what is up. */

  if (mac_address[1] > 0) {
    mac_address[1] --;
    return 1;
  } else {
    return 0;
  }
}

int macconf_request(unsigned char mac_address[6], 
		    slan_session_t *session_obj) {
  control_msg_t *msg;
  uint32 length;
  uchar *p;
  int timeout, rval;

 start_over:
  msg = new_control_msg(MACCONF, MACCONF_REQ, session_obj);
  msg_append_tlv(msg, MAC_ADDRESS, 6, mac_address);
  timeout = MACCONF_TIMEMAX;
  rval = control_msg_send(msg, session_obj, &timeout);
  control_msg_destruct(msg);
  if (rval < 0) return MACCONF_TCP_ERROR;
  if (rval > 0) return MACCONF_ETIMEOUT;

  timeout = MACCONF_TIMEMAX;
  while(1) {
    rval = control_msg_recv(&msg, session_obj, &timeout);
    if (rval < 0) return MACCONF_TCP_ERROR;
    if (rval > 0) return MACCONF_ETIMEOUT;
    if (msg->type == MACCONF && msg->subtype == MACCONF_ACK) break;
    if (msg->type == MACCONF && msg->subtype == MACCONF_FAIL) {
      control_msg_destruct(msg);
      return 1;
    }
    if (msg->type == MACCONF && msg->subtype == MACCONF_NACK) {
      control_msg_destruct(msg);
      logmsg(MSG_WARNING,"Server rejected proposed MAC address\n");
      if (new_mac_address(mac_address)) goto start_over;
      else return 1;
    }
    logmsg(MSG_DEBUG,"Received unexpected control message in "
	   "macconf_request() protocol = %0lX, type = %0lX\n",
	   msg->type, msg->subtype);
    control_msg_destruct(msg);
  }

  p = NULL;
  msg_search_tlv(msg, SESSION_ID, &length, &p);
  if (length != 4) {
    control_msg_destruct(msg);
    logmsg(MSG_CRITICAL,"Server did not send us a session id\n");
    return 1;
  }

  memcpy(&session_obj->id, p, 4);
  control_msg_destruct(msg);
  return 0;
}

u_int32_t macaddr_lookup(uchar *);
int macconf_req_wait(slan_session_t *session_obj, 
		     uchar *mac_address) {
  control_msg_t *msg;
  uint32 length;
  int rval, timeout;
  uchar *macaddr;

 start_over:
  timeout = MACCONF_TIMEMAX;
  while(1) {
    rval = control_msg_recv(&msg, session_obj, &timeout);
    if (rval < 0) return MACCONF_TCP_ERROR;
    if (rval > 0) return MACCONF_ETIMEOUT;
    if (msg->type == MACCONF && msg->subtype == MACCONF_REQ) break;
    logmsg(MSG_DEBUG,"Received unexpected control message in macconf_req_"
	   "wait() protocol = %0lX, type = %0lX\n",msg->type, msg->subtype);
    control_msg_destruct(msg);
  }

  msg_search_tlv(msg, MAC_ADDRESS, &length, &macaddr);
  if (length != 6) {
    msg = new_control_msg(MACCONF, MACCONF_FAIL, session_obj);
    timeout = MACCONF_TIMEMAX;
    rval = control_msg_send(msg, session_obj, &timeout);
    control_msg_destruct(msg);
    if (rval < 0) return MACCONF_TCP_ERROR;
    if (rval > 0) return MACCONF_ETIMEOUT;
    return 1;
  }

  /* Require two items to accept a mac address: One, the locally assigned
     bit, two, the same address is not already present in our mac address
     mapping. 
     
     More through treatment would actually probe the LAN to make sure the
     address is not in use, but this would probably be slow, and I don't
     know how to do an 802 link-layer ping at this time, and its unlikely
     that it would be necessary in pretty much all cases. */
  if ((macaddr[0] & 0x02) && !macaddr_lookup(macaddr)) {
    memcpy(mac_address, macaddr, length);
    control_msg_destruct(msg);

    msg = new_control_msg(MACCONF, MACCONF_ACK, session_obj);
    msg_append_tlv(msg, SESSION_ID, 4, (uchar *) &session_obj->id);

    timeout = MACCONF_TIMEMAX;
    rval = control_msg_send(msg, session_obj, &timeout);
    control_msg_destruct(msg);
    if (rval < 0) return MACCONF_TCP_ERROR;
    if (rval > 0) return MACCONF_ETIMEOUT;
    return 0;
  } else {
    control_msg_destruct(msg);
    msg = new_control_msg(MACCONF, MACCONF_NACK, session_obj);

    timeout = MACCONF_TIMEMAX;
    rval = control_msg_send(msg, session_obj, &timeout);
    control_msg_destruct(msg);
    if (rval < 0) return MACCONF_TCP_ERROR;
    if (rval > 0) return MACCONF_ETIMEOUT;
    goto start_over;
  }
}

