/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/control/random_interface.c,v 1.1.1.1 1999/12/09 01:16:34 koni Exp $ 
  $Author: koni $ $Date: 1999/12/09 01:16:34 $ $Name: slan-release-0-04 $
  $Revision: 1.1.1.1 $

*/
   
#ifndef WIN32
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include <sys/types.h>
#include <pthread.h>
#include <syslog.h>
#else
#include "slan_windows.h"
#undef _WIN32_WINNT
#define _WIN32_WINNT 0x4000
#include "wincrypt.h"
#endif

#include "slan_types.h"
#include "slan_alloc.h"

#ifdef HAVE_GMP_H
#include <gmp.h>
#else
#include "slan_gmp3.h"
#endif

#include <sym_cipher.h>
#include <blowfish.h>


static cipher_t *blowfish_gen;
static u_int32_t plaintext_sequence[2];

static pthread_mutex_t random_lock;

#define CIPHER_BLOCK_BITS (64)
#define CIPHER_BLOCK_BYTES (8)

#ifdef linux
/* Initial entropy source on Linux */
static void get_devrandom_entropy(unsigned char *, int);
void (*get_entropy_bytes)(unsigned char *, int) = get_devrandom_entropy;

#else
static void get_windows_entropy(unsigned char *, int);
void (*get_entropy_bytes)(unsigned char *, int) = get_windows_entropy;

#endif

#ifdef linux
static void get_devrandom_entropy(unsigned char *entropy, int bytes) {
  FILE *f;

#ifdef DEBUG
  f = fopen("/dev/urandom","r");
#else
  f = fopen("/dev/random","r");
#endif
  if (f == NULL) {
    perror("Could not open /dev/random for entropy source: ");
    syslog(LOG_CRIT,"Cannot start SLAN: /dev/random could not be opened for"
	   " an entropy source.\n");
    abort();
  }

  fread(entropy,1,bytes,f);
  fclose(f);
}
#endif

#ifdef WIN32
static void get_windows_entropy(unsigned char *entropy, int bytes) {
   BOOL res;
   HCRYPTPROV hcp;
   UINT errval;
 
    /* we need to acquire a cryptographic context to generate random numbers.
      talk about reaching around your ass to grab your elbow.. ;) */
 
   res = CryptAcquireContext(&hcp, NULL, NULL, PROV_RSA_FULL, 0);
   if(!res && GetLastError() == NTE_BAD_KEYSET)
         res = CryptAcquireContext(&hcp, NULL, NULL, PROV_RSA_FULL, CRYPT_NEWKEYSET);
   if(!res)
   {
         errval = GetLastError();
         abort(); // this should be handled more elegantly ;)
   }
 
   res = CryptGenRandom(hcp, bytes, entropy);
   if(!res)
   {
         errval = GetLastError();
         abort(); // and this should just never happen.
   }
 
   CryptReleaseContext(hcp, 0);
}
#endif


void random_initialize() {
  /* This will store 40 bytes (256 bit key + 64 bit plaintext) of "true
     randomness" gathered from the operating system (system dependant) */
  unsigned char sysrandom[40];

  /* Obtain a pointer to a blowfish object to use as a random number
     generator. This works by keying the cipher with some "entropy" which we
     get from the system by "get_entropy_bytes", a function pointer which
     points to the appropriate system dependant routine for "true randomness"
  */ 
  blowfish_gen = get_blowfish();
  
  /* Get 256 + 64 bits of entropy to use as a key and initial plaintext */
  get_entropy_bytes(sysrandom, 40);

  /* Here, we may want to call some function to get a stored random state off
     the disk, to mix in with the sysrandom generated above, if we don't have
     a reliable entropy source on a particular system. Or, that could be done
     in the system dependant function, if necessary. */
  
  blowfish_gen->set_key(blowfish_gen, sysrandom, 32);

  memcpy(plaintext_sequence,sysrandom + 32, 4);
  memcpy(plaintext_sequence + 1, sysrandom + 36, 4);

#ifndef WIN32
  pthread_mutex_init(&random_lock, NULL);
#endif
}

/* This function builds a string of <num> random bytes, stored in <bytes>. The 
   <bytes> array must be allocated by the caller and of sufficient size to
   hold <num> bytes. */
void random_byte_string(unsigned char *bytes, int num) {
  /* This temporary memory is not really necessary, as we can just use the
     random_bytes array directly -- but I like doing it this way as its more 
     explicit not as easy to screw up, and it is imperative that this stuff
     is handled correctly. */
  unsigned char ciphertext[CIPHER_BLOCK_BYTES];

#ifndef WIN32
  pthread_mutex_lock(&random_lock);
#endif
  while(num>0) {
    /* Blowfish will encrypt in place, and we want to ensure that the
       plaintext sequence (essentially the sequence number) is progressing
       independantly of the randomizing operation (block encryption) */
    memcpy(ciphertext,plaintext_sequence,CIPHER_BLOCK_BYTES);

    /* Convert the plaintext to something "random" */
    blowfish_gen->block_encrypt(blowfish_gen, ciphertext);

    if (CIPHER_BLOCK_BYTES < num) 
      memcpy(bytes,ciphertext,CIPHER_BLOCK_BYTES);
    else
      memcpy(bytes,ciphertext,num);

    /* Increment the plaintext sequence */
    plaintext_sequence[1] += 1;
    if (plaintext_sequence[1] == 0) plaintext_sequence[0] += 1;

    num -= CIPHER_BLOCK_BYTES;
    bytes += CIPHER_BLOCK_BYTES;
  }
#ifndef WIN32
  pthread_mutex_unlock(&random_lock);
#endif
}

/* Build a random number of size 2^bits, essentially unifrom between
   2^(bits) and 2^(bits+1). The number is returned as a mpz_t integer. This
   mpz_t integer must be initialized already by the caller. */
void random_uniform_2exp(mpz_t op, int bits) {
  int i, num_bytes;
  unsigned char *random_bytes;

  /* Calculate the number of bytes we need to build a <bits> size mpz
     integer. */
  num_bytes = (bits)/CHAR_BIT + 1;
  random_bytes = slan_calloc(num_bytes, sizeof(unsigned char));

  random_byte_string(random_bytes, num_bytes);
    
  /* Built and mpz integer from the random bit string. We do this not so
     efficiently by grabbing a byte at a time from the string and adding it to 
     the mpz integer, then shifting left, etc. This could probably be done
     more efficiently by building a random string with a even blocksize number 
     of bytes, and adding u_int32_t's, then truncating to proper size as done
     below. But I didn't think of that 'til now, when its already done this
     way.  */
  mpz_set_ui(op,0);
  for(i=0;i<num_bytes;i++) {
    unsigned char temp_ui;
    temp_ui = random_bytes[i];
    /* This does nothing the on the very first iteration size (of both
       loops) since op is 0, but it would have to be undone if done after
       the addition, outside the loop as it would leave the 32 right most
       bits as zeros. This is simpler for me. */
    mpz_mul_2exp(op,op,CHAR_BIT);
    mpz_add_ui(op,op,temp_ui);
  }

  if (mpz_cmp_ui(op,0) < 0) {
#ifndef WIN32
    syslog(LOG_ERR,"Warning: Random number generation resulted in a "
	    "negative number... somehow. (auto corrected)\n");
#endif
    /* Correct the problem */
    mpz_neg(op,op);
  }

  /* Shift right so that the size of the number here is at least <bits> in
     size, and then set the high order bit. Note that some long string on
     zeros in front of the number may cause an effective right shift longer
     than what we want, so we must set the high order bit to ensure a number
     of 2^bits size */
  if (mpz_sizeinbase(op,2) > bits) 
    mpz_fdiv_q_2exp(op,op,mpz_sizeinbase(op,2)-bits);
  mpz_setbit(op,bits-1);

  if (mpz_cmp_ui(op,0) < 0) {
#ifndef WIN32
    syslog(LOG_ERR,"Warning: Random number generation (after truncation) "
	   "resulted in a negative number. (auto corrected)\n");
#endif
    /* Correct that problem */
    mpz_neg(op,op);
  }

  slan_memfree(random_bytes);
}

