/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/control/rsa_auth.c,v 1.1.1.1 1999/12/09 01:16:34 koni Exp $ 
  $Author: koni $ $Date: 1999/12/09 01:16:34 $ $Name: slan-release-0-04 $
  $Revision: 1.1.1.1 $

*/
   
#ifndef WIN32
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <syslog.h>
#else
#include "slan_windows.h"
#endif

#include "slan_types.h"
#include "slan_alloc.h"
#include "slan_config.h"

#ifdef HAVE_GMP_H
#include <gmp.h>
#else
#include "slan_gmp3.h"
#endif

#include "control_tcp.h"
#include "random_interface.h"
#include "rsaauth.h"
#include "sha.h"
#include "log_message.h"

extern slan_conf_t *slan_global_config;

enum { RSAAUTH_REQUEST = 0, RSAAUTH_NONCE, 
       RSAAUTH_HASH, RSAAUTH_ACCEPT, RSAAUTH_ACCEPT_ACK, RSAAUTH_REJECT,
       RSAAUTH_REFUSE, RSAAUTH_ABORT, RSAAUTH_NUM_MSGTYPE };

enum { RSAAUTH_FINGERPRINT = 0, RSAAUTH_NONCEVALUE, RSAAUTH_HASHVALUE, 
       RSAAUTH_ERROR };

enum { RSAAUTH_ACCEPTED = 0, RSAAUTH_EREFUSE, RSAAUTH_ENOMEM, 
       RSAAUTH_EBADNONCE, RSAAUTH_EABORT, RSAAUTH_EREJECT, 
       RSAAUTH_ETIMEOUT, RSAAUTH_EBADKEY, RSAAUTH_TCP_ERROR, RSAAUTH_NUM_ERR };

static char *rsaauth_error[RSAAUTH_NUM_ERR] = 
{ "(No Error) RSA Authentication accepted",
  "Peer refused RSA authentication",
  "Memory allocation failure",
  "Invalid nonce sent by peer",
  "Protocol could not be completed (aborted)",
  "Peer failed RSA Authentication",
  "Protocol Timeout",
  "Cannot find key matching sent fingerprint",
  "TCP transmission failure"
};

typedef struct {
  slan_session_t *session_obj;
  int nonce_bits;
  rsakey_t *key;
  mpz_t nonce1_mpz;
  mpz_t nonce2_mpz;
  unsigned char *keyhash;
  int keyhash_length;
  unsigned char *error_string;
  int error_length;
} rsaauth_obj_t;

#define RSAAUTH_TIMEMAX (10000)

static rsaauth_obj_t *new_rsaauth_obj(slan_session_t *session_obj) {
  rsaauth_obj_t *obj;

  obj = slan_malloc(sizeof(rsaauth_obj_t));
  obj->session_obj = session_obj;
  obj->key = slan_global_config->rsakey;  

  /* We will require the nonce values to be the same number of bits as the
     modulus */
  obj->nonce_bits = mpz_sizeinbase(slan_global_config->rsakey->modulus_mpz,2) 
    - 1;

  mpz_init(obj->nonce1_mpz);
  mpz_init(obj->nonce2_mpz);

  /* One of the nonce values will be overwritten with the value sent by the
     peer. Which one depends on who initiates the authentication. We set
     random values for both here to make the code below simpler. */
  random_uniform_2exp(obj->nonce1_mpz,obj->nonce_bits);
  random_uniform_2exp(obj->nonce2_mpz,obj->nonce_bits);

  /* This serves as a secret to protect against man-in-the-middle attacks of
     Diffie-Hellman. We will verify, folded into the RSA auth challenge, that
     the key used by the legitimate server is indeed the same as that used by
     the client (for the symmetric encryption) -- the makes it very difficult
     for an attacker to sit in the middle of the DH key-exchange, setting
     different keys with the client and server, and then letting the server
     authenticate as usual  */
  obj->keyhash_length = session_obj->keyhash_length;
  obj->keyhash = session_obj->keyhash;

  obj->error_length = 0;
  obj->error_string = NULL;

  return obj;
}

static void rsaauth_obj_destruct(rsaauth_obj_t *obj) {
  mpz_clear(obj->nonce1_mpz);
  mpz_clear(obj->nonce2_mpz);

  if (obj->error_string != NULL) slan_memfree(obj->error_string);

  slan_memfree(obj);
}


static void rsaauth_build_mpz(mpz_t number, unsigned char *data, 
			      int length) {
  mpz_t temp;
  int i;

  mpz_init(temp);
  mpz_set_ui(number,0);
  for(i=0;i<length;i++) {
    mpz_set_ui(temp,(unsigned long)data[i]);
    mpz_mul_2exp(temp,temp,i*8);
    mpz_add(number,number,temp);
  }
  mpz_clear(temp);
}

/* Note! Switched to add instead of xor (hence nonce_sum_mpz rather than the
   nonce_xor_mpz) because GMP version 2 does not have the XOR function. We
   don't really need it anyway. In fact, I'm not sure the sum variable is
   needed at all since we are hashing everything away anyway. */
static void rsaauth_build_hash(rsaauth_obj_t *obj, mpz_t hash_mpz) {
  mpz_t nonce_sum_mpz;
  int hash_document_length, offset;
  unsigned char *hash_document;
  unsigned char hash_value[20];

  mpz_init(nonce_sum_mpz);
  mpz_set_ui(nonce_sum_mpz, 0);
  mpz_add(nonce_sum_mpz, obj->nonce1_mpz, obj->nonce2_mpz);

  hash_document_length = mpz_sizeinbase(obj->nonce1_mpz,16) +
    mpz_sizeinbase(obj->nonce2_mpz,16) + mpz_sizeinbase(nonce_sum_mpz,16) +
    obj->keyhash_length;

  hash_document = slan_malloc(hash_document_length + 10);
  
  offset = 0;
  mpz_get_str(hash_document,16,obj->nonce1_mpz);
  offset += mpz_sizeinbase(obj->nonce1_mpz, 16);

  mpz_get_str(hash_document + offset, 16, obj->nonce2_mpz);
  offset += mpz_sizeinbase(obj->nonce2_mpz, 16);
  
  mpz_get_str(hash_document + offset, 16, nonce_sum_mpz);
  offset += mpz_sizeinbase(nonce_sum_mpz, 16);

  memcpy(hash_document + offset, obj->keyhash, obj->keyhash_length);
  hash_document[hash_document_length] = 0;

  sha_digest(hash_document,hash_document_length,hash_value);
  mpz_clear(nonce_sum_mpz);
  slan_memfree(hash_document);

  rsaauth_build_mpz(hash_mpz,hash_value,20);
}
  
static int rsaauth_wait_accept(rsaauth_obj_t *obj) {
  control_msg_t *msg;
  int rval, timeout;
  
  timeout = RSAAUTH_TIMEMAX;
  while(1) {
    rval = control_msg_recv(&msg, obj->session_obj, &timeout);
    if (rval < 0) return RSAAUTH_TCP_ERROR;
    if (rval > 0) return RSAAUTH_ETIMEOUT;
    if (msg->type == RSAAUTH && msg->subtype == RSAAUTH_ACCEPT) break;
    if (msg->type == RSAAUTH && msg->subtype == RSAAUTH_REJECT) {
      control_msg_destruct(msg);
      return RSAAUTH_EREJECT;
    }
    logmsg(MSG_DEBUG,"Received unexpected control message in rsaauth_wait_accept() "
	   "protocol = %0lX, type = %0lX\n",msg->type, msg->subtype);
    control_msg_destruct(msg);
  }

  control_msg_destruct(msg);
  msg = new_control_msg(RSAAUTH, RSAAUTH_ACCEPT_ACK, obj->session_obj);
  timeout = RSAAUTH_TIMEMAX;
  rval = control_msg_send(msg, obj->session_obj, &timeout);
  control_msg_destruct(msg);
  if (rval < 0) return RSAAUTH_TCP_ERROR;
  if (rval > 0) return RSAAUTH_ETIMEOUT;
  return RSAAUTH_ACCEPTED;
}

static int rsaauth_send_hash(rsaauth_obj_t *obj) {
  mpz_t hash_mpz;
  uchar *hash;
  control_msg_t *msg;
  uint32 size;
  int rval, timeout;

  mpz_init(hash_mpz);
  rsaauth_build_hash(obj, hash_mpz);
  
  mpz_powm(hash_mpz,hash_mpz,obj->key->private_mpz,obj->key->modulus_mpz);
  hash = mpz_get_str(NULL,16,hash_mpz);
  mpz_clear(hash_mpz);

  msg = new_control_msg(RSAAUTH, RSAAUTH_HASH, obj->session_obj);
  size = strlen(hash);
  msg_append_tlv(msg, RSAAUTH_HASHVALUE, size, hash);

  free(hash);
  timeout = RSAAUTH_TIMEMAX;
  rval = control_msg_send(msg, obj->session_obj, &timeout);
  control_msg_destruct(msg);
  if (rval < 0) return RSAAUTH_TCP_ERROR;
  if (rval > 0) return RSAAUTH_ETIMEOUT;

  return rsaauth_wait_accept(obj);
}

static int rsaauth_test_nonce2(rsaauth_obj_t *obj, int nonce_length, unsigned
			       char *nonce) {
  control_msg_t *msg;
  int size, rval, timeout;

  if (nonce_length !=0 ) {
    mpz_set_str(obj->nonce2_mpz,nonce,16);
  }

  if (nonce_length==0 || mpz_sizeinbase(obj->nonce2_mpz,2)!=obj->nonce_bits) {
    logmsg(MSG_WARNING,"Expected nonce of size %d, received %d bits.\n", 
	   obj->nonce_bits, mpz_sizeinbase(obj->nonce2_mpz,2)); 
    msg = new_control_msg(RSAAUTH, RSAAUTH_ABORT, obj->session_obj);
    size = strlen(rsaauth_error[RSAAUTH_EBADNONCE]);
    msg_append_tlv(msg,RSAAUTH_ERROR,size,rsaauth_error[RSAAUTH_EBADNONCE]);
    rval = control_msg_send(msg, obj->session_obj, &timeout);
    control_msg_destruct(msg);
    if (rval < 0) return RSAAUTH_TCP_ERROR;
    if (rval > 0) return RSAAUTH_ETIMEOUT;
    return RSAAUTH_EBADNONCE;
  }

  return rsaauth_send_hash(obj);
}

static int rsaauth_send_nonce1(rsaauth_obj_t *obj) {
  control_msg_t *msg;
  uchar *nonce;
  uint32 nonce_length;
  int rval, timeout;

  msg = new_control_msg(RSAAUTH, RSAAUTH_NONCE, obj->session_obj);
  nonce = mpz_get_str(NULL, 16, obj->nonce1_mpz);
  nonce_length = strlen(nonce);
  msg_append_tlv(msg, RSAAUTH_NONCEVALUE, nonce_length, nonce);
  slan_memfree(nonce);
  
  timeout = RSAAUTH_TIMEMAX;
  rval = control_msg_send(msg, obj->session_obj, &timeout);
  control_msg_destruct(msg);
  if (rval < 0) return RSAAUTH_TCP_ERROR;
  if (rval > 0) return RSAAUTH_ETIMEOUT;

  timeout = RSAAUTH_TIMEMAX;
  while(1) {
    rval = control_msg_recv(&msg, obj->session_obj, &timeout);
    if (rval < 0) return RSAAUTH_TCP_ERROR;
    if (rval > 0) return RSAAUTH_ETIMEOUT;
    if (msg->type == RSAAUTH && msg->subtype == RSAAUTH_NONCE) break;
    if (msg->type == RSAAUTH && msg->subtype == RSAAUTH_ABORT) {
      uint32 elength;
      uchar *err_string;

      msg_search_tlv(msg, RSAAUTH_ERROR, &elength, &err_string);
      if (elength) {
	obj->error_string = strdup(err_string);
      }
      control_msg_destruct(msg);
      return RSAAUTH_EABORT;
    }
    logmsg(MSG_DEBUG,"Received unexpected control message in "
	   "rsaauth_send_nonce1() protocol = %0lX, type = %0lX\n",
	   msg->type, msg->subtype);
    control_msg_destruct(msg);
  }

  msg_search_tlv(msg, RSAAUTH_NONCEVALUE, &nonce_length, &nonce);
  
  rval = rsaauth_test_nonce2(obj, nonce_length, nonce);
  control_msg_destruct(msg);
  return rval;
}  

/* For now, we are just going to refuse RSA authentication when the
   key is unknown. Later, we should send a different message
   indicating that we can't continue because the key is unknown. */
static int send_unknown_key(slan_session_t *session_obj) {
  control_msg_t *msg;
  int timeout;
 
  msg = new_control_msg(RSAAUTH, RSAAUTH_REFUSE, session_obj);
  timeout = RSAAUTH_TIMEMAX;
  control_msg_send(msg, session_obj, &timeout);
  control_msg_destruct(msg);
  return RSAAUTH_EBADKEY;
}


int rsaauth_listen(slan_session_t *session_obj) {
  rsaauth_obj_t *obj;
  uint32 fingerprint_length;
  uchar *fingerprint;
  control_msg_t *msg;
  int timeout, rval;

  timeout = RSAAUTH_TIMEMAX;
  while(1) {
    rval = control_msg_recv(&msg, session_obj, &timeout);
    if (rval < 0) return RSAAUTH_TCP_ERROR;
    if (rval > 0) return RSAAUTH_ETIMEOUT;
    if (msg->type == RSAAUTH && msg->subtype == RSAAUTH_REQUEST) break;
    logmsg(MSG_DEBUG,"Received unexpected control message in rsaauth_listen()"
	   " protocol = %0lX, type = %0lX\n",msg->type, msg->subtype);
    control_msg_destruct(msg);
  }

  msg_search_tlv(msg, RSAAUTH_FINGERPRINT, &fingerprint_length, &fingerprint);
  if (fingerprint_length == 0) {
    control_msg_destruct(msg);
    return send_unknown_key(session_obj);
  } 
  if (memcmp(slan_global_config->rsakey->fingerprint, fingerprint, 20)) {
    control_msg_destruct(msg);
    return send_unknown_key(session_obj);
  }

  control_msg_destruct(msg);
  obj = new_rsaauth_obj(session_obj);
  rval = rsaauth_send_nonce1(obj);
  rsaauth_obj_destruct(obj);
  return rval;
}

static int rsaauth_accept(rsaauth_obj_t *obj) {
  control_msg_t *msg;
  int timeout, rval;

  timeout = RSAAUTH_TIMEMAX;
  while(1) {
    rval = control_msg_recv(&msg, obj->session_obj, &timeout);
    if (rval < 0) return RSAAUTH_TCP_ERROR;
    if (rval > 0) return RSAAUTH_ETIMEOUT;
    if (msg->type == RSAAUTH && msg->subtype == RSAAUTH_ACCEPT_ACK) break;
    logmsg(MSG_DEBUG,"Received unexpected control message in rsaauth_accept() "
	   "protocol = %0lX, type = %0lX\n",msg->type, msg->subtype);
    control_msg_destruct(msg);
  }
  control_msg_destruct(msg);

  return RSAAUTH_ACCEPTED;
}

/* This function is called to check the hash value sent by the peer
   (server). The hash is verified by decrypting with the server's
   public key, and comparing to our own computation of the hash. If
   they match, it proves that the server knows the private key. */
static int rsaauth_verify_hash(rsaauth_obj_t *obj, int hash_length, unsigned
				char *hash) {
  mpz_t our_hash_mpz, peer_hash_mpz;
  control_msg_t *msg;
  int timeout, rval;

  if (hash_length == 0) {
    msg = new_control_msg(RSAAUTH, RSAAUTH_REJECT, obj->session_obj);
    timeout = RSAAUTH_TIMEMAX;
    rval = control_msg_send(msg, obj->session_obj, &timeout);
    control_msg_destruct(msg);
    if (rval < 0) return RSAAUTH_TCP_ERROR;
    if (rval > 0) return RSAAUTH_ETIMEOUT;

    return RSAAUTH_EREJECT;
  }

  mpz_init(peer_hash_mpz);
  mpz_set_str(peer_hash_mpz,hash,16);
  mpz_powm(peer_hash_mpz,peer_hash_mpz,obj->key->public_mpz, obj->key->modulus_mpz);
  mpz_init(our_hash_mpz);
  rsaauth_build_hash(obj, our_hash_mpz);

  if (mpz_cmp(our_hash_mpz, peer_hash_mpz)) {
    mpz_clear(our_hash_mpz);
    mpz_clear(peer_hash_mpz);
    logmsg(MSG_ERROR,"RSA Authentication error. Server identity could not be"
	   " proven.\n");
    msg = new_control_msg(RSAAUTH, RSAAUTH_REJECT, obj->session_obj);
    timeout = RSAAUTH_TIMEMAX;
    rval = control_msg_send(msg, obj->session_obj, &timeout);
    control_msg_destruct(msg);
    return RSAAUTH_EREJECT;
  }

  mpz_clear(our_hash_mpz);
  mpz_clear(peer_hash_mpz);
  msg = new_control_msg(RSAAUTH, RSAAUTH_ACCEPT, obj->session_obj);
  timeout = RSAAUTH_TIMEMAX;
  rval = control_msg_send(msg, obj->session_obj, &timeout);
  control_msg_destruct(msg);
  if (rval < 0) return RSAAUTH_TCP_ERROR;
  if (rval > 0) return RSAAUTH_ETIMEOUT;
  
  return rsaauth_accept(obj);
}  

static int rsaauth_send_nonce2(rsaauth_obj_t *obj) {
  uchar *nonce, *hash;
  uint32 hash_length, size;
  control_msg_t *msg;
  int timeout, rval;

  msg = new_control_msg(RSAAUTH, RSAAUTH_NONCE, obj->session_obj);
  nonce = mpz_get_str(NULL, 16, obj->nonce2_mpz);
  size = strlen(nonce);
  msg_append_tlv(msg, RSAAUTH_NONCEVALUE, size, nonce);
  slan_memfree(nonce);

  timeout = RSAAUTH_TIMEMAX;
  rval = control_msg_send(msg, obj->session_obj, &timeout);
  control_msg_destruct(msg);
  if (rval < 0) return RSAAUTH_TCP_ERROR;
  if (rval > 0) return RSAAUTH_ETIMEOUT;
  
  timeout = RSAAUTH_TIMEMAX;
  while(1) {
    rval = control_msg_recv(&msg, obj->session_obj, &timeout);
    if (rval < 0) return RSAAUTH_TCP_ERROR;
    if (rval > 0) return RSAAUTH_ETIMEOUT;
    if (msg->type == RSAAUTH && msg->subtype == RSAAUTH_HASH) break;
    if (msg->type == RSAAUTH && msg->subtype == RSAAUTH_ABORT) {
      uint32 elength;
      uchar *err_string;
      msg_search_tlv(msg, RSAAUTH_ERROR, &elength, &err_string);
      if (elength) {
	obj->error_string = strdup(err_string);
      }
      control_msg_destruct(msg);
      return RSAAUTH_EABORT;
    }
    logmsg(MSG_DEBUG,"Received unexpected control message in rsaauth_send_nonce2() "
	   "protocol = %0lX, type = %0lX\n",msg->type, msg->subtype);
    control_msg_destruct(msg);
  }

  msg_search_tlv(msg, RSAAUTH_HASHVALUE, &hash_length, &hash);
  rval = rsaauth_verify_hash(obj, hash_length, hash);
  control_msg_destruct(msg);
  return rval;
}

static int rsaauth_test_nonce1(rsaauth_obj_t *obj, int nonce_length, uchar *nonce) {
  control_msg_t *msg;
  uint32 size;
  int timeout, rval;

  if (nonce_length !=0 ) {
    mpz_set_str(obj->nonce1_mpz,nonce,16);
  }
  if (nonce_length==0 || mpz_sizeinbase(obj->nonce1_mpz,2)!=obj->nonce_bits) {
    logmsg(MSG_WARNING,"Expected nonce of size %d, received %d bits.\n", 
	   obj->nonce_bits, mpz_sizeinbase(obj->nonce1_mpz,2)); 
    msg = new_control_msg(RSAAUTH, RSAAUTH_ABORT, obj->session_obj);
    size = strlen(rsaauth_error[RSAAUTH_EBADNONCE]);
    msg_append_tlv(msg, RSAAUTH_ERROR, size, rsaauth_error[RSAAUTH_EBADNONCE]);

    timeout = RSAAUTH_TIMEMAX;
    rval = control_msg_send(msg, obj->session_obj, &timeout);
    control_msg_destruct(msg);
    if (rval < 0) return RSAAUTH_TCP_ERROR;
    if (rval > 0) return RSAAUTH_ETIMEOUT;
    return RSAAUTH_EBADNONCE;
  }

  return rsaauth_send_nonce2(obj);
}

/* This function is called when we are waiting for a response from the 
   peer, after sending a request. The only two possibilities are a
   REFUSE or a NONCE. NONCE indicates that the server posses the key
   indicated by our request and is willing to continue. REFUSE
   inidicates either the server does not possess the key inidicated in 
   our request or does not wish to continue. */
static int rsaauth_request_sent(rsaauth_obj_t *obj) {
  control_msg_t *msg;
  uchar *nonce;
  uint32 nonce_length;
  int timeout, rval;

  timeout = RSAAUTH_TIMEMAX;
  while(1) {
    rval = control_msg_recv(&msg, obj->session_obj, &timeout);
    if (rval < 0) return RSAAUTH_TCP_ERROR;
    if (rval > 0) return RSAAUTH_ETIMEOUT;
    if (msg->type == RSAAUTH && msg->subtype == RSAAUTH_NONCE) break;
    if (msg->type == RSAAUTH && msg->subtype == RSAAUTH_REFUSE) {
      control_msg_destruct(msg);
      return RSAAUTH_EREFUSE;
    }
    logmsg(MSG_DEBUG,"Received unexpected control message in rsaauth_request_sent() "
	   "protocol = %0lX, type = %0lX\n",msg->type, msg->subtype);
    control_msg_destruct(msg);
  }

  logmsg(MSG_INFO,"RSA_AUTH: Server has agreed to authenticate using "
	 "key labeled:\n%s\n",slan_global_config->rsakey->label);
  msg_search_tlv(msg, RSAAUTH_NONCEVALUE, &nonce_length, &nonce);
  rval = rsaauth_test_nonce1(obj, nonce_length, nonce);
  control_msg_destruct(msg);
  return rval;
}


/* This function is called to initiate an RSA authentication with the
   peer, usually (in fact, always at time of writing) called by the
   client to authenticate the server. It is called with the an
   rsa_keydata_t to specify the key to use for authentication, as well as 
   some data (keyhash) to be used to check for man in the middle
   attack. The intent for this is that a portion of the SHA hash of the
   key previously set by a diffie-hellman key exchange is added to the 
   document signed by the server's private key, making the signature
   dependant on the Diffie-Hellman key calculated. This defeats the
   man-in-the-middle attack because man-in-the-middle attack on the
   Diffie-Hellman key exchange will result in different keys computed
   by the server and client, requiring the man-in-the-middle to stay
   in the middle, decrypting then re-encrypting the traffic. However,
   the private key is unknown to the attacker, so it is very difficult 
   to make the RSA authentication of the server complete correctly,
   because the attacker cannot sign documents, and can't get the
   server to sign a document with the correct keyhash. */
int rsaauth_initiate(slan_session_t *session_obj) {
  rsaauth_obj_t *obj;
  control_msg_t *msg;
  int rval, timeout;

  obj = new_rsaauth_obj(session_obj);
  msg = new_control_msg(RSAAUTH, RSAAUTH_REQUEST, session_obj);
  
  msg_append_tlv(msg, RSAAUTH_FINGERPRINT, 20, 
		 slan_global_config->rsakey->fingerprint);

  timeout = RSAAUTH_TIMEMAX;
  rval = control_msg_send(msg, obj->session_obj, &timeout);
  control_msg_destruct(msg);
  if (rval < 0) return RSAAUTH_TCP_ERROR;
  if (rval > 0) return RSAAUTH_ETIMEOUT;
  
  rval = rsaauth_request_sent(obj);
  rsaauth_obj_destruct(obj);

  if (rval != RSAAUTH_ACCEPTED) {
    logmsg(MSG_CRITICAL,"Cannot start SLAN: RSA Authentication failed.\n");
    if (obj->error_string!=NULL) 
      logmsg(MSG_ERROR,"Peer has sent error string: %s\n",obj->error_string);   
    logmsg(MSG_ERROR,"Local error code string: %s\n",rsaauth_error[rval]);
  }

  return rval;
}

