/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/control/server_pamauth.c,v 1.1.1.1 1999/12/09 01:16:35 koni Exp $ 
  $Author: koni $ $Date: 1999/12/09 01:16:35 $ $Name: slan-release-0-04 $
  $Revision: 1.1.1.1 $

*/
   
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <security/pam_appl.h>

#include "slan_alloc.h"

#include "slan_session.h"
#include "control_tcp.h"
#include "client_pamauth.h"
#include "log_message.h"

#define CLIENTAUTH_TIMEMAX (120000)

/* The documentation on how this function is supposed to work with PAM is
   quite unclear -- I certainly hope PAM is going to free the memory
   allocated. 

   PAM calls this when it wants to ask the client questions. We gobble up
   these questions and send them to the client. The client will answer those
   messages which were prompts and do whatever with the others. */
static int slan_pamconv(int num_msg, const struct pam_message **pam_msg,
			 struct pam_response **resp, void *sesobj) {
  slan_session_t *session_obj;
  struct pam_response *pam_resp;
  int i, timeout, rval, j;
  control_msg_t *msg;
  uint32 length;
  uchar *response;
  uint32 type;

  /* We tell the pam library to pass us the session obj when this conversation 
     function is called. Thus, we avoid need for global state. */
  session_obj = (slan_session_t *) sesobj;

  msg = new_control_msg(CLIENTAUTH, PAM_SENDCLIENT, session_obj);

  type = htonl(num_msg);
  msg_append_tlv(msg, NUM_MSG, sizeof(uint32), (uchar *) &type);

  for(i=0;i<num_msg;i++) {
    switch(pam_msg[i]->msg_style) {
    case PAM_PROMPT_ECHO_OFF:
      type = htonl(PROMPT_NOECHO);
      msg_append_tlv(msg, PAM_MSG + i, sizeof(uint32), (uchar *)
		     &type);
      msg_append_tlv(msg, MSG_DATA + i, strlen(pam_msg[i]->msg), 
		     (uchar *) pam_msg[i]->msg);
      break;
    case PAM_PROMPT_ECHO_ON:
      type = htonl(PROMPT_ECHO);
      msg_append_tlv(msg, PAM_MSG + i, sizeof(uint32), (uchar *)
		     &type);
      msg_append_tlv(msg, MSG_DATA + i, strlen(pam_msg[i]->msg), 
		     (uchar *) pam_msg[i]->msg);
      break;
    case PAM_ERROR_MSG:
      type = htonl(WARNING);
      msg_append_tlv(msg, PAM_MSG + i, sizeof(uint32), (uchar *)
		     &type);
      msg_append_tlv(msg, MSG_DATA + i, strlen(pam_msg[i]->msg), 
		     (uchar *) pam_msg[i]->msg);
      break;
    case PAM_TEXT_INFO:
      type = htonl(INFORMATION);
      msg_append_tlv(msg, PAM_MSG + i, sizeof(uint32), (uchar *)
		     &type);
      msg_append_tlv(msg, MSG_DATA + i, strlen(pam_msg[i]->msg), 
		     (uchar *) pam_msg[i]->msg);
      break;
    default:
      logmsg(MSG_ERROR,"PAM Authentication error: Unknown pam message %d\n", 
	     pam_msg[i]->msg_style);
      type = htonl(SLANPAM_UNKNOWN);
      msg_append_tlv(msg, PAM_MSG + i, sizeof(uint32), (uchar *)
		     &type);
      msg_append_tlv(msg, MSG_DATA + i, strlen(pam_msg[i]->msg), 
		     (uchar *) pam_msg[i]->msg);
      break;
    }
  }
  timeout = CLIENTAUTH_TIMEMAX;
  rval = control_msg_send(msg, session_obj, &timeout);
  control_msg_destruct(msg);
  if (rval) return PAM_CONV_ERR;

  timeout = CLIENTAUTH_TIMEMAX;
  while(1) {
    rval = control_msg_recv(&msg, session_obj, &timeout);
    if (rval) return PAM_CONV_ERR;
    if (msg->type == CLIENTAUTH && msg->subtype == PAM_CLIENTRESP) break;
    logmsg(MSG_DEBUG,"Received unexpected control message in client_pamauth() "
	   "protocol = %0lX, type = %0lX\n",msg->type, msg->subtype);
    control_msg_destruct(msg);
  }

  pam_resp = slan_calloc(num_msg, sizeof(struct pam_response));
  if (pam_resp == NULL) {
    logmsg(MSG_ERROR,"Failed allocating memory in slan_pamconv() (%d bytes)\n",
	   sizeof(struct pam_response)*num_msg);
    return PAM_CONV_ERR;
  }

  for(i=0;i<num_msg;i++) {
    if (pam_msg[i]->msg_style == PAM_PROMPT_ECHO_OFF || 
	pam_msg[i]->msg_style == PAM_PROMPT_ECHO_ON) {
      response = NULL;
      msg_search_tlv(msg, PAM_RESPONSE + i, &length, &response);
      if (length > 0) {
	pam_resp[i].resp = calloc(1, length+1);
	if (pam_resp[i].resp == NULL) {
	  logmsg(MSG_ERROR,"Failed allocating memory in slan_pamconv() (%d bytes)\n",
		 length+1);
	  for(j=0;j<i;j++) {
	    if (pam_resp[j].resp) free(pam_resp[j].resp);
	  }
	  free(pam_resp);
	  return PAM_CONV_ERR;
	}
	memcpy(pam_resp[i].resp, response, length);	    
	pam_resp[i].resp_retcode = 0;

	/* Hack the PAM conversation here and attempt to grab the 
	   username, assuming that any PAM_PROMPT_ECHO_ON response
	   from the user is the login/username */
	if (pam_msg[i]->msg_style == PAM_PROMPT_ECHO_ON) {
	  session_obj->username_length = length;
	  session_obj->username = malloc(length + 1);
	  if (session_obj->username == NULL) {
	    logmsg(MSG_ERROR,"Failed allocating memory in slan_pamconv() "
		   "(%d bytes)\n", length+1);
	  } else {
	    memcpy(session_obj->username, response, length);
	    session_obj->username[length] = 0;
	  }
	} 
      } else {
	/* TLV not found */
	pam_resp[i].resp = NULL;
	pam_resp[i].resp_retcode = 0;
      }
    } else {
      /* Not a prompt message, we don't expect a response */
      pam_resp[i].resp = NULL;
      pam_resp[i].resp_retcode = 0;
    }
  }
  control_msg_destruct(msg);
  *resp = pam_resp;
  return PAM_SUCCESS;
}

static int server_pam_authenticate(slan_session_t *session_obj) {
  pam_handle_t *pamh;
  struct pam_conv slan_conv;
  control_msg_t *msg;
  int rval, pam_rval, timeout;

  slan_conv.conv = slan_pamconv;
  slan_conv.appdata_ptr = session_obj; 

  pam_rval = pam_start("slan", NULL, &slan_conv, &pamh);

  if (pam_rval != PAM_SUCCESS) {
    logmsg(MSG_ERROR,"PAM Authentication Failure: pam_start() (%s)\n",
	    pam_strerror(pamh, pam_rval));
    msg = new_control_msg(CLIENTAUTH, AUTH_REJECT, session_obj);

    timeout = CLIENTAUTH_TIMEMAX;
    rval = control_msg_send(msg, session_obj, &timeout);
    control_msg_destruct(msg);
    pam_end(pamh, pam_rval);
    if (rval < 0) return CLIENTAUTH_TCP_ERROR;
    if (rval > 0) return CLIENTAUTH_ETIMEOUT;
    return 1;
  }

  pam_rval = pam_authenticate(pamh, PAM_SILENT);
  if (pam_rval != PAM_SUCCESS) {
    syslog(MSG_ERROR,"PAM Authentication Failure: pam_authenticate(): %s\n",
	   pam_strerror(pamh, pam_rval));
    msg = new_control_msg(CLIENTAUTH, AUTH_REJECT, session_obj);
    timeout = CLIENTAUTH_TIMEMAX;
    rval = control_msg_send(msg, session_obj, &timeout);
    control_msg_destruct(msg);
    pam_end(pamh, pam_rval);
    if (rval < 0) return CLIENTAUTH_TCP_ERROR;
    if (rval > 0) return CLIENTAUTH_ETIMEOUT;
    return 1;
  }

  pam_end(pamh, PAM_SUCCESS);
  /* Later, we may want to do something a bit more fancy, like check the
     user's credentials and/or open a session for the user and use that for
     accounting, etc. */ 
  msg = new_control_msg(CLIENTAUTH, AUTH_OKAY, session_obj);

  timeout = CLIENTAUTH_TIMEMAX;
  rval = control_msg_send(msg, session_obj, &timeout);
  control_msg_destruct(msg);
  if (rval < 0) return CLIENTAUTH_TCP_ERROR;
  if (rval > 0) return CLIENTAUTH_ETIMEOUT;

  return 0;
}

/* This function is called when server wishes to authenticate the client. It
   tells the client that authentication is going to occur -- if the client
   responds that it is ready, it proceeds with PAM authentication. */
int server_clientauth_initiate(slan_session_t *session_obj) {
  control_msg_t *msg;
  int timeout, rval;

  /* Tell client we are going to authenticate it */
  msg = new_control_msg(CLIENTAUTH, AUTH_REQUEST, session_obj);
  timeout = CLIENTAUTH_TIMEMAX;
  rval = control_msg_send(msg, session_obj, &timeout);
  control_msg_destruct(msg);
  if (rval < 0) return CLIENTAUTH_TCP_ERROR;
  if (rval > 0) return CLIENTAUTH_ETIMEOUT;
  
  timeout = CLIENTAUTH_TIMEMAX;
  while(1) {
    rval = control_msg_recv(&msg, session_obj, &timeout);
    if (rval < 0) return CLIENTAUTH_TCP_ERROR;
    if (rval > 0) return CLIENTAUTH_ETIMEOUT;
    if (msg->type == CLIENTAUTH && msg->subtype == AUTH_READY) break;
    logmsg(MSG_DEBUG,"Received unexpected control message in client_pamauth() "
	   "protocol = %0lX, type = %0lX\n",msg->type, msg->subtype);
    control_msg_destruct(msg);
  }
  control_msg_destruct(msg);
  return server_pam_authenticate(session_obj);
}
    

