/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/hash_algorithms/sha.c,v 1.1.1.1 1999/12/09 01:16:35 koni Exp $ 
  $Author: koni $ $Date: 1999/12/09 01:16:35 $ $Name: slan-release-0-04 $
  $Revision: 1.1.1.1 $

*/
   
/* KONI August 21, 2000

   Secure Hash Algorithm (SHA-1) Implementation for Lightlink SLAN

   Information source: Schneier, Bruce   "Applied Cryptography" Second 
   Edition

   Revision HISTORY:

   File Created: August 21, 2000
   Functionality Tested but output not verified to be SHA: August 21, 2000
   
   Modified on Novemeber 5, 2000. Endianess Errors corrected, added
              64-bit message length token at the end of the message
	      to match specifications.
   November 5, 2000: Passed test vectors in SHA-1 publication FIPS 180-1
           http://www.itl.nist.gov/fipspubs/fip180-1.htm
  

   TODO:

   Big-Endian implementation not verified!
*/
#ifndef WIN32
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#else
#include "slan_windows.h"
#endif

#include "slan_types.h"

/* Bitwise rotate left */
#define RL(word,amount)  (((word) << (amount)) | ((word) >> (32 - (amount))))

INLINE_FUNC void byte_reverse(u_int32_t *word) {
#if __BYTE_ORDER == __LITTLE_ENDIAN
  u_int32_t copy;
  copy = (*word >> 16) | (*word << 16);
  *word = (((copy & 0xFF00FF00L) >> 8) | ((copy & 0x00FF00FFL) << 8));
#endif
}

/* This function creates the expanded hashing block that the main SHA
   algorithm uses to create its digest. It copies the original 512 bit 
   block into <process_block>, pading message if necessary, as
   specified by algorithm, then expands to fill up 80 32-bit
   words. */
static void sha_block(unsigned char *message, unsigned length, unsigned
		      i, u_int32_t process_block[]) {
  int t;

  if (i >= length) {
    /* This case happens when we have a final short block that is
       longer than 56 bytes, preventing us from putting a 64-bit
       representation of the length of the message in that
       block. Hence, we have to process an empty block to get the
       length in. Here we create that empty block with the length. */
    memset(process_block,0x0,64);
    /* If a message is exactly a multiple of 512 bits in length, we 
       have to add the padding "1" bit in the final block that
       contains the length */
    if (i == length) memset(process_block,0x80,1);
    byte_reverse(process_block);
    process_block[15] = length * 8;
  } else if (length - i < 64) {
    /* A final, short block, of the message. We pad the block with a
       single 1 bit and then zeros to 512 bits. If we have room,
       without overwriting the message or the "1" bit, we place a
       64-bit representation of the length of the message in bits. If
       we do not have room for this, we skip this stage and use
       another 512 bit block to get the length. */
    memcpy(process_block, message + i, length-i);
    ((unsigned char *) process_block)[length-i] = 0x80;
    memset(((unsigned char *) process_block) + (length - i + 1), 0x0,
	   64 - (length - i + 1));
    for(t=0;t<((length-i)/4)+1;t++)
      byte_reverse(process_block + t);
    if (length - i < 56) {
      process_block[15] = (length * 8);
    }
  } else {  
    memcpy(process_block, message + i, 64);
    for(t=0;t<16;t++)
      byte_reverse(process_block + t);
  }

  
  for(t=16;t<80;t++) {
    process_block[t] = process_block[t-3] ^ process_block[t-8] ^
      process_block[t-14] ^ process_block[t-16];
    process_block[t] = RL(process_block[t],1);
  }
}
     
/* This function computes the SHA message digest of the message
   pointed to by <message>, of <length> bytes. The digest is 160-bit
   value returned as 20 byte array in <digest>, which must be allocated
   by the caller */
void sha_digest(unsigned char *message, unsigned length, unsigned char 
		*digest) {
  u_int32_t A,B,C,D,E;
  u_int32_t a,b,c,d,e;
  int i, t, stop;

  /* Initialize digest value */
  A = 0x67452301L;
  B = 0xefcdab89L;
  C = 0x98badcfeL;
  D = 0x10325476L;
  E = 0xc3d2e1f0L;

  /* This should cause an extra iteration of the loop below, if we
     need to use an empty block to meet the requirements for padding
     the message */
  if ((length % 64) < 56 && (length % 64) != 0) stop = length;
  else stop = length + 56;

  for(i=0;i<stop;i+=64) {
    u_int32_t temp;
    u_int32_t process_block[80];
    sha_block(message, length, i, process_block);

    a = A; b = B; c = C; d = D; e = E;
    for(t=0;t<20;t++) {
      temp = RL(a,5) + ((b & c) | (~b & d)) + e + process_block[t] + 
	0x5a827999;
      e = d;
      d = c;
      c = RL(b,30);
      b = a;
      a = temp;
    }
    for(t=0;t<20;t++) {
      temp = RL(a,5) + (b ^ c ^ d) + e + process_block[t+20] + 
	0x6ed9eba1;
      e = d;
      d = c;
      c = RL(b,30);
      b = a;
      a = temp;
    }
    for(t=0;t<20;t++) {
      temp = RL(a,5) + ((b & c) | (b & d) | (c & d)) + e + 
	process_block[t+40] + 0x8f1bbcdc;
      e = d;
      d = c;
      c = RL(b,30);
      b = a;
      a = temp;
    }
    for(t=0;t<20;t++) {
      temp = RL(a,5) + (b ^ c ^ d) + e + 
	process_block[t+60] + 0xca62c1d6;
      e = d;
      d = c;
      c = RL(b,30);
      b = a;
      a = temp;
    }

    A += a;
    B += b;
    C += c;
    D += d;
    E += e;
  }
    
  ((u_int32_t *) digest)[0] = A;
  ((u_int32_t *) digest)[1] = B;
  ((u_int32_t *) digest)[2] = C;
  ((u_int32_t *) digest)[3] = D;
  ((u_int32_t *) digest)[4] = E;

  byte_reverse(((u_int32_t *) digest));
  byte_reverse(((u_int32_t *) digest) + 1);
  byte_reverse(((u_int32_t *) digest) + 2);
  byte_reverse(((u_int32_t *) digest) + 3);
  byte_reverse(((u_int32_t *) digest) + 4);
}

