/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/platform/linux/slan_client.c,v 1.2 1999/12/09 02:15:49 koni Exp $ 
  $Author: koni $ $Date: 1999/12/09 02:15:49 $ $Name: slan-release-0-04 $
  $Revision: 1.2 $

*/
   
/* The Lightlink Secure LAN project

   KONI - December 2000
   slan_client.c

   Linux client program
*/

/* Standard C library includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>
#include <errno.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* to secure memory (lock from paging to disk) */
#include <sys/mman.h>

/* Network stuff */
#include <sys/socket.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <net/route.h>
#include <net/if_arp.h>

/* GNU multiprecision math library */
#ifdef HAVE_GMP_H
#include <gmp.h>
#else
#include "slan_gmp3.h"
#endif

/* session state type definitions and constants */
#include "slan_session.h"
#include "slan_version.h"
#include "slan_config.h"

/* memory wrapper */
#include "slan_alloc.h"
#include "log_message.h"

/* Cryptography */
#include "blowfish.h"
#include "sym_cipher.h"
#include "sha.h"
#include "mac.h"
#include "hmac_sha1.h"

/* Setup and Control Protocol components */
#include "dh_keyexchange.h"
#include "rsaauth.h"
#include "random_interface.h"
#include "control_tcp.h"
#include "client_pamauth.h"
#include "mac_conf.h"

/* Transport protocol stack */
#include "proto_stack.h"
#include "stack_register.h"

/* This variable is global to the entire slan program */
slan_conf_t *slan_global_config;

static char *preset_username = NULL;
static char *preset_password = NULL;

/* These file-scope variables are used to indicate shutdown status (requested
   by signal from operating system) and whether or not to configure the tap0
   interface and fork, or just connect and drop (development only) */
static volatile int shutdown_now = 0;
static int connect_only = 0;

/* This is set if the client terminates because the server hung up or an
   exceptional circumstance occurs. This prevents the shutdown process from
   restoring default routing for the real network. Otherwise, the user may
   continue using the network under false assumption that the security
   measures in SLAN are still in effect. */
static int dont_restore = 0;

/* Main state information for the SLAN session is stored in this struct */
static slan_session_t *session_obj;

#define PID_FILENAME "/var/tmp/slan.pid"
#define PIPE_FILENAME "/var/tmp/.slan_statuspipe"

#define NOBODY_UID (99)
#define NOBODY_GID (99)

static void client_setup_transport_layer(slan_session_t *session_obj,
					 char *server_address) {
  proto_stack_t *last, *current;
  int header_save;

  header_save = 0;
  /* A stack terminator at the top -- this is simply a dummy protocol
     layer that drops any packet passed to it. It can be used to catch
     bad or dropped packets and debug them. Otherwise, it isn't necessary and
     doesn't do much. */
  current = last = slan_malloc(sizeof(proto_stack_t));
  terminate_register(current);
  header_save += current->header_reserve;
  
  /* Packets received from the slan device sent down the stack from the top */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  slan_link_register(current, 0, session_obj);
  header_save += current->header_reserve;
  session_obj->transport_top = current;
  current->prev = last;
  last = current;

  /* Optional fragmenting protocol layer -- currently not implemented */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  fragment_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* All packets being sent through the slan tunnel get tagged with sequence
     numbers before adding a MAC and encrypting. This used to limit the
     possibilities for message replay attacks. */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  sequence_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* After adding the sequence number, we add a Message Authentication Code
     (MAC) to allow the receiver to verify the integrity of the packet and the 
     sender of the data */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  message_sign_register(current, session_obj->mac);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* Lastly, we enrypt the packet for privacy during transmission. Currently,
     I am expecting the symmetric cipher used to be a configurable option, so
     the cipher used depends on the field in the session_obj */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  symmetric_cipher_register(current, session_obj->cipher);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* Session ID numbers are prepended to identify the session the message
     belongs to */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  session_id_register(current);
  header_save += current->header_reserve;
  *((u_int32_t *) current->data) = session_obj->id;
  current->prev = last;
  last = current;

  /* Transmit the secured packet via the UDP channel */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  client_udplink_register(current, server_address, 
			  slan_global_config->transport_port, session_obj);
  header_save += current->header_reserve;
  session_obj->transport_bottom = current;
  current->prev = last;
  last = current;

  /* Terminate the bottom of the stack */
  current = current->next = slan_malloc(sizeof(proto_stack_t));
  terminate_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  session_obj->header_save = header_save;
}

static void forget_keys(slan_session_t *session_obj) {
  int i;

  /* Forget keys. NULL pointers may exist if shutdown happens before the 
     session is configured fully */
  if (session_obj->mackey!=NULL) {
    for(i=0;i<session_obj->mackey_length;i++)
      session_obj->mackey[i] = 0x0;
    free(session_obj->mackey);
  }
  session_obj->mackey = NULL;

  if (session_obj->key!=NULL) {
    for(i=0;i<session_obj->key_length;i++)
      session_obj->key[i] = 0x0;
    slan_memfree(session_obj->key);
  }
  session_obj->key = NULL;

  if (session_obj->keyhash!=NULL) {
    for(i=0;i<session_obj->keyhash_length;i++)
      session_obj->keyhash[i] = 0x0;
    slan_memfree(session_obj->keyhash);
  }
  session_obj->keyhash = NULL;

}

static void client_shutdown(slan_session_t *session_obj) {
  
  syslog(LOG_INFO,"Connection closing...\n");

  session_obj->cipher->close(session_obj->cipher);
  session_obj->mac->close(session_obj->mac);
  forget_keys(session_obj);

  /* Transport layer destruction should go here */

  /* Closing the control channel (TCP) socket alerts the server to the fact
     that we are shuting down. */
  shutdown(session_obj->control_tcp->fd, SHUT_RDWR);
  free(session_obj->control_tcp);

  /* This should cause the process to exit */
  pthread_exit(NULL);
}

static void change_key(slan_session_t *session_obj) {

  forget_keys(session_obj);
  if (dh_initiate(session_obj)) {
    syslog(LOG_ERR,"Change session key: DH exchange failed!\n");
    syslog(LOG_ERR,"The key was not changed.\n");
  } else {
    syslog(LOG_INFO,"Setting new MAC and cipher keys....\n"); fflush(stdout);
    session_obj->cipher->set_key(session_obj->cipher, session_obj->key,
				 session_obj->key_length); 
    
    session_obj->mac->set_key(session_obj->mac, session_obj->mackey,
			      session_obj->mackey_length); 
    /* Reset sequence numbers: Maybe this should be randomized a bit */
    session_obj->sequence_down = 0;
    session_obj->sequence_up = 0;
    
    session_obj->key_timestamp = time(NULL);
    /* Unblock the transport layer. */
  }
  session_obj->key_change = 0;
}

/* This is called by the signal handler */
static void interrupt_shutdown(int signal) {
  shutdown_now = 1;
}

static void client_unexpected_shutdown(slan_session_t *session_obj) {

  syslog(LOG_EMERG,"\n\nThe connection to the server has unexpectedly "
	  "closed.\n");
  syslog(LOG_EMERG,"The session will be terminated (sorry!).\n\n");

  /* Don't restore default routes -- force user to know (network will stop
     working) that SLAN has terminated without user's request */
  dont_restore = 1;

  /* !! This should restart the connection automatically, if possible, instead 
        of full shutdown. !! */
  client_shutdown(session_obj);
}


static void *control_loop(void *arg) {
  slan_session_t *session_obj;
  struct pollfd ufds;
  control_msg_t *msg, *new_msg;
  int rval, timeout;

  session_obj = arg;
  for(;;) {
    ufds.fd = session_obj->control_tcp->fd;
    ufds.events = ( POLLIN | POLLPRI );
    ufds.revents = 0;

    /* Give up after 2 seconds, check to see if we should do anything
       and then go back to polling. */
    poll(&ufds, 1, 2000);
    pthread_mutex_lock(session_obj->lock);
    if (shutdown_now) {
      logmsg(MSG_NORMAL,"Interrupt Received, shutdown started.\n");
      client_shutdown(session_obj);
    }

    if (ufds.revents & POLLHUP) {
      /* If the server hangs up, we shutdown immediately. */
      client_unexpected_shutdown(session_obj);
    }
    if (ufds.revents &  POLLPRI) {
      /* Our protocol doesn't use out of band data. So we just ignore it */
      logmsg(MSG_ERROR,"Out of band data waiting... why is that?\n");
      continue;
    }
    if (ufds.revents & POLLERR) {
      /* I don't know what might cause this to occur, so for now, we'll
	 shutdown the connection if we have an error */ 
      logmsg(MSG_ERROR,"Error condition on control channel.");
      client_shutdown(session_obj);
    }
    if (ufds.revents & POLLIN) {
      logmsg(MSG_DEBUG,"POLLIN on control channel\n");
      timeout = 100;
      rval = control_msg_recv(&msg, session_obj, &timeout);
      if (rval < 0) {
	logmsg(MSG_ERROR,"Closing session due to tcp error\n");
	client_shutdown(session_obj);
      }
      if (rval > 0) {
	logmsg(MSG_WARNING,"Control message expected but no data received\n");
	pthread_mutex_unlock(session_obj->lock);
	continue;
      }

      switch(msg->type) {
      case KEYCHANGE:
	if (msg->subtype == KEYCHANGE_READY) {
	  control_msg_destruct(msg);
	  change_key(session_obj);
	} else {
	  session_obj->key_change = 1;
	  new_msg = new_control_msg(KEYCHANGE, KEYCHANGE_READY, session_obj);

	  timeout = 1000;
	  rval = control_msg_send(new_msg, session_obj, &timeout);
	  control_msg_destruct(new_msg);
	  if (rval < 0) {
	    logmsg(MSG_ERROR,"Closing session due to tcp error\n");
	    control_msg_destruct(msg);
	    client_shutdown(session_obj);
	  }
	  if (rval > 0) {
	    session_obj->key_change = 0;
	    control_msg_destruct(msg);
	    continue;
	  }

	  pthread_mutex_unlock(session_obj->lock);
	  control_msg_destruct(new_msg);
	  control_msg_destruct(msg);
	  usleep(1000);
	  pthread_mutex_lock(session_obj->lock);
	  change_key(session_obj);	  
	}
	pthread_mutex_unlock(session_obj->lock);
	continue;
	break;
      case SHUTDOWN:
	/* This should not return */
	client_shutdown(session_obj);
	break;
      case KEEPALIVE:
	if (msg->subtype != KEEPALIVE_PING) {
	  /* This should not happen. */
	  logmsg(MSG_WARNING,"Not expecting a keepalive response here.\n");
	  control_msg_destruct(msg);
	} else {
	  new_msg = new_control_msg(KEEPALIVE, KEEPALIVE_RESPONSE, 
				    session_obj);
	  timeout = 1000;
	  rval = control_msg_send(new_msg, session_obj, &timeout);
	  control_msg_destruct(new_msg);
	  if (rval < 0) {
	    logmsg(MSG_ERROR,"Closing session due to tcp error\n");
	    control_msg_destruct(msg);
	    client_shutdown(session_obj);
	  }
	  if (rval > 0) {
	    control_msg_destruct(msg);
	    continue;
	  }
	}
	break;
      default:
	logmsg(MSG_WARNING,"Unhandled control message type: %d\n",msg->type);
	control_msg_destruct(msg);
	break;
      }
    }
    
    /* Check the session key's age. Change it if it should be expired. This is
       the client's responsibility to enforce. */
    if ((time(NULL) - session_obj->key_timestamp) > KEY_LIFETIME &&
	session_obj->key_change == 0) {
      session_obj->key_change = 1;
      control_msg_destruct(msg);

      new_msg = new_control_msg(KEYCHANGE, KEYCHANGE_REQUEST, session_obj);

      timeout = 1000;
      rval = control_msg_send(new_msg, session_obj, &timeout);
      control_msg_destruct(new_msg);
      if (rval < 0) {
	logmsg(MSG_ERROR,"Closing session due to tcp error\n");
	client_shutdown(session_obj);
      }
      if (rval > 0) {
	continue;
      }
    }
    pthread_mutex_unlock(session_obj->lock);
  }
}

/* This function continuouly reads either UDP socket for incoming traffic or
   the SLAN socket for outgoing traffic, using the builtin transport layer
   functions. They are non-blocking, using select to test for data before
   reading.  */
static void *transport_loop(void *arg) {
  slan_session_t *session_obj;
  unsigned char buf[10240];
  unsigned char *p;
  slan_message_t slan_message;
  int slan_socket, udp_socket, max_fd;
  fd_set read_fds;

  session_obj = arg;

  slan_message.buffer = buf;
  slan_message.maxspace = 4096;

  p = buf + session_obj->header_save;
  
  slan_message.header = slan_message.data = p;
  slan_message.trailer = p;
  slan_message.length = 0;

  udp_socket = session_obj->udp_socket;
  slan_socket = session_obj->slan_socket;
  max_fd = ((udp_socket > slan_socket)?udp_socket:slan_socket) + 1;

  FD_ZERO(&read_fds);
  
  for(;;) {
    FD_SET(udp_socket, &read_fds);
    FD_SET(slan_socket, &read_fds);
    
    select(max_fd, &read_fds, NULL, NULL, NULL);
    
    pthread_mutex_lock(session_obj->lock);

    if (FD_ISSET(slan_socket, &read_fds) && !session_obj->key_change) {
      session_obj->transport_top->down(session_obj->transport_top, 
				       &slan_message);
      slan_message.header = slan_message.data = p;
      slan_message.trailer = p;
      slan_message.length = 0;
    }
    
    if (FD_ISSET(udp_socket, &read_fds)) {
      session_obj->transport_bottom->up(session_obj->transport_bottom, 
					&slan_message);
      slan_message.header = slan_message.data =p;
      slan_message.trailer = p;
      slan_message.length = 0;
    }
    pthread_mutex_unlock(session_obj->lock);
  }   
}

static int configure_mac_address(unsigned char *mac_address) {
  struct ifreq ifr;
  struct sockaddr sa;
  int sock;


  strcpy(ifr.ifr_name,"tap0");
  sock = socket(PF_INET, SOCK_DGRAM, 0);

  sa.sa_family = ARPHRD_ETHER;
 

  memcpy(sa.sa_data, mac_address, 6);
  memcpy((char *) &ifr.ifr_hwaddr, &sa, sizeof(struct sockaddr));

  if (ioctl(sock, SIOCSIFHWADDR, &ifr) < 0) {
    logmsg(MSG_ERROR,"Failed to set hardware address for interface (%s)\n",
	   strerror(errno));
  }

  /* TODO: Dynamically figure out header requirements for the tunnel and set
     the MTU appropriately */
  ifr.ifr_mtu = 1420;
  if (ioctl(sock, SIOCSIFMTU, &ifr) < 0) {
    logmsg(MSG_ERROR, "Failed to set maximum transfer unit (MTU) for slan "
	   "interface (%s)\n",strerror(errno));
  }

  close(sock);
  return 0;

}

/* On signal SIGUSR1 -- dump status to /var/tmp socket */
static void signal_statusupdate(int signal) {
  int fd;
  FILE *status_file = NULL;

  fd = open(PIPE_FILENAME, O_WRONLY|O_NONBLOCK);
  if (fd < 0) { 
    syslog(LOG_ERR,"Can not open fifo to communicate status information. "
	    "(%s).\n",strerror(errno));
    return;
  } else {
    status_file = fdopen(fd, "w");
  }

  fprintf(status_file,"SLAN is (still) running\n");
  fprintf(status_file,"Server is %s\n",slan_global_config->server_address);
  fprintf(status_file,"Bytes received: %u kB\tsent: %u kB\n",
	  session_obj->bytes_in/1024, session_obj->bytes_out/1024);
  fprintf(status_file,"Packets received: %u\tsent: %u\n", 
	  session_obj->packets_in, session_obj->packets_out);
  fprintf(status_file,"Keys last changed %d seconds ago.\n",
	  (int) (time(NULL) - session_obj->key_timestamp));
  fprintf(status_file,"---- END OF STATUS REPORT ----\n");
  fflush(status_file);
  fclose(status_file);
}


/* This function attempts to find the PID of a running SLAN client, if one
   exists, by looking for a file created in /var/run which has the PID of
   a running SLAN client, and checking to see if /proc reports that the 
   name of the program running with that PID (if there is one) is
   slan_client. If all is satisfied, returns the PID of the running SLAN
   client. Note that this will fail to detect a running client if it is run
   under a different name than slan_client. */
static pid_t find_runningpid() {
  pid_t pid, return_value;
  FILE *f;
  char inputline[256];

  f = fopen(PID_FILENAME,"r");
  if (f == NULL) {
    return -1;
  }
  fgets(inputline,255,f);
  inputline[255]=0;
  pid = atol(inputline);
  if (pid < 0 || pid > 0xFFFF) return -1;
  
  fclose(f);
  
  sprintf(inputline,"/proc/%u/stat",pid);
  f = fopen(inputline,"r");
  if (f == NULL) return -1;
  
  fgets(inputline,255,f);
  inputline[255] = 0;
  if (strstr(inputline, "slan_client")) {
    return_value = pid;
  } else {
    return_value = -1;
  }

  fclose(f);
  return return_value;
}
  

static void client_getstatus() {
  pid_t pid;
  int fd;
  FILE *f;
  fd_set read_fds, err_fds;
  int rv;
  struct timeval timeout;

  timeout.tv_sec = 5;
  timeout.tv_usec = 0;

  pid = find_runningpid();
  if (pid == -1) {
    fprintf(stderr,"SLAN client does not appear to be running.\n");
    exit(-1);
  }

  unlink(PIPE_FILENAME);
  if (mknod(PIPE_FILENAME, (S_IFIFO | 0600), 0)) {
    fprintf(stderr,"Warning: Can not create fifo to communicate status "
	    "information. (%s).\n",strerror(errno));
    return;
  }
  fd = open(PIPE_FILENAME, O_RDONLY | O_NONBLOCK);
  fchown(fd, NOBODY_UID, NOBODY_GID);

  if (fd == -1) {
    fprintf(stderr,"Error: Unable to open SLAN client communication pipe "
	    "for slan client process #%u (%s)\n",pid,strerror(errno));
    exit(-1);
  }

  if (kill(pid, SIGUSR1)) {
    fprintf(stderr,"Error: Unable to send \"get status\" signal to running "
	    "SLAN client process #%u (%s)\n",pid,strerror(errno));
    if (errno == EPERM) 
      fprintf(stderr,"Are you root?\n");
    exit(-1);
  }

  f = fdopen(fd, "r");
  FD_ZERO(&read_fds);
  FD_ZERO(&err_fds);
  FD_SET(fd, &read_fds);
  FD_SET(fd, &err_fds);

  rv = select(fd + 1, &read_fds, NULL, &err_fds, &timeout);
  if (rv == -1) {
    fprintf(stderr,"Error: select() failed on slan client pipe. (%s)",
	    strerror(errno));
    exit(-1);
  }
  if (rv == 0) {
    fprintf(stderr,"Time out waiting for running SLAN client (%u) to "
	    "respond. SLAN client is\nprobably frozen.\n", pid); 
    fprintf(stderr,"Try -k to kill it.\n");
    exit(-1);
  }
  if (FD_ISSET(fd, &err_fds)) {
    fprintf(stderr,"Error: Exception on select() for slan client pipe.\n");
    fprintf(stderr,"Try again or use -k to kill the running slan client.\n");
    exit(-1);
  }
  if (FD_ISSET(fd, &read_fds)) {
    char statusline[1024];
    do {
      fgets(statusline, 1023, f);
      fprintf(stdout,"%s",statusline);
    } while (strncmp(statusline,"----",4) && !feof(f));
    fclose(f);
    unlink(PIPE_FILENAME);
    exit(0);
  }

  fprintf(stderr,"Internal Error. Can not continue (This should not happen)\n");
  exit(0);

}

/* Parse command line options */
static void comline(int argc, char *argv[], char **config_filename)  {

  static struct option long_options[] = {
    { "status", no_argument, NULL, 's' },
    { "kill", no_argument, NULL, 'k' },
    { "config", required_argument, NULL, 'C' },
    { "connect-only", no_argument, NULL, 'r' },
    { "user", required_argument, NULL, 'u' },
    { "password", required_argument, NULL, 'p' }
  };

  const char optstring[] = "skC:ru:p:";
  pid_t pid;

  opterr = 0;

  while(1) {
    switch(getopt_long(argc, argv, optstring, long_options, NULL)) {
    case -1:
      return;
    case 's':
      client_getstatus();
      break;
    case 'k':
      pid = find_runningpid();
      if (pid > 0) {
	/* Don't do this if we get -1 out of find_runningpid()! -- That will
	   kill everything on the system! */
	if (kill(pid, SIGINT)<0) {
	  fprintf(stderr,"Could not send shutdown signal to slan client "
		  "process %u (%s)\n",pid,strerror(errno));
	} else {
	  fprintf(stderr,"Sent shutdown signal to SLAN client process %u\n",
		  pid);
	}
      } else {
	fprintf(stderr,"Can not find running SLAN client.\n");
      }
      exit(0);
      break;
    case 'C':
      *config_filename = strdup(optarg);
      break;
    case 'r':
      connect_only = 1;
      //fprintf(stderr,"Robo-client connect only not implemented.\n");
      break;
    case 'u':
      preset_username = strdup(optarg);
      break;
    case 'p':
      preset_password = strdup(optarg);
      bzero(optarg,strlen(optarg));
      break;
    default:
      fprintf(stderr,"Unknown Option %c\n",optopt);
      break;
    }
  }

}

typedef struct rtable {
  char interface[32];
  u_int32_t network;
  u_int32_t gateway;
  u_int32_t mask;
  unsigned mtu;
  char hwaddr[32]; /* Should only need 6 here, but we'll leave room in case we
		      accidentally load a non 802 hw address that is longer */
  unsigned flags;
  struct rtable *next;
} route_t;
/* This stores the default routes at startup which we will delete upon
   successful SLAN connection, and then restore when SLAN is shutdown */
static route_t *startup_default_routes = NULL;

/* This stores a host route created to reach the SLAN server with tunnel
   traffic through the real network. We reference it later to delete the
   host route on shutdown. */
static route_t tunnel_route; 

static void load_routing_table() {
  FILE *routing_table;
  int sock;
  struct sockaddr sa;
  u_int32_t network, gateway, mask, flags, mtu;
  char interface[32];
  char inputline[256];
  unsigned char macaddr[6];
  struct ifreq ifr;
  route_t **p;

  p = &startup_default_routes;

  routing_table = fopen("/proc/net/route","r");
  if (routing_table == NULL) {
    logmsg(MSG_ERROR,"Error opening /proc/net/route (%s)\n",strerror(errno));
    logmsg(MSG_ERROR,"Unable to inspect your system's network routing table. "
	   "Can not continue.\n");
    logmsg(MSG_FATAL,"Please check your physical network configuration.\n");
  }

  sock = socket(PF_INET, SOCK_DGRAM, 0);
  if (sock < 0) {
    logmsg(MSG_ERROR,"Can't open datagram socket in load_routing_table() (%s). "
	   "Very strange\n");
    logmsg(MSG_FATAL,"Unable to inspect your system's network routing table.\n");
  }

  /* Chew off the header line which just contains column labels */
  while(getc(routing_table)!='\n');

  fgets(inputline,256,routing_table);
  while(!feof(routing_table)) {
    sscanf(inputline,"%30s %x %x %x %*x %*x %*x %x %x",
	   interface, &network, &gateway, &flags, &mask, &mtu);
    if (network == 0x0 && (flags & (RTF_UP|RTF_GATEWAY))) {
      /* This is a default route */
      (*p) = malloc(sizeof(route_t));
      if ((*p) == NULL) {
	logmsg(MSG_FATAL,"Out of memory allocating %d bytes in load_routing_table()\n",
	       sizeof(route_t));
      }

      /* Attempt to find hw address (MAC address) for the interface */
      memset(&ifr,0x0,sizeof(struct ifreq));
      strncpy(ifr.ifr_name,interface, IFNAMSIZ);

      /* We expect an ETHER Hardware address family... this will fail if it is
	 not. */
      memset(&sa, 0x0, sizeof(struct sockaddr));
      sa.sa_family = ARPHRD_ETHER;
      
      memcpy((char *) &ifr.ifr_hwaddr, &sa, sizeof(struct sockaddr));
      if (ioctl(sock, SIOCGIFHWADDR, &ifr) < 0) {
	logmsg(MSG_CRITICAL,"Can not read mac address for phyiscal interface on default "
	       "route. (%s)\n", strerror(errno));
      } else {
	memcpy(macaddr, (unsigned char *) ifr.ifr_hwaddr.sa_data, 6);
	logmsg(MSG_DEBUG,"Default route on interface %s : %02X:%02X:%02X:%02X:%02X:%02X\n",
	       interface, (unsigned int) macaddr[0], (unsigned int) macaddr[1],
	       (unsigned int) macaddr[2], (unsigned int) macaddr[3],
	       (unsigned int) macaddr[4], (unsigned int) macaddr[5]);
	memcpy((*p)->hwaddr,macaddr,6);
      }

      strncpy((*p)->interface, interface,31);
      (*p)->network = network;
      (*p)->gateway = gateway;
      (*p)->flags = flags;
      (*p)->mask = mask;
      (*p)->mtu = mtu;
      (*p)->next = NULL;
      p = &((*p)->next);

    }
    fgets(inputline,256,routing_table);

  }

  if (routing_table == NULL) {
    logmsg(MSG_WARNING,"No default route found for current network configuration.\n");
  }
  fclose(routing_table);
  close(sock);
}

static void fill_routing_entry(struct rtentry *rte, route_t *route) {
  struct sockaddr_in sa;

  memset(rte, 0x0, sizeof(struct rtentry));
  sa.sin_family = PF_INET;

  sa.sin_addr.s_addr = route->network;
  memcpy(&rte->rt_dst, &sa, sizeof(struct sockaddr_in));
  
  sa.sin_addr.s_addr = route->gateway;
  memcpy(&rte->rt_gateway, &sa, sizeof(struct sockaddr_in));
  
  sa.sin_addr.s_addr = route->mask;
  memcpy(&rte->rt_genmask, &sa, sizeof(struct sockaddr_in));

  rte->rt_dev = route->interface;

  rte->rt_flags = route->flags;
}


/* This function is called after successful client startup. It adds a host
   route to the server via the real network interface, and disables ALL
   default routes. */
static void modify_routing_table(char *server_address) {
  struct rtentry rte;
  struct hostent *hostname_hent;
  route_t *p;
  int sock;
  
  sock = socket(PF_INET, SOCK_DGRAM, 0);
  if (sock < 0) {
    logmsg(MSG_ERROR,"Failed to open socket for routing table control (%s)\n",
	   strerror(errno));
  }

  /* First, we take the first default route we find and use it as our tunnel
     route. We assume that the kernel will choose the first default route from
     the routing table (unless somebody actually uses the metric feature) and
     that it is how we are currently communicating with the SLAN server. */
  p = startup_default_routes;
  
  /* Create a copy of the default route and modify it to make a host route to
     the SLAN server */
  memcpy(&tunnel_route, p, sizeof(route_t));

  hostname_hent = gethostbyname(server_address);
  if (hostname_hent == NULL) {
    logmsg(MSG_CRITICAL,"Unable to resolve server hostname %s (%s)\n",
	   server_address, hstrerror(h_errno));
  }
  else {
    memcpy(&tunnel_route.network, hostname_hent->h_addr_list[0], 4);
    tunnel_route.mask = 0xFFFFFFFF;
    tunnel_route.next = NULL;
    tunnel_route.flags = (RTF_UP|RTF_GATEWAY|RTF_HOST);
    fill_routing_entry(&rte, &tunnel_route);
    
  
    if (ioctl(sock, SIOCADDRT, &rte) < 0) {
      logmsg(MSG_CRITICAL, "Failed to set host route for SLAN tunnel to server. (%s)\n",
	     strerror(errno));
      logmsg(MSG_DEBUG, "Destination %0X\n",tunnel_route.network);
    }
  }

  /* Now delete all default routing table entries */
  while(p!=NULL) {
    fill_routing_entry(&rte, p);
    if (ioctl(sock, SIOCDELRT, &rte) < 0) {
      logmsg(MSG_FATAL,"Error deleting default route on interface %s (%s)\n",p->interface,
	     strerror(errno));
    }
    p = p->next;
  }

  close(sock);
}

static void restore_startup_routing(void) {
  int sock;
  struct rtentry rte;
  route_t *p;

  sock = socket(AF_INET, SOCK_DGRAM, 0);
  if (sock < 0) {
    logmsg(MSG_ERROR,"Cannot open socket for routing table restoration (%s)\n",
	   strerror(errno));
    return;
  }

  /* Restore all the default routes we had at startup */
  p = startup_default_routes;
  while(p!=NULL) {
    fill_routing_entry(&rte, p);
    if (ioctl(sock, SIOCDELRT, &rte) < 0) {
      logmsg(MSG_ERROR,"Failed to restore a default route (%s)\n",strerror(errno));
    }
    p = p->next;
  }


  /* Delete our host route to the server */
  fill_routing_entry(&rte, &tunnel_route);
  if (ioctl(sock, SIOCDELRT, &rte) < 0) {
    logmsg(MSG_ERROR,"Failed to delete host route to server (%s)\n",strerror(errno));
  }

  close(sock);
  return;
}

static void client_startup(slan_session_t *session_obj,
			   int *server_addr, int *server_route) {
  unsigned char null_key[32];
  unsigned char macaddr[6]; 

  session_obj->control_tcp  = 
    client_open_socket(slan_global_config->server_address, server_addr, 
		       slan_global_config->control_port);
  session_obj->lock = slan_calloc(1, sizeof(pthread_mutex_t));
  pthread_mutex_init(session_obj->lock, NULL);

  load_routing_table();
  if (startup_default_routes != NULL) {
    memcpy(macaddr, startup_default_routes->hwaddr, 6);
  } else {
    logmsg(MSG_WARNING,"Did not find a MAC address on a default route, using a random one.\n");
    random_byte_string(macaddr, 6);
  }
  macaddr[0] = 0xFE;
  macaddr[1] = 0xFD;

  /* Establish whether or not the server is compatible with the client. In the 
     future this may instruct the server (or client) to do fancy things like
     enable or disable certain features or bug fixes. For now, it really
     doesn't do anything. */
  logmsg(MSG_NORMAL,"Checking version compatibility...");
  if (check_version(session_obj->control_tcp->fd)<0) {
    logmsg(MSG_ERROR,"Error, version of server and client incompatible.\n");
    exit(-1);
  } else {
    logmsg(MSG_INFO,"Server Version okay");
    logmsg(MSG_NORMAL,"\n");
  }

  logmsg(MSG_INFO,"Initializing to default key values...\n"); fflush(stdout);
  /* The DH key exchange will use control message functionality which expects
     to have these things set up and working, except we don't know the key
     until after DH completes. So we'll set them up with null keys (0x0) so
     they function, but they will not provide any security ofcourse. */
  memset(null_key, 0x0, 32);
  session_obj->mac = get_hmac_sha1();
  session_obj->cipher = get_blowfish();
  session_obj->mac->set_key(session_obj->mac, null_key, 32);
  session_obj->cipher->set_key(session_obj->cipher, null_key, 32);
  session_obj->key_timestamp = time(NULL);

  /* Establish the session key via Diffie-Hellman key exchange. The value
     computed by this exchange will be used to produce a symmetric key for
     data privacy and a Message Authentication Code (MAC) key for protecting
     the messages against alterations. We also compute a one-way hash of the
     key value computed and use that to defeat a Diffie-Hellman
     man-in-the-middle attack by using it in the RSA authentication challenge
     which authenticates the server. */
  logmsg(MSG_NORMAL,"Establishing session key with server (Diffie-Hellman)...\n");
  fflush(stdout);
  if (dh_initiate(session_obj)) {
    logmsg(MSG_FATAL,"Diffie-Hellman key exchange failed.\n");
    shutdown(session_obj->control_tcp->fd, SHUT_RDWR);
    exit(1);
  }

  logmsg(MSG_INFO,"Diffie-Hellman Completed. Setting session keys....\n"); 

  /* Setup the symmetric cipher. */
  session_obj->cipher->set_key(session_obj->cipher, session_obj->key,
			       session_obj->key_length); 

  /* Setup the message authentication code */
  session_obj->mac->set_key(session_obj->mac, session_obj->mackey,
			    session_obj->mackey_length); 
  
  logmsg(MSG_INFO,"Symmetric Encryption key: %d bits.\n",
	  session_obj->key_length * 8);
  logmsg(MSG_INFO,"Message Authentication Code key: %d bits.\n",
	  session_obj->mackey_length * 8);
  fflush(stdout);

  /* Authenticate the server. This is a required part of the protocol for any
     service provider. The public key should be given in the service set
     definitions loaded above (some day) */
  logmsg(MSG_NORMAL,"Establishing authenticity of server (RSA)...\n");
  if (rsaauth_initiate(session_obj)) { 
    logmsg(MSG_FATAL,"Server identity cannot be verified (FAILED RSA CHALLENGE)\n");
    shutdown(session_obj->control_tcp->fd, SHUT_RDWR);
    exit(-1);
  }
  logmsg(MSG_INFO, "Server identity verified.\n\n");

  logmsg(MSG_NORMAL, "Client (user) authentication...\n");
  if (client_clientauth_wait(session_obj)) {
    logmsg(MSG_FATAL,"Client authentication to server failed.\n");
    shutdown(session_obj->control_tcp->fd, SHUT_RDWR);
    exit(-1);
  }
  logmsg(MSG_INFO,"Client (you) accepted.\n\n"); fflush(stdout);

  logmsg(MSG_NORMAL,"Negotiating a hardware address with server...\n");
  fflush(stdout);
  

  if (macconf_request(macaddr, session_obj)) {
    logmsg(MSG_FATAL,"Unable to settle dispute with server. Unable to obtain "
	   "a hardware address\n");
    shutdown(session_obj->control_tcp->fd, SHUT_RDWR);    
    exit(-1);
  }
  logmsg(MSG_INFO,"MAC Addresss: %02X:%02X:%02X:%02X:%02X:%02X\n",
	  macaddr[0],
	  macaddr[1],
	  macaddr[2],
	  macaddr[3],
	  macaddr[4],
	  macaddr[5]);
  client_setup_transport_layer(session_obj,
			       slan_global_config->server_address);
  configure_mac_address(macaddr);
  modify_routing_table(slan_global_config->server_address);
}

static void create_pidfile(void) {
  FILE *f;
  pid_t pid;

  pid = getpid();
  f = fopen(PID_FILENAME, "w");

  if (f == NULL) {
    fprintf(stderr,"Warning: Can not open %s to write process id information. "
	    "(%s)\n",PID_FILENAME,strerror(errno));
    fprintf(stderr,"You will not be able to read status information from the "
	    "running client.\n");
    fprintf(stderr,"SLAN will have to be terminated manually using kill(1)\n");
    fprintf(stderr,"Send the slan_client process with the smallest PID\n");
    fprintf(stderr,"an interrupt signal (SIGINT or kill -2)\nto terminate");
    fprintf(stderr,"your SLAN session.\n");
    return;
  }

  fprintf(f,"%u\n",pid);
  fclose(f);

  return;
}


/* This is the entry point for the client side user space daemon program 

   TO BE COMMENTED MORE LATER */
int main(int argc, char *argv[]) {
  int ipconfig_sock;
  unsigned int server_route, server_addr;
  pthread_t transport_thread, control_thread;
  char *config_filename;
  pid_t pid;
  pthread_attr_t thread_attr;

  configure_logmsg(0,1,MSG_DEBUG);
#ifndef DEBUG
  /* Lock ALL memory -- This process is important to functioning of the
     network for the client, and should never be swapped out. As well (and
     more importantly), the memory used by this process holds vital security
     sensitive information, such as cryptographic keys, and must never be
     written to permanent store. */
  if (mlockall(MCL_FUTURE)) {
    perror("Failed locking memory");
    fprintf(stderr,"This process must be run as root.\n");
    exit(-1);
  }
#endif

  /* Before doing anything, gather entropy and setup our random number
     generator. This will be used to provide security, so it better be good. */
  random_initialize();

  /* This function call never returns if command line parsing fails. In 
     fact, it only returns if we are going to start the SLAN client for
     real */
  config_filename = NULL;
  comline(argc, argv, &config_filename);

  if (slan_client_load_configuration(config_filename)) {
    fprintf(stderr,"Cannot start slan: Failed loading configuation file.\n");
    exit(-1);
  }

  pid = find_runningpid();
  if (pid>0) {
    fprintf(stderr,"Appears to be an SLAN client already running (pid %u)\n",
	    pid);
    fprintf(stderr,"Use --status to try to read its status.\n");
    exit(-1);
  }

  if (!connect_only) {
    struct ifreq ifr;
    strcpy(ifr.ifr_name,"tap0");

    
    /* Check for tap0 device and take it down if it is running */
    ipconfig_sock = socket(AF_INET, SOCK_DGRAM, 0);

    /*ifr.ifr_flags = IFF_NOARP | IFF_BROADCAST;
    if (ioctl(ipconfig_sock, SIOCSIFFLAGS, &ifr) < 0) {
      fprintf(stderr,"Can't find interface tap0 (%s)\n",strerror(errno));
      fprintf(stderr,"Is the device loaded (insmod/modprobe)?\n");
      exit(-1);
    }
    */
  } else {
    ipconfig_sock = -1;
  }

  /* Allocate and setup a session object which will hold some state
     information for the active connection. 

     TO BE DONE:
     (1) The server address is to be determined from the service set
     information, when that functionality is added */
  session_obj = slan_calloc(1, sizeof(slan_session_t));

  client_startup(session_obj, &server_addr, &server_route);

  if (connect_only) {
    close(session_obj->control_tcp->fd);
    exit(0);
  }

  fprintf(stdout,"Session startup okay, SLAN secured network link running.\n");

  /* Lock memory again! Locks are not inherited after fork() */
  if (mlockall(MCL_CURRENT|MCL_FUTURE)) {
    perror("Failed locking memory");
    fprintf(stderr,"This process must be run as root.\n");
    exit(-1);
  }

  pthread_attr_init(&thread_attr);
  pthread_attr_setstacksize(&thread_attr, 65535);
  if (setegid(SLANRUN_GID) || seteuid(SLANRUN_UID)) {
    fprintf(stderr,"SLAN (Warning): Can't change user to drop root "
	    "priveledges\n");
  } else {
    fprintf(stderr,"SLAN client: Root priveledge dropped.\n");
}

#ifndef DEBUG
  if ((pid=fork())!=0) {
    if (pid == -1) {
      fprintf(stderr,"Can not fork into background! (%s)\n",strerror(errno));
      fprintf(stderr,"Exiting. Something must be way messed up.\n");
      exit(-1);
    } else {
      fprintf(stderr,"Forked into background. Child is [%d]\n",pid);
      return 0;
    }
  } else {
    fclose(stderr);
    fclose(stdout);
    fclose(stdin);
  }
#endif

  create_pidfile();
  pthread_create(&transport_thread, &thread_attr, transport_loop, session_obj);
  
  pthread_create(&control_thread, &thread_attr, control_loop, session_obj);

  signal(SIGINT, interrupt_shutdown);
  signal(SIGUSR1, signal_statusupdate);
    
  /* When the control thread exits, we are done. Other things may terminate
     the process entirely however.  */
  pthread_join(control_thread, NULL);
  pthread_kill(transport_thread, 15);
  
  if (seteuid(0))
    syslog(MSG_WARNING,"Failed to regain root priveledges (%m)\n");

  restore_startup_routing();
  close(session_obj->control_tcp->fd);

  syslog(LOG_INFO,"Slan session finished.\n"); fflush(stdout);
  return 0;
}


/* Unused code */

/* This function is never called by the client program, but required to in
   order for some shared code between the client and server to link properly
   into the client executible */    
u_int32_t macaddr_lookup(unsigned char *p) {

  return 0;
}
