/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/platform/linux/slan_device-2.2.c,v 1.1.1.1 1999/12/09 01:16:38 koni Exp $ 
  $Author: koni $ $Date: 1999/12/09 01:16:38 $ $Name: slan-release-0-04 $
  $Revision: 1.1.1.1 $

*/
   
/*

  KONI Oct. 25-27, 2000
  Stage 0 device driver for "lightlink secure lan" project
  
  Oct. 27: Stage 0 demonstrated. This kernel module redirects packets sent 
  to this interface from the IP stack to the slan daemon process, and packets
  sent from the slan daemon process to the IP stack.

  This code was derived in part from loopback.c, skeleton.c, dummy.c, and 
  some other specific (real) device drivers in linux/drivers/net
*/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/socket.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/in.h>
#include <linux/init.h>

#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/if_ether.h>	/* For the statistics structure. */
#include <linux/if_arp.h>	/* For ARPHRD_ETHER */

/* This structure gets pushed onto the packets received from the kernel,
   and pulled from the packets received from the daemon process. The presense 
   of the magic number at the top of a packet indicates the difference. An 
   IP packet cannot have 0xFF00FF00 (our magic number here) as the first 32
   bit word of its header */
struct slan_header {
  __u32 magic_number;
  __u16 our_protocol;
  __u16 orig_protocol;
};

/* slan_xmit()

   This function does all the work:

   It seperates the packets by detecting the magic number, and redirects 
   appropriately. In both cases, netif_rx() is called to pass the packet to
   the kernel. */
static int slan_xmit(struct sk_buff *skb, struct device *dev) {
  struct net_device_stats *stats = (struct net_device_stats *)dev->priv;
  struct slan_header *p;
  
  if (skb == NULL || dev == NULL) 
    printk(KERN_DEBUG "Lightlink Secure Lan Error: Null SKB pointer\n");

  /* This code is stolen from loopback.c. I think it is cloning the skb in
     case a higher layer whichs to hold on to it (and therefore calling 
     dev_kfree_skb() won't muck up TCP or something.

     TODO: Verify that. */
  if(atomic_read(&skb->users) != 1)	{
    struct sk_buff *skb2=skb;    
    skb=skb_clone(skb, GFP_ATOMIC);    
    if(skb==NULL) {
      kfree_skb(skb2);
      return 0;
    }
    kfree_skb(skb2);
  }
  else
    skb_orphan(skb);

  /* Check the presence of the "magic number" */
  if (*((__u32 *) skb->data) == 0xFF00FF00) {
    /* This number at the head of the packet indicates that it came from the
       slan daemon program, not from the IP stack. So, we pull the slan header
       off, revealing the original packet, change the protocol type in the 
       skb structure to the original protocol, and direct the packet to the
       IP stack. */
    p = (struct slan_header *) skb->data;
    skb->protocol = p->orig_protocol;
    skb_pull(skb, sizeof(struct slan_header));

    skb->mac.raw = skb->data;
    skb->h.raw = skb->data;

    /* We do not need to checksum anything as nothing went on the network */
    skb->ip_summed = CHECKSUM_UNNECESSARY;
    skb->dev = dev;

    stats->rx_packets++;
    stats->rx_bytes += skb->len;

    /*    printk(KERN_INFO"Received %d bytes from slan daemon, attempting to redirect to IP stack.\n",skb->len); */
    netif_rx(skb);

    return 0;

  } else {
    /* In this case, we add our header and redirect the packet to the slan
       daemon program */
    int room;

    /* We must check for room before doing a push. If for some reason we
       do not have room, executing the skb_push will cause the kernel to 
       crash! */
    room = skb_headroom(skb);
    if (room < sizeof(struct slan_header)) {
      printk(KERN_INFO"Lightlink Secure Lan: Insufficent room (%d bytes) in packet received from kernel\n",room);
      /* We cannot redirect this packet, dump it in the ashcan */
      dev_kfree_skb(skb);
      stats->tx_errors++;
    } 
    else {
      /* We have room to slap down our header */
      p = (struct slan_header *) skb_push(skb, sizeof(struct slan_header));
      p->magic_number = 0xFF00FF00;
      p->orig_protocol = skb->protocol;
      p->our_protocol = 0x7777;
    
      skb->h.raw = (unsigned char *) p;
      skb->mac.raw = (unsigned char *) p;
      
      skb->protocol=0x7777;
      skb->dev=dev;
      skb->ip_summed = CHECKSUM_UNNECESSARY;

      /* Redirect (receive) packet to daemon process */
      /*      printk(KERN_INFO"Sent packet to daemon process: %d bytes\n",skb->len); */
      netif_rx(skb); 

      stats->tx_bytes+=skb->len;
      stats->tx_packets++;
    }
    return(0);
  }
}

/* slan_get_stats()
   
   return a pointer to device private data which has our recorded statistics
   for the device <dev> */
static struct net_device_stats *slan_get_stats(struct device *dev) {
  return (struct net_device_stats *)dev->priv;
}

/* TODO: Help! What the hell will we do about multicast data? */
/* set_multicast_list()

   fake multicast ability for now like the linux/drivers/net/dummy.c
   dummy network driver */
static void set_multicast_list(struct device *dev)
{
}

/* slan_open()

   This function is called by the kernel when a user configures (via ifconfig)
   the slan interface. */
static int slan_open(struct device *dev) {
  int i;
  /* Make up a hardware address for our fake device driver. Use a 0x00 for 
     the first octet because a multicast or broadcast packet cannot have
     that as an address and we don't want to make up multicast addresses. */
  for(i=0;i<ETH_ALEN;i++) dev->dev_addr[i] = "\0SLAN0"[i];

  dev->start = 1;
  dev->tbusy = 0;

  MOD_INC_USE_COUNT;
  return 0;
}

/* slan_close()

   This function is called by ifconfig to take down the device */
static int slan_close(struct device *dev) {

  MOD_DEC_USE_COUNT;
  dev->start = 0;
  dev->tbusy = 1;
  return 0;
}

/* This function is called at module load time by insmod or modprobe. It
   sets up the device structure so the kernel can open/close the device
   and whatever else. */
__initfunc(int slan_init(struct device *dev))
{
	/* Initialize the device structure. */
	dev->hard_start_xmit	= slan_xmit;

	/* Get memory for network statistics information, like that reported 
	   by ifconfig */
	dev->priv = kmalloc(sizeof(struct net_device_stats), GFP_KERNEL);
	if (dev->priv == NULL) {
	  printk(KERN_INFO"Lightlink Secure Lan: Error, cannot allocate memory for device statistics.\n");
	  return -ENOMEM;
	}
	memset(dev->priv, 0, sizeof(struct net_device_stats));

	/* The kernel knows how to manipulate the device by these pointers
	   in the device (dev) structure. */
	dev->get_stats	= slan_get_stats;
	dev->open =       slan_open;
	dev->stop =       slan_close;
	dev->set_multicast_list = set_multicast_list;

	/* We are mostly pretending to be an ethernet device, so we use
	   the generic ethernet device setup function to set defualt values
	   for our fake device. We overide some settings below */
	ether_setup(dev);

	/* Override some ethernet defualts:
	   We don't need a queue.
	   No ARP on our interface.
	   We don't want multicast going on
	   Reduce MTU size because we are going to push some extra headers 
	*/
	dev->tx_queue_len = 0;
	dev->flags |= IFF_NOARP;
	dev->flags &= ~IFF_MULTICAST;
	dev->mtu = 1400;

	dev->hard_header = NULL;
	dev->rebuild_header = NULL;
	dev->set_mac_address = NULL;
	dev->hard_header_parse  = NULL;
	dev->hard_header_cache = NULL;
	dev->header_cache_update = NULL;
	  
	dev->hard_header_len = sizeof(struct slan_header);
	dev->addr_len = 0;
	return 0;
}

#ifdef MODULE

__initfunc(static int slan_probe(struct device *dev))
{
	slan_init(dev);
	return 0;
}

static char slan_name[16];

static struct device dev_slan = {
		slan_name, 	/* Needs to be writeable */
		0, 0, 0, 0,
	 	0x0, 0,
	 	0, 0, 0, NULL, slan_probe };

/* Module setup code: Stolen from other network drivers in linux/drivers/net */
int init_module(void)
{
	/* Find a name for this unit */
	int err=dev_alloc_name(&dev_slan,"slan%d");
	if(err<0)
		return err;
	if (register_netdev(&dev_slan) != 0)
		return -EIO;
	return 0;
}

void cleanup_module(void)
{
	unregister_netdev(&dev_slan);
	kfree(dev_slan.priv);
	dev_slan.priv = NULL;
}
#endif /* MODULE */

