/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/platform/linux/slan_roboconnect.c,v 1.1.1.1 1999/12/09 01:16:39 koni Exp $ 
  $Author: koni $ $Date: 1999/12/09 01:16:39 $ $Name: slan-release-0-04 $
  $Revision: 1.1.1.1 $

*/
   
/* The Lightlink Secure LAN project

   KONI - December 2000
   slan_client.c

   Linux client program
*/

/* Standard C library includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

/* to secure memory (lock from paging to disk) */
#include <sys/mman.h>

/* Network stuff */
#include <sys/socket.h>
#include <sys/poll.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <net/route.h>

/* GNU multiprecision math library */
#ifdef HAVE_GMP_H
#include <gmp.h>
#else
#include "slan_gmp3.h"
#endif

/* session state type definitions and constants */
#include "slan_session.h"
#include "slan_version.h"
#include "slan_config.h"

/* Cryptography */
#include "blowfish.h"
#include "sym_cipher.h"
#include "sha.h"
#include "mac.h"
#include "hmac_sha1.h"

/* Setup and Control Protocol components */
#include "dh_keyexchange.h"
#include "rsaauth.h"
#include "random_interface.h"
#include "client_auth.h"
#include "ip_conf.h"
#include "control_msg.h"
#include "control_tcp.h"
#include "client_pamauth.h"

/* Transport protocol stack */
#include "proto_stack.h"
#include "stack_register.h"

slan_conf_t *slan_global_config;

static int shutdown_now = 0;

static void client_setup_transport_layer(slan_session_t *session_obj,
					 char *server_address) {
  proto_stack_t *last, *current;
  int header_save;

  header_save = 0;
  /* A stack terminator at the top -- this is simply a dummy protocol
     layer that drops any packet passed to it. It can be used to catch
     bad or dropped packets and debug them. Otherwise, it isn't necessary and
     doesn't do much. */
  current = last = malloc(sizeof(proto_stack_t));
  terminate_register(current);
  header_save += current->header_reserve;
  
  /* Packets received from the slan device sent down the stack from the top */
  current = current->next = malloc(sizeof(proto_stack_t));
  slan_link_register(current, 0, session_obj);
  header_save += current->header_reserve;
  session_obj->transport_top = current;
  current->prev = last;
  last = current;

  /* Optional fragmenting protocol layer -- currently not implemented */
  current = current->next = malloc(sizeof(proto_stack_t));
  fragment_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* All packets being sent through the slan tunnel get tagged with sequence
     numbers before adding a MAC and encrypting. This used to limit the
     possibilities for message replay attacks. */
  current = current->next = malloc(sizeof(proto_stack_t));
  sequence_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* After adding the sequence number, we add a Message Authentication Code
     (MAC) to allow the receiver to verify the integrity of the packet and the 
     sender of the data */
  current = current->next = malloc(sizeof(proto_stack_t));
  message_sign_register(current, session_obj->mac);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* Lastly, we enrypt the packet for privacy during transmission. Currently,
     I am expecting the symmetric cipher used to be a configurable option, so
     the cipher used depends on the field in the session_obj */
  current = current->next = malloc(sizeof(proto_stack_t));
  symmetric_cipher_register(current, session_obj->cipher);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* Session ID numbers are prepended to identify the session the message
     belongs to */
  current = current->next = malloc(sizeof(proto_stack_t));
  session_id_register(current);
  header_save += current->header_reserve;
  *((u_int32_t *) current->data) = session_obj->id;
  current->prev = last;
  last = current;

  /* Transmit the secured packet via the UDP channel */
  current = current->next = malloc(sizeof(proto_stack_t));
  client_udplink_register(current, server_address, session_obj);
  header_save += current->header_reserve;
  session_obj->transport_bottom = current;
  current->prev = last;
  last = current;

  /* Terminate the bottom of the stack */
  current = current->next = malloc(sizeof(proto_stack_t));
  terminate_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  session_obj->header_save = header_save;
}

static void forget_keys(slan_session_t *session_obj) {
  int i;

  /* Forget keys. NULL pointers may exist if shutdown happens before the 
     session is configured fully */
  if (session_obj->mackey!=NULL) {
    for(i=0;i<session_obj->mackey_length;i++)
      session_obj->mackey[i] = 0x0;
  }

  /* Usally diffie-hellman returns one big long key which we set the mackey
     pointer to something halfway through and the cipher key to the memory
     returned from DH. So we only free the cipher key. */
  if (session_obj->key!=NULL) {
    for(i=0;i<session_obj->key_length;i++)
      session_obj->key[i] = 0x0;
    free(session_obj->key);
  }

}

static void client_shutdown(slan_session_t *session_obj) {
  
  fprintf(stdout,"Connection closing...\n"); fflush(stdout);

  session_obj->cipher->close(session_obj->cipher);
  session_obj->mac->close(session_obj->mac);
  forget_keys(session_obj);

  /* Transport layer destruction here */
  close(session_obj->control_fd);

  /* This should cause the process to exit */
  pthread_exit(NULL);
}

static void change_key(slan_session_t *session_obj) {

  forget_keys(session_obj);
  if (dh_initiate(session_obj)) {
    fprintf(stderr,"Change session key: DH exchange failed!\n");
    fprintf(stderr,"The key was not changed.\n");
    client_shutdown(session_obj);
  } else {
    fprintf(stdout,"Setting MAC and cipher keys....\n"); fflush(stdout);
    session_obj->mackey_length = session_obj->key_length/2;
    session_obj->key_length -= session_obj->mackey_length;
    session_obj->mackey = session_obj->key + session_obj->key_length;
    
    session_obj->cipher->set_key(session_obj->cipher, session_obj->key,
				 session_obj->key_length); 
    
    session_obj->mac->set_key(session_obj->mac, session_obj->mackey,
			      session_obj->mackey_length); 
    /* Reset sequence numbers: Maybe this should be randomized a bit,
       bleeding off some bits from the DH key exchange. */
    session_obj->sequence_down = 0;
    session_obj->sequence_up = 0;
    
    session_obj->key_timestamp = time(NULL);
    /* Unblock the transport layer. */
  }
  session_obj->key_change = 0;
}

static void interrupt_shutdown(int signal) {
  shutdown_now = 1;
}

static void client_unexpected_shutdown(slan_session_t *session_obj) {

  fprintf(stderr,"\n\nThe connection to the server has unexpectedly "
	  "closed.\n");
  fprintf(stderr,"The session will be terminated (sorry!).\n\n");

  client_shutdown(session_obj);
}


static void *control_loop(void *arg) {
  slan_session_t *session_obj;
  struct pollfd ufds;
  control_msg_t *msg, *new_msg;
  u_int32_t type, subtype;

  session_obj = arg;
  for(;;) {
    ufds.fd = session_obj->control_fd;
    ufds.events = ( POLLIN | POLLPRI | POLLERR | POLLHUP );
    ufds.revents = 0;

    /* Give up after 2 seconds, check to see if we should do anything
       and then go back to polling. */
    poll(&ufds, 1, 2000);
    pthread_mutex_lock(session_obj->lock);
    if (shutdown_now) {
      fprintf(stderr,"Interrupt Received, shutdown started.\n");
      client_shutdown(session_obj);
    }

    if (ufds.revents == POLLHUP) {
      /* If the server hangs up, we shutdown immediately. */
      client_unexpected_shutdown(session_obj);
    }
    if (ufds.revents == POLLPRI) {
      /* Our protocol doesn't use out of band data. So we just ignore it */
      fprintf(stderr,"Out of band data waiting... why is that?\n");
      continue;
    }
    if (ufds.revents == POLLERR) {
      /* I don't know what might cause this to occur, so for now, we'll
	 shutdown the connection if we have an error */ 
      fprintf(stderr,"Error condition on control channel.");
      client_shutdown(session_obj);
    }
    if (ufds.revents == POLLIN) {
      msg = control_msg_receive(session_obj, &type, &subtype, 0);
      if (msg == NULL) {
	pthread_mutex_unlock(session_obj->lock);
	continue;
      }

      switch(type) {
      case KEYCHANGE:
	if (subtype == KEYCHANGE_READY) {
	  change_key(session_obj);
	  control_msg_destruct(msg);
	} else {
	  session_obj->key_change = 1;
	  new_msg = new_control_msg(KEYCHANGE, KEYCHANGE_READY, session_obj);
	  control_msg_send(new_msg);
	  pthread_mutex_unlock(session_obj->lock);
	  control_msg_destruct(new_msg);
	  control_msg_destruct(msg);
	  usleep(1000);
	  pthread_mutex_lock(session_obj->lock);
	  change_key(session_obj);
	  
	}
	pthread_mutex_unlock(session_obj->lock);
	continue;
	break;
      case SHUTDOWN:
	/* This should not return */
	client_shutdown(session_obj);
	break;
      case KEEPALIVE:
	if (subtype != KEEPALIVE_PING) {
	  /* This should not happen. */
	  fprintf(stderr,"Not expecting a keepalive response here.\n");
	} else {
	  new_msg = new_control_msg(KEEPALIVE, KEEPALIVE_RESPONSE, 
				    session_obj);
	  control_msg_send(new_msg);
	  control_msg_destruct(new_msg);
	  fprintf(stderr,"Sent keepalive response.\n");
	}
	break;
      default:
	fprintf(stderr,"Unhandled control message type: %d\n",type);
	break;
      }
      control_msg_destruct(msg);
    }
    
    /* Check the session key's age. Change it if it should be expired. This is
       the client's responsibility to enforce. */
    if ((time(NULL) - session_obj->key_timestamp) > KEY_LIFETIME &&
	session_obj->key_change == 0) {
      session_obj->key_change = 1;
      new_msg = new_control_msg(KEYCHANGE, KEYCHANGE_REQUEST, session_obj);
      control_msg_send(new_msg);
      control_msg_destruct(new_msg);
    }
    pthread_mutex_unlock(session_obj->lock);
  }
}

/* This function continuouly reads either UDP socket for incoming traffic or
   the SLAN socket for outgoing traffic, using the builtin transport layer
   functions. They are non-blocking, using select to test for data before
   reading.  

   TODO: Make this a multithreaded operation and allow blocking calls?
   Performance at the client end seems to be fine with the current
   implementation. */
/* Incremental sleeping step size. The longer we have nothing to do the more
   we wait before looking for packets to process */
#define SLEEP_STEP (2)
#define SLEEP_MAX  (1000)
/* Maximum number of packets to process in one direction before switching to
   the other. Blocking will cause immediate switch. */
#define PACKETS_PER_ATTEMPT (10)
static void *transport_loop(void *arg) {
  slan_session_t *session_obj;
  unsigned char buf[10240];
  unsigned char *p;
  slan_message_t slan_message;
  int packet;

  session_obj = arg;

  slan_message.buffer = buf;
  slan_message.maxspace = 4096;

  p = buf + session_obj->header_save;
  
  slan_message.header = slan_message.data = p;
  slan_message.trailer = p;
  slan_message.length = 0;

  session_obj->down_blocking = session_obj->up_blocking = 0;
  for(;;) {
    /* This stuff will be later handled by slan_free/slan_alloc */

    /* Count the number of back to back packets we send in the main loop, as
       a measure of how busy we are to adapt sleeping time. */
    packet = 0;

    /* Action on the control channel (such as a change key event) may require
       that the transport layer halt. Idealing, only the outgoing channel
       should halt. This may/will need to be modified. */
    pthread_mutex_lock(session_obj->lock);
      
    /* This is an attempt at adaptive throttling for processing the network
       packets. If either direction doesn't have packets, to send, it will
       set the blocking flag. When both channels are blocking, we'll drop
       out of the loop and decide whether or not to sleep and for how
       long. 
       
       This is experimental. */ 
    while(!(session_obj->down_blocking && session_obj->up_blocking) &&
	  packet < PACKETS_PER_ATTEMPT) {
      if (!session_obj->key_change) {
	/* If a key change is in about to happen, process incoming packets
	   only */
	session_obj->transport_top->down(session_obj->transport_top,
					 &slan_message);
	slan_message.header = slan_message.data = p;
	slan_message.trailer = p;
	slan_message.length = 0;
      }

      session_obj->transport_bottom->up(session_obj->transport_bottom,
					&slan_message);
      slan_message.header = slan_message.data =p;
      slan_message.trailer = p;
      slan_message.length = 0;
      packet++;
    }
    pthread_mutex_unlock(session_obj->lock);
    if (packet < PACKETS_PER_ATTEMPT)
      usleep((PACKETS_PER_ATTEMPT - packet) * SLEEP_STEP);
    session_obj->up_blocking = session_obj->down_blocking = 0;
#if 0
   /* If both layers are blocking, back off and sleep, otherwise, if we've
       suddenly gotten busy, reset the sleeptime to zero. */
    if ((session_obj->up_blocking && session_obj->down_blocking) ||
	session_obj->transport_block) {
      if (sleeptime < SLEEP_MAX) {
	sleeptime *= SLEEP_STEP;
      } else {
	fprintf(stderr,"Max sleeptime reached\n");
      }
      usleep((unsigned long) sleeptime);
    } else {
      sleeptime = 1.0f;
    }
#endif
  }
}


/* Temporary */

/* Get ready to reach around your ass to grab the other elbow.... */
static int  linux_configure_network(ip_info_t *ip_conf, 
				    unsigned int server_route,
				    unsigned int server_addr)  { 
  int sock;
  struct ifreq ifr;
  struct sockaddr_in sa;
  struct rtentry rt;
  struct in_addr ns;
  FILE *f;

  sock = socket(AF_INET, SOCK_DGRAM, 0);

  strcpy(ifr.ifr_name,"slan0");

  sa.sin_family = AF_INET;
  memcpy(&sa.sin_addr, &ip_conf->ip_address, 4);
  memcpy(&ifr.ifr_addr, &sa, sizeof(struct sockaddr));
  if (ioctl(sock, SIOCSIFADDR, &ifr) < 0) {
    fprintf(stderr,"Could not set ipaddress for interface.\n");
    perror("Error code:");
    return 1;
  } 

  memcpy(&sa.sin_addr, &ip_conf->subnet_mask, 4);
  memcpy(&ifr.ifr_netmask, &sa, sizeof(struct sockaddr));

  if (ioctl(sock, SIOCSIFNETMASK, &ifr) < 0) {
    fprintf(stderr,"Could not set subnet mask for interface.\n");
    perror("Error code:");
    return 1;
  } 

  memcpy(&sa.sin_addr, &ip_conf->subnet_bcast, 4);
  memcpy(&ifr.ifr_broadaddr, &sa, sizeof(struct sockaddr));
  if (ioctl(sock, SIOCSIFBRDADDR, &ifr) < 0) {
    fprintf(stderr,"Could not set broadcast for interface.\n");
    perror("Error code:");
    return 1;
  } 

  /* Start the interface going like a normal ethernet except we don't do arp */
  ifr.ifr_flags = IFF_UP | IFF_RUNNING | IFF_NOARP | IFF_BROADCAST;
  if (ioctl(sock, SIOCSIFFLAGS, &ifr) < 0) {
    fprintf(stderr,"Failed setting interface flags.\n");
    perror("Error code:");
    return 1;
  } 
  
  /* Now do routing */
  
  /* Clear the route entry structure */
  memset(&rt, 0x0, sizeof(struct rtentry));
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = htonl(INADDR_ANY);
  
  memcpy(&rt.rt_dst, &sa, sizeof(struct sockaddr_in));
  memcpy(&rt.rt_genmask, &sa, sizeof(struct sockaddr_in));
  
  rt.rt_flags = RTF_UP | RTF_GATEWAY;
  while(ioctl(sock, SIOCDELRT, &rt) == 0) {
    fprintf(stdout,"Deleted default route from routing table.\n");
  }

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = server_addr;
  memcpy(&rt.rt_dst, &sa, sizeof(struct sockaddr_in));

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = 0xFFFFFFFF;
  memcpy(&rt.rt_genmask, &sa, sizeof(struct sockaddr_in));

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = server_route;
  memcpy(&rt.rt_gateway, &sa, sizeof(struct sockaddr_in));

  rt.rt_flags = RTF_HOST | RTF_GATEWAY | RTF_UP;
  if (ioctl(sock, SIOCADDRT, &rt) < 0) {
    perror("Could not set routing information for slan server route.\n");
    return 1;
  }

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = 0x0;
  memcpy(&rt.rt_dst, &sa, sizeof(struct sockaddr_in));

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = 0x0;
  memcpy(&rt.rt_genmask, &sa, sizeof(struct sockaddr_in));

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = ip_conf->router;
  memcpy(&rt.rt_gateway, &sa, sizeof(struct sockaddr_in));

  rt.rt_flags = RTF_GATEWAY | RTF_UP;
  if (ioctl(sock, SIOCADDRT, &rt) < 0) {
    perror("Could set default gateway in routing table.\n");
    return 1;
  }
  close(sock);
  
  /* Set nameserver configuration */
  f = fopen("/etc/resolv.conf","w");
  if (f == NULL) return 1;

  ns.s_addr = ip_conf->nameserver_1;
  fprintf(f,"nameserver %s\n",inet_ntoa(ns));
  if (ip_conf->nameserver_2) {
    ns.s_addr = ip_conf->nameserver_2;
    fprintf(f,"nameserver %s\n",inet_ntoa(ns));
  }
  fclose(f);

  return 0;
}

static void client_restore_routes(unsigned int server_route,
				  unsigned int server_addr) {
  int sock;
  struct sockaddr_in sa;
  struct rtentry rt;

  sock = socket(AF_INET, SOCK_DGRAM, 0);

  if (sock < 0) {
    perror("Cannot open socket for routing table configuration: ");
    return;
  }

  /* Clear the route entry structure */
  memset(&rt, 0x0, sizeof(struct rtentry));
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = htonl(INADDR_ANY);
  
  memcpy(&rt.rt_dst, &sa, sizeof(struct sockaddr_in));
  memcpy(&rt.rt_genmask, &sa, sizeof(struct sockaddr_in));
  
  rt.rt_flags = RTF_UP | RTF_GATEWAY;
  if (ioctl(sock, SIOCDELRT, &rt)<0){
    fprintf(stderr,"Failed deleting default route\n");
  }

  sa.sin_addr.s_addr = server_route;
  memcpy(&rt.rt_gateway, &sa, sizeof(struct sockaddr_in));
  if (ioctl(sock, SIOCADDRT, &rt)<0) {
    fprintf(stderr,"Failed to restore default routing.\n");
  }

  sa.sin_addr.s_addr = server_addr;
  memcpy(&rt.rt_dst, &sa, sizeof(struct sockaddr_in));

  sa.sin_addr.s_addr = 0xFFFFFFFF;
  memcpy(&rt.rt_genmask, &sa, sizeof(struct sockaddr_in));

  rt.rt_flags |= RTF_HOST;
  if (ioctl(sock, SIOCDELRT, &rt)<0) {
    fprintf(stderr,"Failed to delete route to slan server.\n");
  }

  /* Shut down the slan0 interface here? */

  close(sock);
  fprintf(stdout,"Slan routing deleted, normal routing restored.\n");
  return;
}

static char *shut_up_compiler(u_int32_t addr) {
  struct in_addr x;
  x.s_addr = addr;

  return inet_ntoa(x);
}

/* This function inspects the system's routing table (as per /proc/net/route)
   and finds the default route. If there are more than one it will
   (incorrectly most likely) use the last one in the table. */
static unsigned int route_table_get_default() {
  unsigned int server_target_route =0x0;
  unsigned char inputline[256];
  FILE *rtable;

  rtable = fopen("/proc/net/route","r");
  if (rtable == NULL) perror("Could not open /proc/net/route.\n");
  while(getc(rtable)!='\n');
  while(!feof(rtable)) {
    unsigned int target, gate, flags;
    fgets(inputline,256,rtable);
    sscanf(inputline,"%*s %x %x %x",&target,&gate,&flags);
    if (target == 0x0 && (flags & 0x1)) {
      server_target_route = gate;
    }
  }

  fclose(rtable);
  return server_target_route;
}


/* This is the entry point for the client side user space daemon program 

   TO BE COMMENTED MORE LATER */
int main(int argc, char *argv[]) {
  slan_session_t *session_obj;
  unsigned char null_key[32];
  ip_info_t *ip_conf;
  unsigned int server_route, server_addr;
  pthread_t transport_thread, control_thread;

  if (argc!=2) {
    fprintf(stderr,"%s: usage\n",argv[0]);
    fprintf(stderr,"%s <server address>\n",argv[0]);
    exit(1);
  }

  /* Lock ALL memory -- This process is important to functioning of the
     network for the client, and should never be swapped out. As well (and
     more importantly), the memory used by this process holds vital security
     sensitive information, such as cryptographic keys, and must never be
     written to permanent store. */
#ifdef linux
  if (mlockall(MCL_FUTURE)) {
    perror("Failed locking memory");
    fprintf(stderr,"This process must be run as root.\n");
    exit(-1);
  }
#endif
#ifdef WIN32
  /* Lucas: 

     If there is the equivalent of mlockall() in windows, we should probably
     just use that and forget it. Otherwise, we'll have to lock the pages of
     memory used for the transport layer, cipher object, and session object,
     as well as some other memory probably.
     
     Probably what we'll want to do in that case is write our own
     malloc/calloc/free calls that allocates and locks a page at a time as we
     need, handing out pointers to parts of the locked page as we can fit the
     individual requests in. */
#endif


  /* Before doing anything, gather entropy and setup our random number
     generator. This will be used to provide security, so it better be good. */
  random_initialize();

  /* Allocate and setup a session object which will hold some state
     information for the active connection. 

     TO BE DONE:
     (1) The server address is to be determined from the service set
     information, when that functionality is added */
  session_obj = calloc(1, sizeof(slan_session_t));

  session_obj->control_fd = client_open_socket(argv[1],&server_addr);
  session_obj->lock = calloc(1, sizeof(pthread_mutex_t));
  pthread_mutex_init(session_obj->lock, NULL);

  server_route = route_table_get_default();
  if (server_route == 0x0) {
    fprintf(stderr,"Error: Cannot find a default route\n");
    exit(1);
  }

  /* Establish whether or not the server is compatible with the client. In the 
     future this may instruct the server (or client) to do fancy things like
     enable or disable certain features or bug fixes. For now, it really
     doesn't do anything. */
  fprintf(stdout,"Checking version compatibility..."); fflush(stdout);
  if (check_version(session_obj->control_fd)<0) {
    fprintf(stderr,"Error, version of server and client incompatible.\n");
    close(session_obj->control_fd);
    exit(-1);
  } else {
    fprintf(stderr,"Server Version okay.\n");
  }

  fprintf(stdout,"Initializing to default key values...\n"); fflush(stdout);
  /* The DH key exchange will use control message functionality which expects
     to have these things set up and working, except we don't know the key
     until after DH completes. So we'll set them up with null keys (0x0) so
     they function, but they will not provide any security ofcourse. */
  memset(null_key, 0x0, 32);
  session_obj->mac = get_hmac_sha1();
  session_obj->cipher = get_blowfish();
  session_obj->mac->set_key(session_obj->mac, null_key, 32);
  session_obj->cipher->set_key(session_obj->cipher, null_key, 32);
  session_obj->key_timestamp = time(NULL);

  /* Establish the session key via Diffie-Hellman key exchange. The value
     computed by this exchange will be used to produce a symmetric key for
     data privacy and a Message Authentication Code (MAC) key for protecting
     the messages against alterations. We also compute a one-way hash of the
     key value computed and use that to defeat a Diffie-Hellman
     man-in-the-middle attack by using it in the RSA authentication challenge
     which authenticates the server. */
  fprintf(stdout,"Establishing session key with server (Diffie-Hellman)...\n");
  fflush(stdout);
  if (dh_initiate(session_obj)) {
    fprintf(stderr,"Cannot continue, Diffie-Hellman key exchange failed.\n");
    exit(1);
  }

  fprintf(stdout,"Diffie-Hellman Completed. Setting session keys....\n"); 
  fflush(stdout);

  /* Setup the symmetric cipher. 

     TODO: Set this up so that the control channel enrypts and verifies all
     further communications 

     Also, the symmetric cipher actually used may depend on the service set
     preferences set by the user or service provider, or by the version
     information extracted above from the server.  */
  session_obj->cipher->set_key(session_obj->cipher, session_obj->key,
			       session_obj->key_length); 

  /* Setup the message authentication code */
  session_obj->mac->set_key(session_obj->mac, session_obj->mackey,
			    session_obj->mackey_length); 
  
  fprintf(stdout,"Symmetric Encryption key: %d bits.\n",
	  session_obj->key_length * 8);
  fprintf(stdout,"Message Authentication Code key: %d bits.\n",
	  session_obj->mackey_length * 8);
  fprintf(stdout,"Finished symmetric session key configuration, link is "
	  "secure.\n\n");
  fflush(stdout);

  /* Authenticate the server. This is a required part of the protocol for any
     service provider. The public key should be given in the service set
     definitions loaded above (some day) */
  fprintf(stdout,"Establishing authenticity of server (RSA)...");
  fflush(stdout);
  if (rsaauth_initiate(session_obj)) { 
    fprintf(stderr,"Server identity cannot be verified: Cannot continue.\n");
    exit(-1);
  }
  fprintf(stdout, "Server identity verified.\n\n");

  fprintf(stdout, "Client (user) authentication...\n");
  session_obj->username = "koni";
  session_obj->password = "OpG*I[&dc";
  if (client_clientauth_wait(session_obj)) {
    fprintf(stderr,"Client authentication to server failed.\n");
    exit(-1);
  }
  fprintf(stdout,"Client (you) accepted.\n\n"); fflush(stdout);

  /* !!!! TEMPORARY CODE !!!!
     
     Let the server dictate to us the IP address it would like to use for
     us. */ 
  fprintf(stdout,"Asking for network configuration parameters...\n");
  fflush(stdout);
  ip_conf = (ip_info_t *) calloc(1, sizeof(ip_info_t));
  if (ipconf_request(ip_conf, session_obj)) {
    fprintf(stdout,"Failed to get ip configuration information from "
	    "server.\n");
    free(ip_conf);
    exit(-1);
  }


  fprintf(stdout,"Ip address is %s\n",shut_up_compiler(ip_conf->ip_address));
  fprintf(stdout,"Netmask is %s\n",shut_up_compiler(ip_conf->subnet_mask));
  fprintf(stdout,"Broadcast address is %s\n",shut_up_compiler(ip_conf->subnet_bcast));
  fprintf(stdout,"Gateway address is %s\n",shut_up_compiler(ip_conf->router));
  fprintf(stdout,"Nameserver #1 is %s\n",shut_up_compiler(ip_conf->nameserver_1));
  fprintf(stdout,"Nameserver #2 is %s\n\n",shut_up_compiler(ip_conf->nameserver_2)); 
  fflush(stdout);

#if 0
  fprintf(stdout,"Attempting to configure local network settings "
	  "automatically..."); fflush(stdout);
  if (linux_configure_network(ip_conf, server_route, server_addr)) {
    fprintf(stdout,"Error, network configuration failed. Cannot continue.\n");
    exit(-1);
  }
  fprintf(stdout,"Done.\n\n"); fflush(stdout);
#else
  fprintf(stdout,"Robo-client: Skipping ip configuration.\n");
#endif
  free(ip_conf);

  /* End of temporary code -- the above will be replaced by something more
     concrete soon. */
  fprintf(stdout,"Session startup okay, SLAN secured network link running.\n");

  client_setup_transport_layer(session_obj,argv[1]);

#if 0  
  /* Loop forever until a shutdown event comes or the server stops
     responding. */
  pthread_create(&transport_thread, NULL, transport_loop, session_obj);
  
  pthread_create(&control_thread, NULL, control_loop, session_obj);

  signal(SIGINT, interrupt_shutdown);

  /* When the control thread exits, we are done. Other things may terminate
     the process entirely however.  */
  pthread_join(control_thread, NULL);

  client_restore_routes(server_route, server_addr);
#endif
  usleep(1000);
  close(session_obj->control_fd);

  fprintf(stdout,"Slan session finished.\n"); fflush(stdout);
  return 0;
}


