/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/platform/linux/slan_server.c,v 1.4 2002/04/02 17:09:39 koni Exp $ 
  $Author: koni $ $Date: 2002/04/02 17:09:39 $ $Name: slan-release-0-04 $
  $Revision: 1.4 $

*/
   
/* The Lightlink Secure LAN project

   KONI - December 2000
   slan_server.c
   
   Highest level file/routines for the multiple client slan server.
*/   

/* Standard C library includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <errno.h>

/* Network Includes */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/poll.h>

/* To lock all memory pages in physical memory */
#include <sys/mman.h>

/* Packet Interface to slan device driver includes */
#include <linux/if.h>
#include <linux/if_tun.h>
#include <fcntl.h>



/* Output messages to the system log facility */
#include <syslog.h>

#include <sys/ioctl.h>

/* GNU multiprecision math library. For public key math. */
#ifdef HAVE_GMP_H
#include <gmp.h>
#else
#include "slan_gmp3.h"
#endif

/* slan state information and constant/type definitions */
#include "slan_version.h"
#include "slan_session.h"
#include "slan_config.h"

/* Cryptography */
#include "blowfish.h"
#include "sym_cipher.h"
#include "sha.h"
#include "mac.h"
#include "hmac_sha1.h"

/* Setup and Control Protocol components */
#include "dh_keyexchange.h"
#include "rsaauth.h"
#include "random_interface.h"
#include "control_tcp.h"
#include "client_pamauth.h"
#include "mac_conf.h"

/* Transport protocol stack */
#include "proto_stack.h"
#include "stack_register.h"
#include "log_message.h"

slan_conf_t *slan_global_config;

/* Mutex locks for multithreaded server. Add more locks by adding a new enum
   constant before TLOCK_NMUTEXES */
enum { TLOCK_SESSION_TABLE = 0, TLOCK_IPADDR_MAP, TLOCK_MACADDR_MAP,
       TLOCK_CONSOLE,
       TLOCK_UNIQUE_ID, TLOCK_IPASSIGN, TLOCK_CONTROL_HANDLE_LIST,
       TLOCK_NMUTEXES };
static pthread_mutex_t mutex_locks[TLOCK_NMUTEXES];

/* Socket handles for the slan device packet interface and the transport layer 
   udp port. Clients all share the same UDP port on the server, at the
   moment. */
static int slan_fd, udp_fd;

/* This is temporarily used as a key (not for security) until the DH
   keyexchange has given us our initial keys. */
static unsigned char null_key[32];

/* Active client connections are stored in a hash table (using linked list
   collision chaining) of slan_session_t objects. */
/* Please make this a power of 2 less one for hash function efficiency */
#define SESSION_TABLE_MASK 0xFFF
#define SESSION_HASH_FUNC(x) ((x) & SESSION_TABLE_MASK)
static slan_session_t *session_table[SESSION_TABLE_MASK + 1]; 

/* This function finds the slan_session structure associated with a current
   client, given the client's id. If the session structure is not found, NULL
   is returned. The session object is returned with the lock held. If the lock
   cannot be attained, NULL is returned, even if the session obj exists. If
   retry is > 0, we busy wait for the lock and try it "retry" times in a
   row. If sleep is nonzero, we sleep for that many microseconds between
   tries. If retry is -1, we block waiting for the lock, and wait for it
   forever. */
slan_session_t *session_lookup(u_int32_t id, u_int32_t retry, u_int32_t sleep) {
  slan_session_t *session_obj;

  pthread_mutex_lock(mutex_locks + TLOCK_SESSION_TABLE);

  session_obj = session_table[SESSION_HASH_FUNC(id)];
  while(session_obj!=NULL && session_obj->id != id)
    session_obj = session_obj->next;

  if (session_obj != NULL) {
    if (pthread_mutex_trylock(session_obj->lock)) {
      pthread_mutex_unlock(mutex_locks + TLOCK_SESSION_TABLE);
      if (retry == 0) {
	logmsg(MSG_WARNING,"session id %d: retry exceeded. Cannot obtain "
	       "lock.\n", id);
	return NULL;
      }
      usleep(sleep);
      return session_lookup(id, retry-1, sleep);
    }
  }

  pthread_mutex_unlock(mutex_locks + TLOCK_SESSION_TABLE);
  return session_obj;
}

/* This function clears any values from the session table. It should be called 
   once at server startup, to make sure all entries in the table start as NULL 
   pointers. As connections arrive, the connections are added to the table
   according to their session id. When a session closes, the entry should be
   removed. */
static void session_table_init(void) {
  int index;
  
  pthread_mutex_lock(mutex_locks + TLOCK_SESSION_TABLE);
  for(index=0;index<=SESSION_TABLE_MASK;index++)
    session_table[index] = NULL;
  pthread_mutex_unlock(mutex_locks + TLOCK_SESSION_TABLE);
}

/* This function addes a slan_session object to the table of active
   sessions. Then, the session object can be retrieved efficiently, when
   needed. */
static void session_table_add(slan_session_t *new_obj) {
  int table_index;
  slan_session_t *session_obj;

  table_index = SESSION_HASH_FUNC(new_obj->id);

  /* Any table operation requires locking the table to prevent another
     simultaneous modification */
  pthread_mutex_lock(mutex_locks + TLOCK_SESSION_TABLE);
  if (session_table[table_index] == NULL) {
    session_table[table_index] = new_obj;
  } else {
    session_obj = session_table[table_index];
    while (session_obj->next != NULL) 
      session_obj = session_obj->next;
    session_obj->next = new_obj;
  }
  new_obj->next = NULL;

  pthread_mutex_unlock(mutex_locks + TLOCK_SESSION_TABLE);
}

/* When an slan session terminates, we delete the session object from the
   table and forget its state information */
static void session_table_delete(u_int32_t id) {
  int table_index;
  slan_session_t *session_obj, *last;

  table_index = SESSION_HASH_FUNC(id);

  pthread_mutex_lock(mutex_locks + TLOCK_SESSION_TABLE);
  if (session_table[table_index]->id == id) {
    session_obj = session_table[table_index];
    session_table[table_index] = session_obj->next;
  } else {
    last = session_obj = session_table[table_index];
    session_obj = session_obj->next;
    while(session_obj != NULL && session_obj->id != id) {
      last = session_obj;
      session_obj = session_obj->next;
    }
    /* ACK! Trying to delete a session that doesn't exist? */
    if (session_obj == NULL) {
      pthread_mutex_unlock(mutex_locks + TLOCK_SESSION_TABLE);
      return;
    }
    last->next = session_obj->next;
  }
  pthread_mutex_unlock(mutex_locks + TLOCK_SESSION_TABLE);
}

/* For now, and for the future -- unless practice determines that this 
   is not a good enough method -- we're going to perform the same
   trick here as in the session_table, to map IP addresses into
   session ids and session_objs.

   We'll find the destination address (should be the slan address of
   the client) in the packet received from the slan device, as a 4
   byte integer (u_int32_t). Then we'll hash on that (with a really
   simple but faster than hell hash function) and scan the linked list 
   found to find the client's connection id. From there we can
   proceed directing the packet. */
typedef struct ipaddr_map_struct {
  u_int32_t id;
  u_int32_t ipaddr;
  struct ipaddr_map_struct *next;
} ipaddr_map_t;

/* Please make this a power of 2 less one for hash function efficiency. 
   

   TODO: Make sure the endianness crap is handled correctly -- I'd expect the
   leading part (in network byte order) of the ip address to be the same or
   very similar. Thus, we want the hash function to mask off the least
   significant portion of the ipaddress. */
#define IPADDR_TABLE_MASK 0xFFF
#define IPADDR_HASH_FUNC(x) ((x) & IPADDR_TABLE_MASK)
static ipaddr_map_t *ipaddr_table[IPADDR_TABLE_MASK + 1];


/* This function returns the session id of a client given the destination 
   ipaddress of a network packet */ 
static u_int32_t ipaddr_lookup(u_int32_t ip) {
  ipaddr_map_t *ipaddr_map;
  u_int32_t return_val;

  pthread_mutex_lock(mutex_locks + TLOCK_IPADDR_MAP);
  ipaddr_map = ipaddr_table[IPADDR_HASH_FUNC(ip)];
  while(ipaddr_map != NULL && ipaddr_map->ipaddr != ip)
    ipaddr_map = ipaddr_map->next;

  if (ipaddr_map == NULL) {
    return_val = 0;
  } else {
    return_val = ipaddr_map->id;
  }
  pthread_mutex_unlock(mutex_locks + TLOCK_IPADDR_MAP);
  return return_val;
}

/* Clear the ipaddr_table to all NULL entries. This function is to be called
   once, when the server starts up */
static void ipaddr_table_init(void) {
  int index;
  
  pthread_mutex_lock(mutex_locks + TLOCK_IPADDR_MAP);
  for(index=0;index<=IPADDR_TABLE_MASK;index++)
    ipaddr_table[index] = NULL;
  pthread_mutex_unlock(mutex_locks + TLOCK_IPADDR_MAP);
}

/* This function adds a ipaddr map entry for a new session object, when a
   client session is started */
static void ipaddr_table_add(u_int32_t session_id, u_int32_t ipaddr) {
  int table_index;
  ipaddr_map_t *ipaddr_map;
  
  table_index = IPADDR_HASH_FUNC(ipaddr);
#ifdef DEBUG
  fprintf(stderr,"IPTABLE ADD %0X %0X\n",ipaddr,table_index);
#endif

  pthread_mutex_lock(mutex_locks + TLOCK_IPADDR_MAP);
  if (ipaddr_table[table_index] == NULL) {
    ipaddr_map = ipaddr_table[table_index] = malloc(sizeof(ipaddr_map_t)); 
  } else {
    ipaddr_map = ipaddr_table[table_index];
    while(ipaddr_map->next != NULL) ipaddr_map = ipaddr_map->next;
    ipaddr_map->next = malloc(sizeof(ipaddr_map_t));
    ipaddr_map = ipaddr_map->next;
  }

  ipaddr_map->next = NULL;
  ipaddr_map->ipaddr = ipaddr;
  ipaddr_map->id = session_id;
  pthread_mutex_unlock(mutex_locks + TLOCK_IPADDR_MAP);
}

/* When client session closes, we delete the ipaddr map entry */
static void ipaddr_table_delete(u_int32_t ipaddr) {
  int table_index;
  ipaddr_map_t *ipaddr_map, *last;

  table_index = IPADDR_HASH_FUNC(ipaddr);
  
  pthread_mutex_lock(mutex_locks + TLOCK_IPADDR_MAP);
  if (ipaddr_table[table_index] == NULL) {
    pthread_mutex_unlock(mutex_locks + TLOCK_IPADDR_MAP);
    return;
  }
  
  if (ipaddr_table[table_index]->ipaddr == ipaddr) {
    ipaddr_map = ipaddr_table[table_index];
    ipaddr_table[table_index] = ipaddr_map->next;
  } else {
    last = (ipaddr_map = ipaddr_table[table_index]);
    ipaddr_map = ipaddr_map->next;
    while(ipaddr_map != NULL && ipaddr_map->ipaddr != ipaddr) {
      last = ipaddr_map;
      ipaddr_map = ipaddr_map->next;
    }

    if (ipaddr_map == NULL) {
      pthread_mutex_unlock(mutex_locks + TLOCK_IPADDR_MAP);
      return;
    }
    last->next = ipaddr_map->next;
  }

  free(ipaddr_map);
  pthread_mutex_unlock(mutex_locks + TLOCK_IPADDR_MAP);
}


#define MACADDR_TABLE_MASK (0xFFF)
#define MACADDR_HASH_FUNC(x) ((((x)[4] & 0x0F) << 8) + (x)[5])
typedef struct macaddr_map_struct {
  u_int32_t id;
  unsigned char macaddr[6];
  struct macaddr_map_struct *next;
} macaddr_map_t;
static macaddr_map_t *macaddr_table[MACADDR_TABLE_MASK + 1];

/* This function returns the session id of a client given the destination 
   macaddress of a network packet. This function is not static because the
   mac_conf.c file will call it to check for mac address collisons with other
   SLAN clients */ 
static void server_unexpected_shutdown(slan_session_t *session_obj);
u_int32_t macaddr_lookup(unsigned char *macaddr) {
  macaddr_map_t *macaddr_map;
  u_int32_t return_val;

  pthread_mutex_lock(mutex_locks + TLOCK_MACADDR_MAP);
  macaddr_map = macaddr_table[MACADDR_HASH_FUNC(macaddr)];
  while(macaddr_map != NULL && memcmp(macaddr_map->macaddr, macaddr,6))
    macaddr_map = macaddr_map->next;

  if (macaddr_map == NULL) {
    return_val = 0;
    pthread_mutex_unlock(mutex_locks + TLOCK_MACADDR_MAP);
  } else {
    time_t last_response;
    slan_session_t *session_obj;

    return_val = macaddr_map->id;
    pthread_mutex_unlock(mutex_locks + TLOCK_MACADDR_MAP);

    /* Quick Hack to deal with hanging sessions */
    session_obj = session_lookup(return_val, 10, 100);
    if (session_obj == NULL) {
      return_val = 0;
    } else {
      last_response = session_obj->last_control_keepalive;
      if (time(NULL) - last_response > 120) {
	

	server_unexpected_shutdown(session_obj);
	return_val = 0;
      } else {
	pthread_mutex_unlock(session_obj->lock);
     } 
    }
  }

  return return_val;
}

/* Clear the macaddr_table to all NULL entries. This function is to be called
   once, when the server starts up */
static void macaddr_table_init(void) {
  int index;
  
  pthread_mutex_lock(mutex_locks + TLOCK_MACADDR_MAP);
  for(index=0;index<=MACADDR_TABLE_MASK;index++)
    macaddr_table[index] = NULL;
  pthread_mutex_unlock(mutex_locks + TLOCK_MACADDR_MAP);
}

/* This function adds a macaddr map entry for a new session object, when a
   client session is started */
static void macaddr_table_add(u_int32_t session_id, unsigned char *macaddr) {
  int table_index;
  macaddr_map_t *macaddr_map;
  
  table_index = MACADDR_HASH_FUNC(macaddr);
#ifdef DEBUG
  fprintf(stderr,"MACTABLE ADD %02X:%02X:%02X:%02X:%02X:%02X %0X\n",
	  (unsigned int) macaddr[0],(unsigned int) macaddr[1],
	  (unsigned int) macaddr[2],(unsigned int) macaddr[3],
	  (unsigned int) macaddr[4],(unsigned int) macaddr[5],
	  table_index);
#endif

  pthread_mutex_lock(mutex_locks + TLOCK_MACADDR_MAP);
  if (macaddr_table[table_index] == NULL) {
    macaddr_map = macaddr_table[table_index] = malloc(sizeof(macaddr_map_t)); 
  } else {
    macaddr_map = macaddr_table[table_index];
    while(macaddr_map->next != NULL) macaddr_map = macaddr_map->next;
    macaddr_map->next = malloc(sizeof(macaddr_map_t));
    macaddr_map = macaddr_map->next;
  }

  macaddr_map->next = NULL;
  memcpy(macaddr_map->macaddr, macaddr, 6);
  macaddr_map->id = session_id;
  pthread_mutex_unlock(mutex_locks + TLOCK_MACADDR_MAP);
}

/* When client session closes, we delete the macaddr map entry */
static void macaddr_table_delete(unsigned char *macaddr) {
  int table_index;
  macaddr_map_t *macaddr_map, *last;

  table_index = MACADDR_HASH_FUNC(macaddr);
  
  pthread_mutex_lock(mutex_locks + TLOCK_MACADDR_MAP);
  if (macaddr_table[table_index] == NULL) {
    pthread_mutex_unlock(mutex_locks + TLOCK_MACADDR_MAP);
    return;
  }
  
  if (memcmp(macaddr_table[table_index]->macaddr, macaddr, 6) == 0) {
    macaddr_map = macaddr_table[table_index];
    macaddr_table[table_index] = macaddr_map->next;
  } else {
    last = (macaddr_map = macaddr_table[table_index]);
    macaddr_map = macaddr_map->next;
    while(macaddr_map != NULL && memcmp(macaddr_map->macaddr, macaddr, 6)) {
      last = macaddr_map;
      macaddr_map = macaddr_map->next;
    }

    if (macaddr_map == NULL) {
      pthread_mutex_unlock(mutex_locks + TLOCK_MACADDR_MAP);
      return;
    }
    last->next = macaddr_map->next;
  }

  free(macaddr_map);
  pthread_mutex_unlock(mutex_locks + TLOCK_MACADDR_MAP);
}


/* ------------------------------------------------------------------------
   This section of code is for maintaining a list of control file 
   descriptors (TCP handles) for active sessions. It is used mainly to 
   build a fd_set for select(), so that the control handler thread can
   wait on control events for any active sessions easily, without having
   to scan the session_table to learn which sessions are active and what
   their control fd's are.
   ------------------------------------------------------------------------ */

typedef struct control_handle_struct {
  int control_fd;
  u_int32_t session_id;
  struct control_handle_struct *next;
} control_handle_t;

static control_handle_t *control_handle_list = NULL;

static void control_handle_add(int fd, u_int32_t session_id) {
  control_handle_t *new_control_handle;
  
  pthread_mutex_lock(mutex_locks + TLOCK_CONTROL_HANDLE_LIST);
  new_control_handle = calloc(1, sizeof(control_handle_t));
  
  new_control_handle->control_fd = fd;
  new_control_handle->session_id = session_id;

  new_control_handle->next = control_handle_list;
  control_handle_list = new_control_handle;
  pthread_mutex_unlock(mutex_locks + TLOCK_CONTROL_HANDLE_LIST);
}

static void control_handle_delete(u_int32_t session_id) {
  control_handle_t *handle, *last;

  pthread_mutex_lock(mutex_locks + TLOCK_CONTROL_HANDLE_LIST);
  last = NULL;
  handle = control_handle_list;
  while(handle!=NULL && handle->session_id != session_id) {
    last = handle; 
    handle = handle->next;
  }

  if (handle != NULL) {
    if (last == NULL) control_handle_list = handle->next;
    else {
      last->next = handle->next;
    }
    free(handle);
  }
  pthread_mutex_unlock(mutex_locks + TLOCK_CONTROL_HANDLE_LIST);  
}


static int control_handle_build_fdset(fd_set *fds) {
  control_handle_t *handle;
  int max_fd;
  
  FD_ZERO(fds);
  max_fd = 0;

  pthread_mutex_lock(mutex_locks + TLOCK_CONTROL_HANDLE_LIST);
  handle = control_handle_list;
  while(handle!=NULL) {
    FD_SET(handle->control_fd, fds);
    if (handle->control_fd > max_fd) max_fd = handle->control_fd;
    handle=handle->next;
  }
  pthread_mutex_unlock(mutex_locks + TLOCK_CONTROL_HANDLE_LIST);
  return max_fd;
}

static control_handle_t *control_handle_search_fdset(fd_set *fds) {
  control_handle_t *handle;

  pthread_mutex_lock(mutex_locks + TLOCK_CONTROL_HANDLE_LIST);
  handle = control_handle_list;
  while(handle!=NULL) {
    if (FD_ISSET(handle->control_fd, fds)) {
      FD_CLR(handle->control_fd, fds);
      pthread_mutex_unlock(mutex_locks + TLOCK_CONTROL_HANDLE_LIST);
      return handle;
    }
    handle=handle->next;
  }
  pthread_mutex_unlock(mutex_locks + TLOCK_CONTROL_HANDLE_LIST);
  return NULL;
}

/* -----------------------------------------------------------------------
   End of section for control handle list management 
*/

/* When a session is starting up, we need to assign an indentification number
   to the client, so we can direct its incoming encrypted packets
   appropriately to the correct transport layer stack. */
static u_int32_t server_get_unique_id(void) {
  static u_int32_t session_ids = 1;
  u_int32_t return_value;

  pthread_mutex_lock(mutex_locks + TLOCK_UNIQUE_ID);
  return_value = session_ids++;
  pthread_mutex_unlock(mutex_locks + TLOCK_UNIQUE_ID);
  
  return return_value;
}

/* This function looks at an slan_message_t as if it were an IP packet,
   looking for the target ip address 16 bytes into the message, as per a
   normal IP header.

   TODO: This seems to fail on DHCP broadcast packets. What up? */

/* This is for detecting ethernet headers for IP packets, which are added by
   DHCPOFFER packets when they come to us and screw up our detection of IP
   address. THIS IS TEMPORARY -- A HACK TO SEE IF DHCP can work through this.
   Since the hardware of slan devices are all the same, we can predict the
   header used by the DHCP server -- it will be the same MAC twice (server's
   slan device, and the client's slan device address -- the same), and then
   0x0008 for the IP protocol number. 

   THIS NEEDS TO BE FIXED -- maybe the server should set the MAC address on
   the client! */

/* This function simply pulls the first 4 bytes from the slan message received 
   at the udp socket. This should be the session id number of the client. This
   will allow us to lookup the session_obj for the client, and direct the
   message through that client's transport layer to decrypt and verify the
   message and send it out to the world  */
static u_int32_t slan_parse_session_id(slan_message_t *message) {
  u_int32_t session_id;

  session_id = *((u_int32_t *) message->header);
  /* Discard this information from the slan_message */
  slan_pop(message, sizeof(u_int32_t));

  return session_id;
}

static void print_mac_address(uchar *s, uchar *addr) {

  sprintf(s,"%02X:%02X:%02X:%02X:%02X:%02X",addr[0],addr[1],addr[2],
	  addr[3],addr[4],addr[5]);
}

static void log_incoming_rogue_packet(slan_message_t *message) {
  uchar source_mac[20];
  uchar target_mac[20];
  u_int16_t type;

  print_mac_address(target_mac, message->header);
  print_mac_address(source_mac, message->header + 6);

  memcpy(&type, message->header + 12, 2);
  type = ntohs(type);

  logmsg(MSG_INFO,"Rogue paoket dropped: %s --> %s type %04X\n",
	 source_mac, target_mac, type);
}
		      
static void dump_hex(unsigned char *buf, int bytes) {
  int i,j;

  for(i=0;i<bytes;i+=16) {
    fprintf(stdout,"%04X: ",i);
    for(j=i;j<bytes && j<i+16;j+=2) {
      fprintf(stdout,"%02X%02X ",buf[j],buf[j+1]);
    }
    fprintf(stdout,"\n");
  }

}

static void slanpacket_send_tunnel(slan_session_t *session_obj, 
				   slan_message_t *message) {

  /* Drop packet if key schedule is running. Someday maybe we'll queue 
     these, but I can't be bothered with it right now */
  if (session_obj->key_change < 0) {
    logmsg(MSG_WARNING,"Packet dropped on session id %d because key change "
	   "in progress.\n",session_obj->id);
    pthread_mutex_unlock(session_obj->lock);
    return;
  }
  
  session_obj->transport_activity += 1;
  session_obj->bytes_out += message->length;
  session_obj->packets_out ++;
  pthread_mutex_unlock(session_obj->lock);
  
  /* Sending down the transport layer is encapsulating and encrypting, and
     ultimately, sending to the client via UDP */
  session_obj->transport_top->down(session_obj->transport_top, message);
  
  pthread_mutex_lock(session_obj->lock);
  session_obj->transport_activity -= 1;
  
  /* Caller held lock, caller free's lock */
}

static slan_message_t *slan_message_clone(slan_message_t *source) {
  slan_message_t *target;

  target = malloc(sizeof(slan_message_t));
  if (target == NULL) return NULL;

  target->buffer = malloc(10240);
  if (target->buffer == NULL) {
    free(target);
    return NULL;
  }

  target->data = target->buffer + (source->data - source->buffer);
  target->header = target->buffer + (source->header - source->buffer);
  target->trailer = target->buffer + (source->trailer - source->trailer);
  
  target->length = source->length;
  memcpy(target->header, source->header, source->length);

  return target;
}
  

/* This function will wait for packets to arrive on the slan PACKET socket and 
   send them through the appropriate transport layer, encrypting and routing
   them to clients. Packets that have unknown destination IP address are
   discarded. Remember that at the server context, packets arriving from the
   SLAN device are return traffic from the internet. This function is run in a
   seperate thread. */
static void *slandevice_process_thread(void *arg) {
  int len;
  int slan_fd;
  unsigned char buf[10240];
  u_int32_t session_id;
  slan_message_t *message;
  slan_session_t *session_obj;
  unsigned char broadcast_addr[6] = {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};

  /* Get the file descriptor, passed as an argument to this function */
  slan_fd = *((int *) arg);

  /* Setup an slan message structure to be used by a transport layer */
  message = malloc(sizeof(slan_message_t));
  message->buffer = buf;
  message->maxspace = 4096;

  /* Loop forever, receiving and routing packets */
  while (1) {
    /* Reset pointers that were probably modified by previous action of a
       transport layer stack on the last packet */

    /* TODO: Some amount of header space is needed for each packet, the exact
       amount may depend on each client's session negotiation and state. Some
       parts of the (current) transport layer, like the hmac_sha1 layer
       actually use a portion of this header space for temp working area, to
       reduce overhead of producing a copy of the message with the hmac crap
       prepended to it.

       For correctness, this should be set to some variable advance from the
       start of the buffer, and that advance is the maximum of the required
       space over all active sessions.

       For now, we are allocating a 10k buffer on the stack and leaving a 4k
       temp/header space at the beginning and forgeting about it. */
    message->data = message->buffer + 4096;
    message->header = message->data;
    message->trailer = message->data;

    /* Read a message from the slan device driver. At the server context, this 
       should be messages returning from the internet, to be encrypted (and
       what not) and sent to the appropriate client. */
    len = read(slan_fd,message->data,message->maxspace);
#ifdef DEBUG
    fprintf(stderr,"Read packet from SLAN device\n");
    dump_hex(message->data,len);
#endif

    /* This ought not happen, but.. */
    if (len < 0) {
      pthread_mutex_lock(mutex_locks + TLOCK_CONSOLE);
      perror("Error reading from slan_socket");
      pthread_mutex_unlock(mutex_locks + TLOCK_CONSOLE);
      continue;
    }
    message->length = len;

    if (memcmp(message->data, broadcast_addr, 6) == 0) {
      /* Damn the broadcast.... */
      int i;
      slan_session_t *p;
      slan_message_t *dup_message;

      pthread_mutex_lock(mutex_locks + TLOCK_SESSION_TABLE);
      for(i=0;i<SESSION_TABLE_MASK;i++) {
	p = session_table[i];
	while(p!=NULL) {
	  slan_session_t *q;
	  if (! pthread_mutex_trylock(p->lock)) {
	    dup_message = slan_message_clone(message);
	    if (dup_message) {
	      slanpacket_send_tunnel(p, dup_message);
	      free(dup_message->buffer);
	      free(dup_message);
	    }
	    q = p->next;
	    pthread_mutex_unlock(p->lock);
	    p = q;
	  } else {
	    /* Race condition? Session object is not locked here! Hopefully
	       the full table lock will hold the fort... */
	    p = p->next;
	  }
	}
      }
      pthread_mutex_unlock(mutex_locks + TLOCK_SESSION_TABLE);
    } else {
      session_id = macaddr_lookup(message->data);
      if (session_id == 0) {
	log_incoming_rogue_packet(message);
	continue;
      }
      
      session_obj = session_lookup(session_id, 10, 10);
      if (session_obj == NULL) {
	logmsg(MSG_WARNING,"session_id %d: Packet dropped. session_obj not "
	       "avail.\n", session_id);
	continue;
      }
      slanpacket_send_tunnel(session_obj, message);
      pthread_mutex_unlock(session_obj->lock);
    }
  }
   
  /* This function currently never returns.
     
  TODO: We need a "graceful" server shutdown mechanism that should perhaps
  cause this function to terminate */
  return NULL;
}

static void forget_keys(slan_session_t *session_obj) {
  int i;

  /* Forget keys. NULL pointers may exist if shutdown happens before the 
     session is configured fully */
  if (session_obj->mackey!=NULL) {
    for(i=0;i<session_obj->mackey_length;i++)
      session_obj->mackey[i] = 0x0;
    free(session_obj->mackey);
  }
  session_obj->mackey = NULL;

  /* Usally diffie-hellman returns one big long key which we set the mackey
     pointer to something halfway through and the cipher key to the memory
     returned from DH. So we only free the cipher key. */
  if (session_obj->key!=NULL) {
    for(i=0;i<session_obj->key_length;i++)
      session_obj->key[i] = 0x0;
    free(session_obj->key);
  }
  session_obj->key = NULL;

  if (session_obj->keyhash!=NULL) {
    for(i=0;i<session_obj->keyhash_length;i++)
      session_obj->keyhash[i] = 0x0;
    free(session_obj->keyhash);
  }
  session_obj->keyhash = NULL;

}

static void server_transport_layer_destruct(slan_session_t *session_obj) {
  proto_stack_t *protocol_layer;

  if (session_obj->transport_top == NULL) return;
  /* TODO:
     Call destructors for protocol layers, when they exist. This is a memory
     leak here because the protocol layers are allocating memory internally. */

  protocol_layer = session_obj->transport_top->next;
  while(protocol_layer!=NULL) {
    proto_stack_t *next;;
    next = protocol_layer->next;
    free(protocol_layer->data);
    free(protocol_layer);
    protocol_layer = next;
  }

  /* Some layers are above our "top" entry point, because the
     slandevice_process_thread performs some of the work that would be done by 
     the protocol stack on the way down, in order to multiplex multiple
     clients. */
  protocol_layer = session_obj->transport_top->prev;
  while(protocol_layer!=NULL) {
    proto_stack_t *prev;
    prev = protocol_layer->prev;
    free(protocol_layer->data);
    free(protocol_layer);
    protocol_layer = prev;
  }
  
  free(session_obj->transport_top->data);
  free(session_obj->transport_top);
}

static void session_destroy(slan_session_t *session_obj) {
  
  forget_keys(session_obj);
  server_transport_layer_destruct(session_obj);

  session_obj->cipher->close(session_obj->cipher);
  session_obj->mac->close(session_obj->mac);

  shutdown(session_obj->control_tcp->fd, SHUT_RDWR);
  free(session_obj->control_tcp);
  free(session_obj->lock);
  free(session_obj);
}


/* This should be called with the lock held on the session_obj */
static void server_session_shutdown(slan_session_t *session_obj) {
  
  logmsg(MSG_DEBUG,"Session shutdown called.\n");

  /* Wait for transport layer to finish anything in progress. */
  while(session_obj->transport_activity > 0) {
    pthread_mutex_unlock(session_obj->lock);
    usleep(1000);
    pthread_mutex_lock(session_obj->lock);
  }

  control_handle_delete(session_obj->id);
  macaddr_table_delete(session_obj->macaddr);
  session_table_delete(session_obj->id);

  /* This is backwards from the client point of view:

     We log received bytes as client received bytes and 
     transmitted bytes as bytes client transmitted to the internet */
  logmsg(MSG_INFO,"Session id %lu: Username: %s\n",session_obj->id,
	 session_obj->username);
  logmsg(MSG_INFO,"Session id %lu: Transmitted %lu kB in %lu packets",
	 session_obj->id, session_obj->bytes_in/1024, session_obj->packets_in);
  logmsg(MSG_INFO,"Session id %lu: Received %lu kB in %lu packets",
	 session_obj->id, session_obj->bytes_out/1024, session_obj->packets_out);
 
  session_destroy(session_obj);
}

static void server_change_key(slan_session_t *session_obj) {

  forget_keys(session_obj);
  if (dh_listen(session_obj)) {
    pthread_mutex_lock(mutex_locks + TLOCK_CONSOLE);
    logmsg(MSG_WARNING,"Session ID: %d: DH keyexchange failed.\n",
	   session_obj->id);
    pthread_mutex_unlock(mutex_locks + TLOCK_CONSOLE);
    /* This is commented out for now. I am observing DH key exchanges failing
       during high server load (when its getting bombed by the robo-connect
       client) because the protocol times out. We should probably set the
       timeouts higher and eliminate the primality testing if the g/p numbers
       passed to the client are ones the client already knows. */
       /*server_session_shutdown(session_obj);*/
    return;
  }

  pthread_mutex_lock(session_obj->lock);
  session_obj->key_change = -1;

  /* Wait for the transport layer to finish anything in progress. */
  while(session_obj->transport_activity > 0) {
    pthread_mutex_unlock(session_obj->lock);
    usleep(1000);
    pthread_mutex_lock(session_obj->lock);
  }
  usleep(1000);
  
  session_obj->cipher->set_key(session_obj->cipher, session_obj->key, 
			       session_obj->key_length);
  session_obj->mac->set_key(session_obj->mac, session_obj->mackey,
			    session_obj->mackey_length); 
  session_obj->sequence_down = 0;
  session_obj->sequence_up = 0;

  session_obj->key_timestamp = time(NULL);
  session_obj->key_change = 0;
  pthread_mutex_unlock(session_obj->lock);
}

static void server_unexpected_shutdown(slan_session_t *session_obj) {

  logmsg(MSG_WARNING,"Unexpected shutdown called for session id %d.\n",
	 session_obj->id);
  server_session_shutdown(session_obj);
}

static void server_control_handler(control_handle_t *handle) {
  control_msg_t *msg;
  slan_session_t *session_obj;
  int rval, timeout;

  session_obj = session_lookup(handle->session_id, 100, 100);
  if (session_obj == NULL) return;
  pthread_mutex_unlock(session_obj->lock);

  timeout = 100;
  rval = control_msg_recv(&msg, session_obj, &timeout);
  if (rval < 0) {
    /* NOTE: Handle TCP Error */
    logmsg(MSG_WARNING,"Forcing session %lu shutdown because of TCP layer error\n",
	   session_obj->id);
    pthread_mutex_lock(session_obj->lock); //?
    server_unexpected_shutdown(session_obj);
    return;
  }
  if (rval > 0) {
    /* timeout! */
    logmsg(MSG_WARNING,"Expected control message from session %lu but no data"
	   " was received\n",session_obj->id);
    return;
  }
  
  switch(msg->type) {
  case KEYCHANGE:
    if (msg->subtype == KEYCHANGE_READY) {
      /* Key change initiated by the server -- the client has responded saying
	 it is ready to enter the key change protocol. Currently, the server
	 is _not_ initiating key changes ever, so this does not get used. */
      control_msg_destruct(msg);
      logmsg(MSG_DEBUG,"Client responding to server initiated key change\n");
      server_change_key(session_obj);
    } else {
      control_msg_t *new_msg;

      /* Key change initiated by client -- we respond saying we're ready
	 to start the keychange protocol and then enter the keychange 
	 function */
      control_msg_destruct(msg);
      if (session_obj->key_change == 0) {
	pthread_mutex_lock(session_obj->lock);
	session_obj->key_change = 1;
	pthread_mutex_unlock(session_obj->lock);

	new_msg = new_control_msg(KEYCHANGE, KEYCHANGE_READY, session_obj);
	timeout = 1000;
	rval = control_msg_send(new_msg, session_obj, &timeout);
	control_msg_destruct(new_msg);
	if (rval < 0) { 
	  logmsg(MSG_WARNING,"Forcing session %lu shutdown because of TCP layer "
		 "error\n", session_obj->id);
	  pthread_mutex_lock(session_obj->lock); //?
	  server_unexpected_shutdown(session_obj);
	  return;
	}
	if (rval > 0) {
	  pthread_mutex_lock(session_obj->lock);
	  session_obj->key_change = 0;
	  pthread_mutex_unlock(session_obj->lock);
	  return;
	}
	usleep(1000);
	server_change_key(session_obj);
      }
    }
    break;
  case SHUTDOWN:
    control_msg_destruct(msg);
    server_session_shutdown(session_obj);
    return; /* Session was destroyed, no lock to unlock */
    break;
  case KEEPALIVE:
    if (msg->subtype != KEEPALIVE_RESPONSE) {
      /* This should not happen. */
      logmsg(MSG_WARNING,"Not expecting a keepalive ping here, from session id "
	     "%d.\n", session_obj->id);
    } else {
      pthread_mutex_lock(session_obj->lock);
      session_obj->last_control_keepalive = time(NULL);
      pthread_mutex_unlock(session_obj->lock);
    }
    control_msg_destruct(msg);
    break;
  default:
    logmsg(MSG_WARNING,"Unhandled control message type %0lX for session id "
	   "%lu\n",msg->type,session_obj->id);
    control_msg_destruct(msg);
    break;
  }
}

#define MAX_PER_QUEST (1024)
#define KEEPALIVE_THRESHOLD (60)
#define KEEPALIVE_KILLTHRESHOLD (600)

static void *control_keepalive_thread(void *arg) {
  uint32 ids_toplug[MAX_PER_QUEST];
  int ids_size, i, rval, timeout;
  uint32 id;
  control_handle_t *handle;
  control_msg_t *new_msg;
  slan_session_t *session_obj;

  for(;;) {
    /* First, while holding the lock on the control handle list, run through
       the control list and see which sessions need a ping (or a forced
       shutdown) and build an array of ids. This is so that when we operate
       on the ids selected, we can do so lethargically and not hold the lock
       on the control list. If a session gets shutdown, we shouldn't be able
       to find it in the session table. */
    pthread_mutex_lock(mutex_locks + TLOCK_CONTROL_HANDLE_LIST);
    handle = control_handle_list;
    ids_size = 0;
    while(handle != NULL && ids_size < MAX_PER_QUEST) {
      id = handle->session_id;
      handle = handle->next;
      session_obj = session_lookup(id, 100, 100);
      if (session_obj == NULL) continue;
      if ((time(NULL) - session_obj->last_control_keepalive) > 
	  KEEPALIVE_THRESHOLD) {
	ids_toplug[ids_size++] = id;
      }
      pthread_mutex_unlock(session_obj->lock);
    }
    pthread_mutex_unlock(mutex_locks + TLOCK_CONTROL_HANDLE_LIST);

    for(i=0;i<ids_size;i++) {
      id = ids_toplug[i];
      session_obj = session_lookup(id, 100, 100);
      if (session_obj == NULL) continue;
      if (session_obj->key_change != 0) {
	/* Don't ping the client during key change events. */
	pthread_mutex_unlock(session_obj->lock);
	continue;
      }
      pthread_mutex_unlock(session_obj->lock);

      if ((time(NULL) - session_obj->last_control_keepalive) > 
	  KEEPALIVE_KILLTHRESHOLD) {
	logmsg(MSG_WARNING,"Session id %lu keepalive kill threshold "
	       "exceeded.\n", session_obj->id);
	server_unexpected_shutdown(session_obj);
      } else {
	new_msg = new_control_msg(KEEPALIVE, KEEPALIVE_PING, session_obj);
	timeout = 100;
	rval = control_msg_send(new_msg, session_obj, &timeout);
	control_msg_destruct(new_msg);
	if (rval < 0) {
	  logmsg(MSG_WARNING,"Forcing shutdown of session %lu due to tcp error\n",
		 session_obj->id);
	  server_unexpected_shutdown(session_obj);
	}
	if (rval > 0) {
	  logmsg(MSG_WARNING,"Socket layer timeout while trying to ping session "
		 "%lu\n",session_obj->id);
	}
      }
      usleep(1000);
    }
    usleep(5000000/(ids_size + 1));
  }
}
      
    

static void *control_process_thread(void *arg) {
  int max_fd, handles_active;
  control_handle_t *handle;
  fd_set read_fds;
  struct timeval tv;

  tv.tv_sec = 0;
  tv.tv_usec = 100000;

  for(;;) {
    max_fd = control_handle_build_fdset(&read_fds);
    if (max_fd == 0) {
      sleep(5);
    } else {
      handles_active = select(max_fd + 1, &read_fds, NULL, NULL, &tv);
      if (handles_active > 0) {
	
	/* We use control_handle_test_fdset to read the read_fds, getting and 
	   dropping the lock on the control_handle_list, and restarting the
	   linear search after processing each handle, because a handle could
	   dissappear by the time we get around to checking it. */
	while((handle = control_handle_search_fdset(&read_fds))) {
	  server_control_handler(handle);
	}
      } else {
	usleep(100000);
      }
    }
  }
}

/* This function listens for packets arriving on the SLAN UDP socket. All
   clients will be sending data to the same UDP port, so we just take them and 
   route them through the client's transport layer, as identified by the
   session id number found in the packet. This function runs in a seperate
   thread, to even out latencies for all clients in both sending and 
   receiving. 
 
   Note: We may be able to use the sending IP address instead of the session
   id, perhaps eliminating the need for a session id. */
static void *udpdata_process_thread(void *arg) {
  int len;
  int udp_fd;
  unsigned char buf[10240];
  int s_length;
  struct sockaddr_in source_addr;
  slan_message_t *message;
  u_int32_t session_id;
  slan_session_t *session_obj;

  /* Get the file descriptor for the UDP socket, passed as an argument to this 
     function */
  udp_fd = *((int *) arg);

  message = malloc(sizeof(slan_message_t));
  message->buffer = buf;
  message->maxspace = 4096;

  /* Loop forever, continuouly receiving UDP packets and directing them to the 
     appropriate transport layer for decryption, verification, and routing to
     the Internet. */
  while(1) {
    /* Reset pointers because the protocol stack's action on the previous
       packet should have moved them about. */
    s_length = sizeof(source_addr);
    /* Please see note above about reserving data at the beginning of the       message. */
    message->data = message->buffer + 4096;
    message->header = message->data;
    message->trailer = message->data;

    len = recvfrom(udp_fd, message->data, message->maxspace, 0, 
		   (struct sockaddr *) &source_addr, &s_length);
    if (len < 0) {
      pthread_mutex_lock(mutex_locks + TLOCK_CONSOLE);
      perror("Error reading from UDP socket");
      pthread_mutex_unlock(mutex_locks + TLOCK_CONSOLE);
      /* This may not be a fatal error, so lets keep looping hoping that this
	 was caused by a temporary resource starvation or something. */
      continue;
    }

    message->length = len;
#ifdef DEBUG
    fprintf(stderr,"Read packet from UDP channel\n");
    dump_hex(message->data,len);
#endif

    /* Try to locate the client's session object for this packet */
    session_id = slan_parse_session_id(message);
    session_obj = session_lookup(session_id, 10, 10);
    /* If we can't find a connection associated with the session id in the
       packet, we just drop the packet on the floor and forget it. */
    if (session_obj == NULL) {
      logmsg(MSG_WARNING,"Session id %lu: packet dropped. session_obj not "
	     "avail.", session_id);
      continue;
    }
#ifdef DEBUG
    logmsg(MSG_DEBUG,"Server: udp thread: Received packet for session id %d\n",
	    session_id);
#endif

    /* Note that a udp packet with valid session id could be received, before
       the transport layer at the server is setup. This would cause a
       segmentation fault! So, we only copy the client_addr if we have a
       pointer to somewhere. */
    if (session_obj->client_addr != NULL) {
      /* Copy the source address of this UDP message, so the return trip knows 
	 where to send the packets. Note that this might be a moving target if 
	 the client is behind a NAT somewhere, otherwise, it ought not change
	 after we've copied it once. */
      memcpy(session_obj->client_addr,&source_addr,sizeof(struct sockaddr_in));
    }
    
    session_obj->transport_activity += 1;

    pthread_mutex_unlock(session_obj->lock);

    /* Route the packet on its way */
    session_obj->transport_bottom->up(session_obj->transport_bottom,message);

    pthread_mutex_lock(session_obj->lock);
    session_obj->transport_activity -= 1;
    pthread_mutex_unlock(session_obj->lock);
  }
  return NULL;
}

/* This function is called when the control protocol setup has completed,
   resulting in a session key establishing trust between the client and server 
   (authentication). This function builds the transport layer protocol stack
   and places the top and bottom entry points into the slan session object.

   TODO: The slan session object should either be locked or not added into the 
   table until this has been completed. The above functions will attempt to
   dereference the bottom and top poionters if packets are received before the 
   server has gotten around to executing this function to completion */
static void server_transport_layer_build(slan_session_t *session_obj) {
  proto_stack_t *last, *current;
  int header_save;
  
  header_save = 0;
  /* A stack terminator at the top -- this is simply a dummy protocol
     layer that drops any packet passed to it. It can be used to catch
     bad or dropped packets and debug them. Otherwise, it isn't necessary and
     doesn't do much. */
  current = last = malloc(sizeof(proto_stack_t));
  terminate_register(current);
  header_save += current->header_reserve;
  current->prev = NULL;
  
  /* Packets received from the slan device sent down the stack from the top */
  current = current->next = malloc(sizeof(proto_stack_t));
  slan_link_register(current, slan_fd, session_obj);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* Optional fragmenting protocol layer -- currently not implemented */
  current = current->next = malloc(sizeof(proto_stack_t));
  fragment_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  /* In the server process, a master thread will receive the slan device
     messages and direct packets to the appropriate transport stack, injecting
     messages into the stack after the slan_link layer */
  session_obj->transport_top = current;
  last = current;


  /* All packets being sent through the slan tunnel get tagged with sequence
     numbers before adding a MAC and encrypting. This used to limit the
     possibilities for message replay attacks. */
  current = current->next = malloc(sizeof(proto_stack_t));
  sequence_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* After adding the sequence number, we add a Message Authentication Code
     (MAC) to allow the receiver to verify the integrity of the packet and the 
     sender of the data */
  current = current->next = malloc(sizeof(proto_stack_t));
  message_sign_register(current, session_obj->mac);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* Lastly, we enrypt the packet for privacy during transmission. Currently,
     I am expecting the symmetric cipher used to be a configurable option, so
     the cipher used depends on the field in the session_obj */
  current = current->next = malloc(sizeof(proto_stack_t));
  symmetric_cipher_register(current, session_obj->cipher);
  header_save += current->header_reserve;
  current->prev = last;
  /* In the server process, a master UDP receiving thread will inject messages 
     just after the session id layer, using the session id in the UDP packet
     to figure out which stack to send the packet to */
  session_obj->transport_bottom = current;
  last = current;

  /* Session ID numbers are prepended to identify the session the message
     belongs to */
  current = current->next = malloc(sizeof(proto_stack_t));
  session_id_register(current);
  header_save += current->header_reserve;
  *((u_int32_t *) current->data) = session_obj->id;
  current->prev = last;
  last = current;

  /* Transmit the secured packet via the UDP channel */
  current = current->next = malloc(sizeof(proto_stack_t));
  /* The udplink register function returns to us a pointer to a struct
     sockaddr_in, which is part of the udplink's local data structure. We set
     the session_obj->client_addr pointer to this pointer, and use that to
     communicate the current UDP/IP address of the client, as per what we
     receive from the client in the messages sent to us. */
  session_obj->client_addr = server_udplink_register(current, udp_fd);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;

  /* Terminate the bottom of the stack */
  current = current->next = malloc(sizeof(proto_stack_t));
  terminate_register(current);
  header_save += current->header_reserve;
  current->prev = last;
  last = current;
  
  current->next = NULL;

  session_obj->header_save = header_save;
}

/* For the moment, tell the client what IP address we expect it to use.

   TODO: Replace this kludge with ability to DHCP the slan interface */
static u_int32_t server_assign_ip() {
  /* Assume for now we are assigning 10.1.x.x addresses */
  static u_int32_t lastassigned_address = 0;
  static u_int32_t broadcast;
  u_int32_t current_address;


  pthread_mutex_lock(mutex_locks + TLOCK_IPASSIGN);

  /* This happens when the first IP address is assigned */
  if (lastassigned_address == 0) {
    lastassigned_address = 
      ntohl(inet_addr(slan_global_config->slan_ipaddress));
    broadcast = ntohl(inet_addr(slan_global_config->broadcast));
  }

  current_address = lastassigned_address + 1;
  if (current_address == broadcast) {
    current_address = ntohl(inet_addr(slan_global_config->slan_ipaddress)) + 1;
  }
  while (ipaddr_lookup(current_address)>0 &&
	 current_address != lastassigned_address) { 
    current_address++;
    if (current_address == broadcast) {
      current_address = 
	ntohl(inet_addr(slan_global_config->slan_ipaddress)) + 1;
    }
  }

  if (current_address == lastassigned_address) {
  /* ACK! entire range of IP addresses are already assigned! */
    logmsg(MSG_CRITICAL,"Cannot assigned IP address for client, IP address "
	   "range is exhausted. ");
    pthread_mutex_unlock(mutex_locks + TLOCK_IPASSIGN);
    return 0;
  }
 
  lastassigned_address = current_address;
  pthread_mutex_unlock(mutex_locks + TLOCK_IPASSIGN);
  return (current_address);
}

#if 0
/* This function sends a message to the client, telling what IP address to use
   for the slan tunnel. This will be used to identify packets returning from
   the internet for the client */
static ip_info_t *server_set_client_ipaddress(slan_session_t *session_obj) {
  u_int32_t ip;
  ip_info_t *ip_conf;

  ip = server_assign_ip();
  if (ip == 0) return NULL;

  session_obj->ipaddr = ip;
  ipaddr_table_add(session_obj->id, ip);

  ip_conf = (ip_info_t *) calloc(1, sizeof(ip_info_t));
  
  /* Temporary -- until DHCP is used. */
  ip_conf->ip_address = htonl(ip);
  ip_conf->subnet_mask = inet_addr(slan_global_config->netmask);
  ip_conf->subnet_bcast = inet_addr(slan_global_config->broadcast);
  ip_conf->router = inet_addr(slan_global_config->router);
  ip_conf->nameserver_1 = inet_addr(slan_global_config->nameserver1);
  ip_conf->nameserver_2 = 0x0;

  return ip_conf;
}
#endif


/* This is a threaded function that performs the session setup and prepares
   the transport layer to handle the client's data payload. When a connection
   is started, we spawn a thread to handle this work. When it is finished the
   thread simply exits. We use a seperate thread because a client's session
   request may experience protocol timeouts or expensive computations that
   could tie up the server for several seconds. We want both existing
   connections to proceed without delays, as well as the ability to setup
   several connections at the same time. */
static void *session_start(void *arg) {
  slan_session_t *session_obj;
  uchar macaddr[20];

  session_obj = (slan_session_t *) arg;

  /* In the future, we'll have do some fancy work here when we get the
     client's version information. For now, we are just going to yes/no the
     connection based on whether or not the numbers match. */
  if (check_version(session_obj->control_tcp->fd) < 0) {
    /* Version mismatch */
    pthread_mutex_lock(mutex_locks + TLOCK_CONSOLE);
    logmsg(MSG_ERROR,"Session ID: %d: Client version incompatible.\n",
	   session_obj->id);
    pthread_mutex_unlock(mutex_locks + TLOCK_CONSOLE);
    session_destroy(session_obj);
    return NULL;
  }

  /* Perform Diffie-Hellman key exchange */
  if (dh_listen(session_obj)) {
    pthread_mutex_lock(mutex_locks + TLOCK_CONSOLE);
    logmsg(MSG_ERROR,"Session ID: %d: DH keyexchange failed.\n",
	   session_obj->id);
    pthread_mutex_unlock(mutex_locks + TLOCK_CONSOLE);
    /* Bail on DH key exchange failure. This will not affect currently running 
       slan sessions */
    session_destroy(session_obj);
    return NULL;
  }

  /* Setup encryption for this client and message authentication */
  session_obj->cipher->set_key(session_obj->cipher, session_obj->key, 
			       session_obj->key_length);
  session_obj->mac->set_key(session_obj->mac, session_obj->mackey,
			    session_obj->mackey_length); 

  /* Expect the client to authenticate us next with RSA. If the client fails
     to initiate RSA authentication, rsaauth_listen will timeout and return
     non-zero, terminating the connection */
  if (rsaauth_listen(session_obj)) {
    pthread_mutex_lock(mutex_locks + TLOCK_CONSOLE);
    logmsg(MSG_ERROR,"Session id %lu: Server failed RSA Authentication.\n",
	    session_obj->id);
    pthread_mutex_unlock(mutex_locks + TLOCK_CONSOLE);
    session_destroy(session_obj);
    return NULL;
  }

  if (server_clientauth_initiate(session_obj)) {
    pthread_mutex_lock(mutex_locks + TLOCK_CONSOLE);
    logmsg(MSG_ERROR,"Session id %lu: Client authentication failed.\n",
	    session_obj->id);
    pthread_mutex_unlock(mutex_locks + TLOCK_CONSOLE);
    session_destroy(session_obj);
    return NULL;    
  }
    
  if (macconf_req_wait(session_obj, macaddr)) {
    logmsg(MSG_ERROR,"Failed negotiation for client MAC address\n");
    session_destroy(session_obj);
    return NULL;
  }
  macaddr_table_add(session_obj->id, macaddr);
  memcpy(session_obj->macaddr, macaddr, 6);

  /* Setup the transport layer */
  server_transport_layer_build(session_obj);

  /* Debugging output: The session is now started. */
  logmsg(MSG_INFO,"Session id %lu: Startup Okay. Username: %s\n",
	 session_obj->id,session_obj->username);

  control_handle_add(session_obj->control_tcp->fd, session_obj->id);
  session_table_add(session_obj);

  pthread_mutex_unlock(session_obj->lock);

  /* We don't need this thread anymore, dump it. */
  return NULL;
}

static tcp_socket_t *server_accept(int fd, struct sockaddr_in *address, 
				   socklen_t *addrlen) {
  int client_fd;
  tcp_socket_t *tcp_socket;

  *addrlen = sizeof(struct sockaddr_in);
  client_fd = accept(fd, (struct sockaddr *) address, addrlen);
  if (client_fd < 0) {
    logmsg(MSG_ERROR,"Failed to accept client on listening tcp socket (%s)\n",
	   strerror(errno));
    return NULL;
  }
  
  tcp_socket = malloc(sizeof(tcp_socket_t));
  if (tcp_socket == NULL) {
    logmsg(MSG_ERROR,"Failed allocating memory in server_accept() (%d bytes)\n",
	   sizeof(tcp_socket_t));
    return NULL;
  }
  tcp_socket->start = tcp_socket->end = tcp_socket->tmp = 0;
  tcp_socket->fd = client_fd;
  
  return tcp_socket;
}

/* This function continuously listens for incoming connections and kicks off a
   session_start thread to do the setup when a connection is found. */ 
static void *server_listen(void *arg) {
  int listen_fd, client_addr_length;
  tcp_socket_t *client_control_tcp;
  slan_session_t *session_obj;
  pthread_attr_t session_start_attr;
  struct sockaddr_in client_address;
  u_int32_t unique_id;

  listen_fd = *((int *) arg);
  pthread_attr_init(&session_start_attr);
  pthread_attr_setdetachstate(&session_start_attr, PTHREAD_CREATE_DETACHED);

  while(1) {
    client_control_tcp = 
      server_accept(listen_fd, &client_address, &client_addr_length);
    if (client_control_tcp == NULL) continue;
  
    /* Debugging output */
    pthread_mutex_lock(mutex_locks + TLOCK_CONSOLE);
    logmsg(MSG_INFO,"Received connection from %s\n", 
	   inet_ntoa(client_address.sin_addr));
    pthread_mutex_unlock(mutex_locks + TLOCK_CONSOLE);
    
    unique_id = server_get_unique_id();
    session_obj = calloc(1, sizeof(slan_session_t));

    session_obj->lock = calloc(1, sizeof(pthread_mutex_t));
    pthread_mutex_init(session_obj->lock, NULL);
    pthread_mutex_lock(session_obj->lock);

    session_obj->id = unique_id;
    session_obj->control_tcp = client_control_tcp;
    session_obj->last_control_keepalive = time(NULL);
        
    session_obj->next = NULL;
    session_obj->ipaddr = 0;
    session_obj->client_addr = NULL;
    
    session_obj->bytes_in = 0;
    session_obj->bytes_out = 0;
    session_obj->packets_in = 0;
    session_obj->packets_out = 0;
    
    session_obj->cipher = get_blowfish();
    session_obj->mac = get_hmac_sha1();
    session_obj->transport_top = NULL;
    session_obj->transport_bottom = NULL;
    session_obj->header_save = 0;
    
    session_obj->key_length = 0;
    session_obj->key = NULL;
    session_obj->mackey = NULL;
    session_obj->mackey_length = 0;
    session_obj->keyhash_length = 0;
    session_obj->keyhash = NULL;
    
    session_obj->username_length = 0;
    session_obj->username = NULL;
    session_obj->password_length = 0;
    session_obj->password = NULL;
    
    /* Set the null key for the mac and cipher. They will be overidden 
       with real keys when the DH keyexchange completes. */
    session_obj->cipher->set_key(session_obj->cipher, null_key, 32);
    session_obj->mac->set_key(session_obj->mac, null_key, 32);

    
    pthread_create(&session_obj->thread_obj, &session_start_attr, 
		   session_start, session_obj);
  }
  /* This function never returns and nobody is waiting for its return value
     anyway. */
  return NULL;
}

/* This function opens the UDP socket for the server. */
static int server_open_udp_socket() {
  int sock;
  struct protoent *protocol;
  struct sockaddr_in address;
  
  protocol = getprotobyname("udp");
  if (protocol == NULL) {
    logmsg(MSG_CRITICAL,"Can not find the operating system protocol number "
	   "for UDP. (FATAL)\n");
    exit(-1);
  }

  sock = socket(PF_INET, SOCK_DGRAM, protocol->p_proto);
  if (sock == -1 ) {
    logmsg(MSG_CRITICAL,"Cannot open network socket for UDP/IP (%s)\n",
	   strerror(errno));
    exit(-1);
  }

  address.sin_family = AF_INET;
  address.sin_addr.s_addr = htonl(INADDR_ANY);
  address.sin_port = htons(slan_global_config->transport_port);

  if (bind(sock, (struct sockaddr *) &address, sizeof(struct sockaddr_in))) {
    logmsg(MSG_CRITICAL,"Cannot bind UDP/IP socket: (%s)\n",
	   strerror(errno));
    exit(-1);
  }

  logmsg(MSG_INFO,"UDP/IP socket created.\n");
  return sock;
}

/* This function opens a socket to the slan device via the PACKET interface. */
static int server_open_slan_socket(char *dev) {
  struct ifreq ifr;
  int fd;

  if ( (fd = open("/dev/net/tun", O_RDWR)) < 0) {
    logmsg(MSG_CRITICAL,"Can not open TUN/TAP device. Can not start SLAN "
	   "server. (%s)\n",strerror(errno));
    exit(-1);
  }

  memset(&ifr, 0, sizeof(ifr));
  ifr.ifr_flags = (IFF_TAP | IFF_NO_PI);
  strncpy(ifr.ifr_name, dev, IFNAMSIZ);

  if (ioctl(fd, TUNSETIFF, (void *) &ifr) < 0) {
    close(fd);
    fprintf(stderr,"ioctl() failed (%s)\n",strerror(errno));
    return -1;
  }
  strcpy(dev, ifr.ifr_name);
  
  return fd;
}

int main(int argc, char *argv[]) {
#ifndef DEBUG
  pid_t pid;
#endif
  int listen_fd;
  char *config_filename;
  char dev[20];
  pthread_attr_t thread_attributes;
  pthread_t listen_thread;
  pthread_t udpdata_thread;
  pthread_t slanlink_thread;
  pthread_t control_thread;
  pthread_t keepalive_thread;

  configure_logmsg(0,1,-1);

  memset(null_key, 0x0, 32);
  random_initialize();

  if (argc==2) {
    config_filename = argv[1];
  } else {
    config_filename = "/etc/slan/slan_server.conf";
  }

  if (slan_server_load_configuration(config_filename)) {
    logmsg(MSG_CRITICAL,"Cannot start slan server: Configuration file could "
	   "not be found.\n");
    logmsg(MSG_INFO,"Please run slan_setup to create a configuration file.\n");
    logmsg(MSG_INFO,"The server will expect by default to find it (as well "
	    "as the \nprivate key file) in /etc/slan/slan_server.conf\n");
    exit(-1);
  }

  session_table_init();
  ipaddr_table_init();
  macaddr_table_init();

  udp_fd = server_open_udp_socket();

  strcpy(dev,"tap0");
  slan_fd = server_open_slan_socket(dev);

  listen_fd = server_open_socket(slan_global_config->control_port);

  logmsg(MSG_INFO,"Server configuration and startup okay.\n");

  /* Lock ALL memory -- This process is important to functioning of the
     network for the client, and should never be swapped out. As well (and
     more importantly), the memory used by this process holds vital security
     sensitive information, such as cryptographic keys, and must never be
     written to permanent store. */
#if defined(linux) && ! defined(DEBUG)

  if (mlockall(MCL_FUTURE)) {
    logmsg(MSG_CRITICAL,"Cannot start SLAN server because memory pages could "
	   "not be locked into physical RAM (%s)\n",strerror(errno));
    logmsg(MSG_CRITICAL,"Most likely reason is the server was not started as "
	   "root\n");
    exit(-1);
  }
#endif

#ifndef DEBUG
  if ((pid = fork())!=0) {
    if (pid == -1) {
      logmsg(MSG_CRITICAL,"Failed to fork process into background. (%s)\n",
	     strerror(errno));
      exit(-1);
    } else {
      logmsg(MSG_INFO,"SLAN server (forked) pid = %d\n",pid);
      return 0;
    }
  } else {
    if (setegid(SLANRUN_GID) || seteuid(SLANRUN_UID)) {
      logmsg(MSG_WARNING,"Failed to change to uid/gid %d/%d (%s)\n",
	     SLANRUN_UID, SLANRUN_GID, strerror(errno));
    } else {
      logmsg(MSG_INFO,"Root priveledge dropped. Running as uid %d gid %d\n",
	     SLANRUN_UID, SLANRUN_GID);
    }
    configure_logmsg(1,0,-1);
    fclose(stderr);
    fclose(stdout);
    fclose(stdin);
  }
#endif

  pthread_attr_init(&thread_attributes);
  pthread_attr_setdetachstate(&thread_attributes,PTHREAD_CREATE_DETACHED);

  pthread_create(&listen_thread, &thread_attributes, server_listen, 
		 &listen_fd);

  pthread_create(&udpdata_thread, &thread_attributes, udpdata_process_thread, 
		 &udp_fd);
 
  pthread_create(&slanlink_thread, &thread_attributes, 
  		 slandevice_process_thread, &slan_fd);
  pthread_create(&control_thread, &thread_attributes, control_process_thread, 
		 NULL);
  pthread_create(&keepalive_thread, &thread_attributes, 
		 control_keepalive_thread, NULL);

  while(1) sleep(7200);
  return 0;
}

