#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <pthread.h>
#include <syslog.h>

/* ---- This file's types and exports ---- */
#include <slan_types.h>
#include <log_message.h>

static int use_syslog;
static int use_console;
static int verbosity_level;

/* For translating our own verbosity codes into the syslog() function's
   verbosity codes. see syslog(3) */
static int syslog_trans[N_MSG] = { LOG_DEBUG, LOG_INFO, LOG_INFO, 
				   LOG_WARNING, LOG_ERR, LOG_CRIT, LOG_ALERT };


static int logmessage_ready = 0;
static pthread_mutex_t logmessage_lock;

/* Set variables which control where a message is copied. Negative value means
   don't modify the current setting. This function should be called at startup
   with 0,1,LOG_INFO, and switched to 1,0,LOG_WARNING when the program forks
   into background. Debugging builds however may do something different. */
int configure_logmsg(int syslog_status, int console_status, int verbosity_status) {
  
  if (logmessage_ready == 0) {
    pthread_mutex_init(&logmessage_lock, NULL);
    logmessage_ready = 1;
  }

  pthread_mutex_lock(&logmessage_lock);
  if (syslog_status >= 0) 
    use_syslog = syslog_status;
  if (console_status >= 0)
    use_console = console_status;
  if (verbosity_status >= 0)
    verbosity_level = verbosity_status;
  pthread_mutex_unlock(&logmessage_lock);

  return 0;
}

/* Write a message to the syslog or console, depending on its priority level
   and the messaging configuration via the variables above. */ 
void logmsg(int priority, volatile uchar *s, ...) {
  va_list ap;
  uchar long_string[4096];

  if (logmessage_ready) {
    va_start(ap, s);
    vsnprintf(long_string ,4096, (uchar *) s, ap);
    va_end(ap);

    pthread_mutex_lock(&logmessage_lock);
    if (priority >= verbosity_level) {
      if (use_console) {
	fprintf(stderr, "%s", long_string);
      }
      if (use_syslog) {
	syslog(syslog_trans[priority], "%s", long_string);
      }
    }
    pthread_mutex_unlock(&logmessage_lock);

  } 
  else {
    fprintf(stderr,"Call configure_logmsg() first!\n");
  }
  if (priority == MSG_FATAL) exit(-1);
}
