/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/transport/client_udplink.c,v 1.1.1.1 1999/12/09 01:16:43 koni Exp $ 
  $Author: koni $ $Date: 1999/12/09 01:16:43 $ $Name: slan-release-0-04 $
  $Revision: 1.1.1.1 $

*/
   
/*

  KONI Oct. 28-29, 2000
  Stage 1 client daemon process for "lightlink secure lan" project

  udp_link.c: This file implements functions to open, send, and receive data
              from a server via UDP.

  Oct. 29: This file split off because of header conflicts with the packet
           interface code.

*/

/* ---- Include files ----- */
#ifndef WIN32
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#else
#include "slan_windows.h"
#endif

#include "slan_alloc.h"
#include "slan_types.h"

#include "proto_stack.h"
#include "slan_version.h"
#include "slan_session.h"

typedef struct {
  SOCKET fd;
  struct sockaddr_in server_addr;
  slan_session_t *session_obj;
} udplink_t;

/* server_open_socket()

   This function opens a socket to the kernel's UDP/IP layer, for sending 
   packets encapsulated packets to the server.

   TODO: Should this try to bind a specific interface and port? If not, it
   will require that the server learn the port and address to use. */
static void server_open_socket(udplink_t *local) {
  SOCKET sock;
  struct protoent *protocol;
  struct sockaddr_in address;
  
  protocol = getprotobyname("udp");
  if (protocol == NULL) {
    fprintf(stderr,"sland: A fatal error has occured, cannot find the "
	    "operating system protocol number for UDP.\n");
#ifndef WIN32
    syslog(LOG_CRIT,"Cannot determine the OS protocol number for UDP.\n");
#endif
    exit(-1);
  }

  sock = socket(PF_INET, SOCK_DGRAM, protocol->p_proto);
  if (sock == -1 ) {
    perror("sland: Cannot open network socket for UDP/IP: ");
#ifndef WIN32
    syslog(LOG_CRIT,"Failed to start SLAN: Cannot open UDP/IP network :"
	   "socket.\n");
#endif
    exit(-1);
  }

  address.sin_family = AF_INET;
  address.sin_addr.s_addr = htonl(INADDR_ANY);
  address.sin_port = 0;

  if (bind(sock, (struct sockaddr *) &address, sizeof(struct sockaddr_in))) {
    perror("sland: Cannot bind UDP/IP socket: ");
#ifndef WIN32
    syslog(LOG_CRIT,"Failed to bind UDP/IP socket.\n");
#endif
    exit(-1);
  }

#ifndef WIN32
  syslog(LOG_INFO,"UDP/IP socket created.\n");
#endif
  local->fd = sock;
}

/* udplink_set_server()

   Attempts to resolve the string containing the server's hostname to an
   IP address. If it was specified as an IP address, gethostbyname() should
   convert it for us to network byte order binary. Then we setup a sockaddr
   struct to use when pushing datagrams to the server. server_addr is a
   file scope variable, statically allocated. */
static void udplink_set_server(udplink_t *local, char *server_address, int port) {
  struct hostent *hostname_hent;
  
  if((local->server_addr.sin_addr.s_addr = inet_addr(server_address)) == INADDR_NONE)
  {
  hostname_hent = gethostbyname(server_address);
  if (hostname_hent == NULL) {
#ifndef WIN32
      fprintf(stderr,"client: Could not resolve server hostname %s: %s\n",
              server_address, hstrerror(h_errno));
      syslog(LOG_CRIT,"Could not resolve server hostname %s (%s)\n",
	     server_address, hstrerror(h_errno));
#endif
    exit(-1);
  }

  memcpy(&local->server_addr.sin_addr.s_addr,hostname_hent->h_addr_list[0],
	 hostname_hent->h_length);
  }

  local->server_addr.sin_port = htons(port);
  local->server_addr.sin_family = AF_INET;
}

static void client_udplink_down(proto_stack_t *stack, slan_message_t
				*message) {
  udplink_t *local;
  int len;

  local = stack->data;
  len = sendto(local->fd,message->data,message->length, 0,
	       (struct sockaddr *) &local->server_addr, sizeof(struct sockaddr_in));
  if (len < message->length) {
#ifndef WIN32
    syslog(LOG_ERR,"Failed sending packet to server (%m)\n");
#endif
  }
  local->session_obj->bytes_out += message->length;
  local->session_obj->packets_out ++;
  stack->next->down(stack->next,message);
}

static void client_udplink_up(proto_stack_t *stack, slan_message_t
			      *message) {
  udplink_t *local;
  int s_length, len;
  struct sockaddr_in source_addr;

  local = stack->data;

#ifdef WIN32
#error Oh.. you are using this
  FD_ZERO(&rfds);
  FD_SET(local->fd, &rfds);
  
  tv.tv_sec = 0;
  tv.tv_usec = 0;
  retval = select(local->fd+1, &rfds, NULL, NULL, &tv);

  if (retval == -1 ) {
    /* Debugging stuff */
#ifndef WIN32
    syslog(LOG_ERR,"Error checking UDP socket for data (%m)\n");
#endif
    local->session_obj->up_blocking = 1;
    slan_free(message);
    return;
  } else if (retval == 0) {
    local->session_obj->up_blocking = 1;
    slan_free(message);
    return;
  }
#endif
  
  local = stack->data;
  s_length = sizeof(source_addr);
  len = recvfrom(local->fd, message->data, message->maxspace, 0,
 		 (struct sockaddr *) &source_addr, &s_length); 
  
  if (len < 0) {
    /* Debugging */
#ifndef WIN32
    syslog(LOG_ERR,"Failed receiving data from the server (%m)\n");
#endif
    slan_free(message);
    return;
  }

#ifdef DEBUG
  fprintf(stderr,"Recieved UDP packet from server.\n");
#endif
  if (memcmp(&source_addr.sin_addr.s_addr,
	     &local->server_addr.sin_addr.s_addr,
	     sizeof(source_addr.sin_addr))) {
    /* Debugging */
#ifndef WIN32
    syslog(LOG_ERR,"While waiting for the server, we got a packet from "
	    "someplace else!\n");    
    syslog(LOG_ERR,"Source of packet reported by network layer is %s\n",
	   inet_ntoa(source_addr.sin_addr));
#endif
    slan_free(message);
    return;
  }

  message->length = len;
  stack->prev->up(stack->prev, message);
}

int client_udplink_register(proto_stack_t *stack_entry, char *address, 
			    int port, slan_session_t *session_obj) {
  udplink_t *local;

  stack_entry->up = client_udplink_up;
  stack_entry->down = client_udplink_down;
  stack_entry->header_reserve = 0;
  stack_entry->trailer_reserve = 0;
  
  local = (udplink_t *) slan_calloc(1, sizeof(udplink_t));
  if (local == NULL) return 1;
  
  server_open_socket(local);
  udplink_set_server(local, address, port);

  local->session_obj = session_obj;

  stack_entry->data = local;
  session_obj->udp_socket = local->fd;
  
  return 0;
}

