/*
  Lightlink SLAN -- A Virtual Private Network System
  (c) 2001 Mark Wright (Koni), Lucas Madar, Art Matrix - Lightlink, Inc.  

  This program is free software; you can redistribute it and/or	
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  $Header: /home/cvsroot/stage6/transport/slan_link.c,v 1.1.1.1 1999/12/09 01:16:43 koni Exp $ 
  $Author: koni $ $Date: 1999/12/09 01:16:43 $ $Name: slan-release-0-04 $
  $Revision: 1.1.1.1 $

*/
   
/*

  KONI Oct. 28-29, 2000
  Stage 1 _client_ daemon process for "lightlink secure lan" project
  
  slan_link.c : This file implements functions to open, send, and receive data
                from the slan device driver running on the local machine.

  Oct 29: File split off from UDP/IP communications because of header file
          conflicts.

*/

/* ---- Include files ----- */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <linux/if_tun.h>
#include <linux/if.h>
#include <sys/types.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/time.h>
#include <linux/ip.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/ioctl.h>
#include <linux/if_arp.h>
#include <sys/poll.h>
#include "proto_stack.h"

#include "slan_version.h"
#include "slan_session.h"

#include "slan_alloc.h"

/* ARG! HACK */
#define ARPHRD_VOID 0xFFFF

typedef struct {
  int fd;
  slan_session_t *session_obj;
} slan_link_t;

/* open_slan_socket()

   This function opens a network socket of type PF_PACKET, which allows us to
   send frames of data directly to a network device driver, or receive frames
   from a network device driver. It also attempts to find the network device
   index number of our device driver, tap0. If it cannot open the socket or
   it cannot find the index number of tap0, we abort the program. */
static int slan_open_socket(char *dev) {
  struct ifreq ifr;
  int fd;  

  if ((fd = open("/dev/net/tun", O_RDWR)) < 0) {
        fprintf(stderr,"Can not open TUN/TAP device (%s)\n",strerror(errno));
    syslog(LOG_CRIT,"Can not open TUN/TAP device. Can not start SLAN server. (%m)\n");
    exit(-1);
  }

  memset(&ifr, 0, sizeof(ifr));
  ifr.ifr_flags = (IFF_TAP|IFF_NO_PI);
  strncpy(ifr.ifr_name, dev, IFNAMSIZ);

  if (ioctl(fd, TUNSETIFF, (void *) &ifr) < 0) {
    close(fd);
    fprintf(stderr,"ioctl() failed (%s)\n",strerror(errno));
    return -1;
  }
  strcpy(dev, ifr.ifr_name);
  
  return fd;
}

static void dump_hex(unsigned char *buf, int bytes) {
  int i,j;

  for(i=0;i<bytes;i+=16) {
    fprintf(stdout,"%04X: ",i);
    for(j=i;j<bytes && j<i+16;j+=2) {
      fprintf(stdout,"%02X%02X ",buf[j],buf[j+1]);
    }
    fprintf(stdout,"\n");
  }

}

/* This gets a message from the slan device. If a message is
   available, it sends it through the protocol stack. If a message is
   not available, it simply exits the protocol stack. */
static void slan_link_down(proto_stack_t *stack, slan_message_t
			   *message) {
  slan_link_t *local;
  int len;

  local = (slan_link_t *) stack->data;
  
  len = read(local->fd, message->data, message->maxspace);
#ifdef DEBUG
  fprintf(stderr,"Reading packet from SLAN device\n");
  dump_hex(message->data,len);
#endif
  if (len < 0) {
    /* Debugging stuff */
    syslog(LOG_ERR,"Error reading data from SLAN device socket (%m)\n");
    slan_free(message);
    return;
  }

  message->length = len;

  stack->next->down(stack->next, message);
}

static void slan_link_up(proto_stack_t *stack, slan_message_t
			 *message) {
  slan_link_t *local;
  int len;

  local = (slan_link_t *) stack->data;

  /*
  slan_reserve(message, sizeof(slan_header_t));
  header = (slan_header_t *) message->header;

  header->magic_number = (0xFF00FF00);
  header->our_protocol = htons(ETH_P_SLAN);
  header->orig_protocol = htons(ETH_P_IP);

  slan_set(message);
 
  
  len = sendto(local->fd, message->data, message->length, 0, 
	       (struct sockaddr *) &local->pl_addr,sizeof(struct
	       sockaddr_ll));
  */

  len = write(local->fd, message->data, message->length);
#define DEBUG
#ifdef DEBUG
  fprintf(stderr,"slan_link: Packet sent to slan device.\n");
  dump_hex(message->data,len);
#endif
  if (len < message->length) {
    /* Debugging */
    syslog(LOG_ERR,"Failed to send a packet to SLAN device driver (%m)\n");
    return;
  }

  local->session_obj->bytes_in += message->length;
  local->session_obj->packets_in ++;
  stack->prev->up(stack->prev, message);
}

int slan_link_register(proto_stack_t *stack_entry, int slan_socket, 
		       slan_session_t *session_obj) {
  slan_link_t *local;
  char dev[20];

  stack_entry->up = slan_link_up;
  stack_entry->down = slan_link_down;
  stack_entry->header_reserve = 0;
  stack_entry->trailer_reserve = 0;

  local = (slan_link_t *) slan_calloc(1, sizeof(slan_link_t));
  if (local == NULL) return -1;

  if (slan_socket == 0) {
    strcpy(dev,"tap0");
    local->fd = slan_open_socket(dev);
#ifdef DEBUG
    if (local->fd>0) {
      fprintf(stderr,"Opened device %s\n",dev);
    } else {
      fprintf(stderr,"Failed to open device %s\n",dev);
      exit(-1);
    }
#endif
  } else {
    local->fd = slan_socket;
  }

  local->session_obj = session_obj;

  stack_entry->data = local;
  session_obj->slan_socket = local->fd;

  return 0;
}

